

module RefacDupDef(duplicateDef) where

import PrettyPrint
import PosSyntax
import Maybe
import TypedIds
import List 
import TiPNT 
import RefacUtils 


{-This refactoring duplicates a defintion(function binding or simple pattern binding) at same level with a new name provided
  by the user. The new name should not cause name clash/capture. 
-}

duplicateDef args
 = do let fileName = ghead "filename" args 
          --fileName'= moduleName fileName
          --modName  = Module fileName' 
          newName  = args!!1 
          row      = read (args!!2)::Int
          col      = read (args!!3)::Int
      if isVarId newName 
        then do modName <-moduleName fileName
                (inscps, _, mod, tokList) <-parseSourceFile fileName    
                let pn = locToPName fileName row col mod                                            
                if (pn /= defaultPN) 
                  then do (mod',((tokList',m),_))<- doDuplicating pn newName (inscps, mod, tokList)
                          if wholeModIsExported mod
                           then do clients <- clientModsAndFiles modName
                                   refactoredClients <- mapM (refactorInClientMod modName (findNewPName newName mod')) clients
                                   writeRefactoredFiles $ ((fileName,m),(tokList',mod')):refactoredClients 
                           else  writeRefactoredFiles [((fileName,m), (tokList',mod'))]
                  else error "Invalid cursor position!" 
        else error "Invalid new function name!"   
   
doDuplicating pn newName (inscps, mod, tokList)
   = runStateT (applyTP ((once_tdTP (failTP `adhocTP` dupInMod
                                            `adhocTP` dupInMatch
                                            `adhocTP` dupInPat
                                            `adhocTP` dupInLet
                                            `adhocTP` dupInAlt
                                            `adhocTP` dupInLetStmt)) `choiceTP` failure) mod) ((tokList,unmodified),locOffset)
        where          
        --1. The definition to be duplicated is at top level.
        dupInMod (mod@(HsModule loc name exps imps ds):: HsModuleP)   
          |findFunOrPatBind  pn ds/=[]=doDuplicating' inscps mod pn  
        dupInMod _ =mzero

        --2. The definition to be duplicated is a local declaration in a match
        dupInMatch (match@(HsMatch loc1 name pats rhs ds)::HsMatchP)
          |findFunOrPatBind pn ds/=[]=doDuplicating' inscps match pn                                                    
        dupInMatch _ =mzero

        --3. The definition to be duplicated is a local declaration in a pattern binding
        dupInPat (pat@(Dec (HsPatBind loc p rhs ds))::HsDeclP)
          |findFunOrPatBind pn ds/=[]=doDuplicating' inscps pat pn                    
        dupInPat _=mzero
  
        --4: The defintion to be duplicated is a local decl in a Let expression
        dupInLet (letExp@(Exp (HsLet ds e))::HsExpP)
          |findFunOrPatBind pn  ds/=[]=doDuplicating' inscps letExp pn 
        dupInLet _=mzero
                
        --5. The defintion to be duplicated is a local decl in a case alternative.
        dupInAlt (alt@(HsAlt loc p rhs ds)::(HsAlt (HsExpP) (HsPatP) [HsDeclP]))
          |findFunOrPatBind pn ds/=[]=doDuplicating'  inscps alt pn 
        dupInAlt _=mzero
      
        --6.The definition to be duplicated is a local decl in a Let statement.
        dupInLetStmt (letStmt@(HsLetStmt ds stmts):: (HsStmt (HsExpP) (HsPatP) [HsDeclP]))
           |findFunOrPatBind pn  ds /=[]=doDuplicating' inscps letStmt pn 
        dupInLetStmt _=mzero

        
        failure=idTP `adhocTP` mod
          where
            mod (m::HsModuleP)
              = error "The selected identifier is not a function/simple pattern name, or is not defined in this module "
 
        findFunOrPatBind pn ds = filter (\d->isFunBind d || isSimplePatBind d) $ definingDecls [pn] ds True

        doDuplicating' inscps parent pn 
           = do let decls           = declsIn parent  
                    duplicatedDecls = definingDecls [pn] decls True
                    (after,before)  = break (defines pn) (reverse decls)
                (f,d) <- freeAndDeclaredNames parent  
                 --f: names that might be shadowd by the new name, d: names that might clash with the new name
                dv <- hsVisibleNames pn decls --dv: names may shadow new name
                let inscpsNames = inScopeNames [inscps]
                    vars        = nub (f `union` d `union` dv)
                if elem newName vars || (isInScopeAndUnqualified newName inscps && findEntity pn duplicatedDecls) 
                   then error ("The new name'"++newName++"' will cause name clash/capture or ambiguity problem after "
                               ++ "duplicating, please select another name!")
                   else do  ((toks,_), locOffset)<-get   
                             {-there maybe fun/simple pattern binding and type signature in the duplicated decls
                               function binding, and type signature are handled differently here: the comment and layout
                               in function binding are preserved.The type signature is outputted by pretty printer, so
                               the comments and layout are NOT preserved.
                             -}
                            let funBinding = filter isFunOrPatBind duplicatedDecls --get the fun binding.
                                typeSig    = filter isTypeSig duplicatedDecls  --get the type signature.
                                --get the start-end location of fuction binding
                                ((startRow,startCol),(endRow,endCol))
                                      = startEndLocIncComments {-startEndLoc-} toks funBinding  
                                {-take those tokens before (and include) the function binding and its following
                                 --white tokens before the 'new line' token. (some times the function may be followed by 
                                 --comments) -}
                                (toks1,_) = break (\t->tokenPos t>(endRow,endCol)) toks
                                 --take those token after (and include) the function binding
                                toks2     = dropWhile (\t->tokenPos t<(startRow,startCol) || isNewLn t) toks
                            put ((toks2,modified),locOffset)     
                            --rename the fun name to the user specified name, and update token stream as well
                            funBinding<-renamePNwithUpdLexer pn newName funBinding
                            --rename function name in type signature  without adjusting the token stream
                            typeSig<-renamePN pn newName typeSig
                            ((toks2,_),locOffset)<-get   -- combine two token list and pretty printed type signature.
                                                       --CommentStart hides the newline charachter the comment,
                            let newLineTok = if toks1/=[] && endsWithNewLn (glast "doDuplicating" toks1)
                                               then [newLnToken]
                                               else [newLnToken,newLnToken] 
                                toks' =
                                  if typeSig /= [] 
                                    then let offset = tokenCol ((ghead "doDuplicating") (dropWhile (\t->isWhite t) toks2))
                                             sigSource = concatMap (\s->whiteSpaces (offset-1)++s++"\n")((lines.render.ppi) typeSig)
                                             t = mkToken Whitespace (0,0) sigSource   --The token type & pos does not matter here.
                                         in  (toks1++newLineTok++[t]++(whiteSpaceTokens (0,0) (startCol-1))++toks2)
                                    else (toks1++newLineTok++(whiteSpaceTokens (0,0) (startCol-1))++toks2) 
                            --make the location information in the duplicated definition unique
                            newBinding <- mkLocsUnique locOffset  $ typeSig++funBinding 
                            put ((toks',modified),(fst locOffset*2, snd locOffset*2))
                            return (replaceDecls parent (reverse before++ newBinding++ reverse after)) 
   


mkLocsUnique::(MonadPlus m, Term t)=>(Int,Int)->t->m t
mkLocsUnique (rowOffset,colOffset) t 
  = mkLocsUnqiue' (allSortedLocs t) t      
  where 
    mkLocsUnqiue' allSourceLocs t=applyTP (full_td (idTP `adhocTP` inLoc)) t
       where inLoc loc@(SrcLoc fileName row col)    --SHOULD ALSO COMPARE FILENAME TO AVOID COINCIDENCE.
               = if isJust (find (==(row, col)) allSourceLocs) --'returning True' means that the id is a declared locally
                   then return (SrcLoc fileName (row+rowOffset) (col+colOffset))
                   else return loc

--Find the the new definition name in PName format. 
findNewPName name 
  =(fromMaybe defaultPN). applyTU (once_buTU (failTU `adhocTU` worker))
     where 
        worker  pname 
           |pNtoName pname == name = Just pname
        worker _ =Nothing 

                   
--Do refactoring in the client module.
-- that is to hide the identifer in the import declaration if it will cause any problem in the client module.
refactorInClientMod serverModName newPName (modName, fileName)
  = do (inscps, exps ,mod ,ts) <- parseSourceFile fileName
       let modNames = willBeUnQualImportedBy serverModName mod   
       if isJust modNames && needToBeHided (pNtoName newPName) exps mod
        then do (mod', ((ts',m), _))<-runStateT (addHiding serverModName mod [newPName]) ((ts,unmodified),(fileName, initValue))
                return ((fileName,m), (ts',mod')) 
        else return ((fileName,unmodified),(ts,mod))
   where
     needToBeHided name exps mod
         =(findNameInUseLoc name (hsModDecls mod)
          || findUnqualEntInExport name (hsModExports mod)
          || causeNameClashInClientExports name mod exps)
          

--Check here:
--get the module name or alias name by which the duplicated definition will be imported automatically.
willBeUnQualImportedBy modName mod
   = let imps = hsModImports mod
         ms   = filter (\(HsImportDecl _ modName1 qualify  as h)->modName==modName1 && (not qualify) && 
                          (isNothing h || (isJust h && ((fst (fromJust h))==True)))) imps
         in if ms==[] then Nothing
                      else Just $ nub $ map getModName ms

         where getModName (HsImportDecl _ modName qualify  as h)
                 = if isJust as then fromJust as
                                else modName