
module RefacLocUtils(module HsTokens,module RefacLocUtils) where

import PosSyntax
import UniqueNames 
import HsLexerPass1 hiding (notWhite)
import HsTokens
import PrettyPrint
import HsExpUtil
import PNT 
-------------------------
import DriftStructUtils  
import StrategyLib
------------------------
import RefacTypeSyn
import Maybe
import List

type Pos = (Int,Int) 
type TokenStream =[(Pos,String)]
type Dict=[(Pos,Pos)]
type HToken=(Token,(Pos,String))


unmodified = 0
modified   = 1

ghead info (h:_) = h
ghead info []    = error $ "ghead "++info++" []"

gtail info (_:h) = h
gtail info [] = error $ "gtail " ++ info ++ " []"

glast info []      = error $ "glast " ++ info ++ " []"
glast info h@(_:_) = last h

defaultPos=(0,0)

getASTTokens = applyTU (full_tdTU (constTU []  `adhocTU` pnt
                                               `adhocTU` literal))
    
   where 
     pnt (PNT pname _ (N (Just (SrcLoc _ row col))))
           = return [((row,col),(pNtoName pname))]
     pnt _ = return []
                                      
     literal (HsInt (SrcLoc _ row col) int)
              = return [((row,col), show int)]
     literal _= return [] 
          
     pNtoName (PN (UnQual i) orig) = i
     pNtoName (PN (Qual modName i) orig) = i

getLexTokens text=filter keep (lexerPass1 text)   
     where 
        keep token =case (fst token) of 
                            Varid       ->True
                            Conid       ->True
                            Varsym      ->True
                            Consym      ->True
                            Qvarid      ->True
                            Qconid      ->True
                            Qvarsym     ->True
                            Qconsym     ->True 
                            IntLit      ->True
                            FloatLit    ->True
                            CharLit     ->True 
                            StringLit   ->True
                            _           ->False

mkDict::TokenStream->TokenStream->Dict
mkDict astTokens lexTokens
   = zipWith (\x y->(fst x, fst y))  astTokens (tail lexTokens)  --'tail' used to get rid of the module name token.

updateAST source mod
   = do let lexTokens = (map snd) (getLexTokens source)
        astTokens<-getASTTokens mod
        let dict = mkDict astTokens lexTokens
        doUpdating dict mod
   where 
      doUpdating::(Monad m,Term t)=>[(Pos,Pos)]->t->m t 
      doUpdating dict =applyTP (full_tdTP (idTP `adhocTP` srcloc))
          where 
               srcloc (SrcLoc fileName row col)
                   =let pos2=lookUpDict (row,col) dict
                    in  if isJust pos2 
                           then return (SrcLoc fileName ((fst.fromJust) pos2) ((snd.fromJust) pos2))
                           else return (SrcLoc fileName row col)  --Problem here?


lookUpDict:: Pos->Dict->Maybe Pos
lookUpDict pos []=Nothing
lookUpDict pos (p:ps)
       =if pos==(fst p) then Just (snd p)
                        else lookUpDict pos ps                       
                           

-- **************FUNCTIONS INVOLVING  COMMENT& LAYOUT PRESERVING**********************
--Compose a new token using the given parameters.
mkToken tokenType pos tokenContent =(tokenType,(pos,tokenContent))

--replace the content of a Token
renameToken (t,(p,_)) newContent = (t,(p,newContent))

whiteSpaces n=replicate n ' '

--Create a list of white space tokens.
whiteSpaceTokens (row,col) n
  |n>0        = (Whitespace,((row,col)," ")):(whiteSpaceTokens (row,col+1) (n-1))
  |otherwise  = []

notWhite  = not.isWhite
isWhiteSpace (t,(_,s))       = t==Whitespace && s==" "
isNewLn (t,(_,s))            = t==Whitespace && s=="\n"
isWhite (t,_)                = t==Whitespace || t==Commentstart || t==Comment || t==NestedComment
isCommentStart (t,(_,s))     = t==Commentstart && s=="--" 
isComment (t,(_,s))          = t==Comment || t ==NestedComment
isNestedComment (t,(_,s))    = t==NestedComment
isMultiLineComment (t,(_,s)) = t==NestedComment && (isJust (find (=='\n') s))

isOpenBracket  (t,(_,s))       = t==Special && s=="("
isCloseBracket (t,(_,s))       = t==Special && s==")"
isOpenSquareBracket  (t,(_,s)) = t==Special && s=="[" 
isCloseSquareBracket (t,(_,s)) = t==Special && s=="]"
isOpenBrace  (t,(_,s))         = t==Special && s=="{" 
isCloseBrace (t,(_,s))         = t==Special && s=="}" 
isComma (t,(_,s))              = t==Special && s=="," 

isConid (t,(_,_))              = t==Conid

isLit (t,(_,s)) = t==IntLit || t==FloatLit || t==CharLit || t==StringLit

isWhereOrLet  t   = isWhere t || isLet t
isWhere (t,(_,s)) = t==Reservedid && s=="where"
isLet   (t,(_,s)) = t==Reservedid && s=="let"
isIn    (t,(_,s)) = t==Reservedid && s=="in" 
isCase  (t,(_,s)) = t==Reservedid && s=="case"  
isDo    (t,(_,s)) = t==Reservedid && s=="do"  
isIf    (t,(_,s)) = t==Reservedid && s=="if"
isForall (t,(_,s)) =t==Reservedid && s=="forall"

isEqual  (t,(_,s))   = t==Reservedop && s=="="
isLambda (t,(_,s))   = t==Reservedop && s=="\\"
isIrrefute (t,(_,s)) = t==Reservedop && s=="~"
isBar   (t,(_,s))    = t==Reservedop && s=="|"

isMinus (t,(_,s))    = t==Varsym && s=="-"

isHiding (t,(_,s))   = s=="hiding" 
  
defaultToken = (Whitespace,(defaultPos," "))    
newLnToken   = (Whitespace, (defaultPos,"\n"))  

--reture True if the last char in the token is '\n'
endsWithNewLn (t,(_,s)) =if s==[] then False 
                                  else (glast "endsWithNewLn"  s=='\n')

--return True if the first char in the token is `\n`
startsWithNewLn (t,(_,s))=if s==[] then False 
                                   else ((ghead "starsWithNewLn" s)=='\n')

startsWithEmptyLn toks=isJust (find isNewLn $ takeWhile (\t->isWhiteSpace t || isNewLn t) toks)

--Return True if there is a '\n' in the token.
findNewLn (_,(_,s))=isJust (find (=='\n') s) {-compare with: isNewLn -}

tokenCol (_,((r,c),_)) = c  
tokenRow (_,((r,c),_)) = r
tokenPos (_,(p,_))     = p
tokenCon (_,(_,s))     = s
tokenLen (_,(_,s))     = length s

setDefaultLoc toks =map (\(t,(_,s))->(t,(defaultPos,s))) toks

--lastNotSpaceToken only used in 'deleteToks', the 'defaultToken' does not cause problems.
lastNotSpaceToken toks=case dropWhile isWhiteSpace (reverse toks) of
                         [] ->defaultToken
                         l -> ghead "lastNotSpaceToken" l

--used in RefacMoveDef.hs. the default value does not cause problems.
firstNotSpaceToken toks=case dropWhile isWhiteSpace toks of
                         [] ->defaultToken
                         l -> ghead "firstNotSpaceToken" l  

--compress multi newlns and the beginning of toks into one newLn. extra
--newLn may be cause by deleting a piece of code while without deleting the following newLns.
--WILL THIS CHANGE THE LAYOUT?
compressPreNewLns toks= let (toks1, toks2)=break (not.(\t->isNewLn t || isWhiteSpace t)) toks
                            newLns=filter isNewLn toks1
                        in if length newLns>1 then (last newLns):toks2  --no problem with 'last'
                                              else toks

compressEndNewLns toks=let (toks1, toks2)=break (not.(\t->isNewLn t || isWhiteSpace t)) (reverse toks)
                           newLns=filter isNewLn toks1
                       in if length newLns>1 then reverse ((ghead "compressEndNewLns" newLns):toks2) --no problem with 'head'
                                             else toks

--Give a token stream covering mutli-lines, calculate the length of the last line  
--This function is used in RefacGenDef.hs
lengthOfLastLine toks=let (toks1,toks2)=break findNewLn $ reverse toks
                      in  if toks2==[] then sum (map tokenLen toks1)
                                       else sum (map tokenLen toks1)+lastLineLenOfToken (ghead "lengthOfLastLine"  toks2)
           where 
               --compute the length of a token, if the token covers multi-line, only count the last line
               --The function  DOES NOT cosider the tab keys
              lastLineLenOfToken (_,(_,s))=(length.(takeWhile (\x->x/='\n')).reverse) s

--get a token stream specified by the start and end position.
getToks::(Pos,Pos)->[HToken]->[HToken]
getToks (startPos,endPos) toks
    =let (_,toks2)=break (\t->tokenPos t==startPos) toks
         (toks21,_)=break (\t->tokenPos t>endPos) toks2 
     in toks21

--COMBINE THIS TO 'replaceToks'

replaceToks1 startPos endPos newToks
  = do ((toks,_),others)<-get                           
       let toks'=replaceToks toks startPos endPos newToks
       put ((toks',modified),others)

--Replace a list of tokens in the token stream by a new list of tokens, adjust the layout as well.
--Conditions to use this function:
--1. Should make sure the start and end positions really exist in the token stream.

replaceToks::[HToken]->Pos->Pos->[HToken]->[HToken]

replaceToks toks startPos endPos newToks
   = let (toks1,toks2)=break (\t->tokenPos t>=startPos) toks
         (toks21,toks22)=break (\t->tokenPos t>endPos) toks2
         nextPos=if toks22==[] then endPos
                               else (tokenPos.ghead "replaceToks") toks22
     in  case (isSingleLine toks21, isSingleLine newToks) of
                 (True,True)   -> replaceToksCase1 toks1 toks21 toks22 nextPos  
                 (True,False)  -> replaceToksCase2 toks1 toks21 toks22 nextPos  
                 (False,True)  -> replaceToksCase3 toks1 toks21 toks22 nextPos  
                 (False,False) -> replaceToksCase4 toks1 toks21 toks22 nextPos 
     where 
         
         --'singleLine' test whether all tokens in 'toks' are in the same line.
         --This function assumes '\n' can only appear at the end of the token.
         --So be careful when creating new tokens from the pretty printer output, which 
         --means the source from pretty printer should be put into several tokens if
         --it covers more than one lines. 
         isSingleLine toks=isNothing $ find findNewLn toks

         lengthOfToks=length.(concatMap tokenCon)

          --replace toks in a single line by toks in a single line. 
         replaceToksCase1 toks1 toks21 toks22 nextPos
            =let oldToksLength=lengthOfToks toks21   
                 newToksLength=lengthOfToks newToks 
             in  toks1++newToks++adjustLayout (newToksLength-oldToksLength) nextPos toks22 

         --replace toks in a single line by toks in multi lines.
         replaceToksCase2 toks1 toks21 toks22 nextPos
           =let oldOffset=lengthOfLastLine toks1+lengthOfToks toks21
                newOffset=lengthOfLastLine newToks 
            in toks1++newToks++adjustLayout (newOffset-oldOffset) nextPos toks22

         --replace toks in multi lines by toks in a single line. 
         replaceToksCase3 toks1 toks21 toks22 nextPos
           =let oldOffset=lengthOfLastLine toks21 
                newOffset=lengthOfLastLine toks1+lengthOfToks newToks 
            in toks1++newToks++adjustLayout (newOffset-oldOffset) nextPos toks22

         --replace toks in multi lines by toks in multi lines.  
         replaceToksCase4 toks1 toks21 toks22 nextPos
           =let oldOffset=lengthOfLastLine toks21 
                newOffset=lengthOfLastLine newToks                   
            in toks1++newToks++adjustLayout (newOffset-oldOffset) nextPos toks22
                   
{-Delete a sequence of tokens specified by the start position and end position from the token stream,
  then adjust the remaining token stream to preserve layout-}

deleteToks::[HToken]->Pos->Pos->[HToken]
deleteToks toks startPos@(startRow, startCol) endPos@(endRow, endCol)
    =let (toks1,toks2)=break (\t->tokenPos t>=(startRow,0)) toks
         (toks21,toks22)=break (\t->tokenPos t>=(endRow+1, 0)) toks2
         --'before'is the tokens preceeding the tokens to be deleted at 'startRow'
         before=takeWhile (\t->tokenPos t<startPos) toks21  
         --'after' is the tokens following the tokens to be deleted at 'endRow'
         after =dropWhile (\t->tokenPos t<=endPos) toks21
     in if after==[]  
             then let toks22'= --handle deleting multi-line comment
                            if isMultiLineComment (lastNotSpaceToken toks21) && toks22/=[] && tokenCol (ghead "deleteToks" toks22)>1  
                                 then  let (row,col)=tokenPos $ ghead "deleteToks" toks22   --no problem with toks22
                                       in whiteSpaceTokens (row,1) (col-1) ++ toks22 
                                 else toks22
                      toks1AndBefore=toks1++before
                  in  if (toks1AndBefore/=[] && endsWithNewLn (glast "deleteToks" toks1AndBefore)) 
                           || (toks22'/=[] && startsWithNewLn (ghead "deleteToks" toks22'))
                       then  toks1++before++toks22'
                       --avoiding layout adjustment by adding a `\n', sometimes may produce extra lines.
                       else  toks1++before++[newLnToken]++toks22'
             else let nextTok=ghead "deleteToks" after  --on problem with 'head'
                  in toks1++before++adjustLayout (startCol-tokenCol nextTok) (tokenPos nextTok) (after++toks22) 

  
--adjust the token stream to reflect the layout change because of adding/removing tokens
adjustLayout::Int->Pos->[HToken]->[HToken]
adjustLayout 0 _ toks = toks  
adjustLayout n pos toks 
   =if layoutRuleApplies pos toks 
      then if n>0 
             then {-start adjusting from next row by adding n whitespaces at the beginnig of each row affected-}
                  addSpaces n (snd pos) toks
             else {-start adjusting from next row by removing n whitespaces from the beginning of each row affected-} 
                  rmSpaces (-n)(snd pos) toks  
      else toks --no need to adjust layout.

    where 

      {-Test whether there is a layout-rule-invoking token (i.e.'where','let','do' or 'of') in the current
        line after position 'pos'. -}
      layoutRuleApplies::Pos->[HToken]->Bool
      layoutRuleApplies pos toks
         = --t is the remained toks in current line from position 'pos'
           let t=dropWhile (not.isSpecialToken) $ filter notWhite $ takeWhile (not.findNewLn) toks
           in length t>1 && tokenCon (t!!1)/="{"           
         where 
            isSpecialToken t=elem (tokenCon t) ["where","let","do","of"]            
  
      {-add/remove n whitespace tokens to/from each row affected by the current layout in the remaining token list.  
       n: the number of spaces to be added/removed
       col is used to determine whether the adjusting should continue.-}
      addSpaces,rmSpaces::Int->Int->[HToken]->[HToken]

      addSpaces n col toks
          =let groupedToks=groupBy (\x y->tokenRow x==tokenRow y) toks
           in  if groupedToks==[] 
                  then [] 
                  else ghead "addSpaces" groupedToks ++ concatMap (insertWhites n col) (tail groupedToks) 
          
         where    
            --Insert n whitespace tokens at the beginning. 
            insertWhites n col toks=let t=firstNotWhiteTok toks
                                    in if t/=defaultToken && tokenCol t>=col
                                          then doAddWhites n toks
                                          else toks 
   
      rmSpaces n col toks
         =let groupedToks=groupBy (\x y->tokenRow x==tokenRow y) toks
          in if groupedToks==[]
               then [] 
               else  ghead "rmSpaces" groupedToks ++ concatMap (rmWhites n col) (tail groupedToks)
           
        where 
         --Remove n whitespaces from the beginning.
         --Assumption: assume there are at least n whitespaces.
         --PROBLEM: This Assumption can not always be guaranteed if
         --comments are allowed to be followed by code in the same line.
         rmWhites n col toks
           =let t=firstNotWhiteTok toks
            in if t/=defaultToken && tokenCol t>col
                then doRmWhites n toks
                else toks 
      
      --Get the first not-white token.
      firstNotWhiteTok toks 
          = fromMaybe defaultToken (find notWhite toks)
  

doRmWhites 0 ts=ts  
doRmWhites n []=[]
doRmWhites n toks@(t:ts)=if isWhiteSpace t then doRmWhites (n-1) ts
                                           else toks 

doAddWhites n []=[]
doAddWhites n ts@(t:_)= whiteSpaceTokens (tokenRow t,0) n ++ts

--Starting from the second line, add/remove 'abs offset' number of white spaces before each line.
--if offset>0 then add white spaces, otherwise remove white spaces.

adjustOffset::Int->[HToken]->Bool->[HToken]           
adjustOffset offset [] _ = []  
adjustOffset offset toks firstLineIncluded  
     = let groupedToks = groupBy (\x y->tokenRow x==tokenRow y) toks  --groupedToks/=[], no problem with 'head'
           --if firstLineIncluded is False, the offset of the first line won't be ajusted.
       in if offset>=0 then if firstLineIncluded 
                               then concatMap (doAddWhites offset) groupedToks
                               else ghead "adjustOffset" groupedToks ++ concatMap (doAddWhites offset) (tail groupedToks)
                       else if firstLineIncluded 
                               then concatMap (doRmWhites  (-offset)) groupedToks
                               else ghead "adjustOffset" groupedToks ++ concatMap (doRmWhites  (-offset)) (tail groupedToks) 

astToSource=toLines.replaceTabBySpaces.render.ppi
    where
       --remove those blank lines. (Is this necessary?) 
       toLines=filter notBlankLine.lines 
 
       notBlankLine s=isJust $ find (\c->c/=' ' && c/='\n' && c/='\t') s

       --Replace Tab by white spaces. (1 Tab=8 white spaces)
       replaceTabBySpaces::String->String
       replaceTabBySpaces []=[]
       replaceTabBySpaces (s:ss)
           =if s=='\t' then whiteSpaces 8 ++replaceTabBySpaces ss
                       else s:replaceTabBySpaces ss

--Given several lines of source, put the source into tokens.
mkTokensNoFstLnIndent  tokenType (row,col) []=[]
mkTokensNoFstLnIndent  tokenType (row,col) (line:lines)
    =if lines==[] then [mkToken tokenType (row,col) line]
                  else [(mkToken tokenType (row,col) line),newLnToken]++
                       mkTokensWithIndent tokenType (row+1,col) lines False
             
mkTokensWithIndent tokenType (row,col) [] addSpace=[]
mkTokensWithIndent tokenType (row,col) (line:lines) addSpace
  = if lines/=[]  
       then  lineToTokens tokenType (row,col) line
              ++mkTokensWithIndent tokenType (row+1,col) lines addSpace
       else  if addSpace
                then  whiteSpaceTokens (row,1) (col-1) ++ [mkToken tokenType (row,col) (line++"  ")]
                else  whiteSpaceTokens (row,1) (col-1) ++ [mkToken tokenType (row,col) line]
   where 
        lineToTokens tokenType (row,col) line
           =whiteSpaceTokens (row,1) (col-1) ++ [mkToken tokenType (row,col) line,
                 mkToken Whitespace (row, col+length line) "\n"] 
 

-------------------------------------------------------------------------------------------------
{- given a piece of AST (declaration, expression ,etc), 
  'startEndLoc' get it start&end position in the program source. -}

--CHECK & RECONSIDER  THIS CLASS AGAIN.
class StartEndLoc t where

   startEndLoc :: [HToken]->t->(Pos,Pos)

instance StartEndLoc HsExpP where
    
  startEndLoc  toks (Exp e)=
      case e of
    
        HsId ident@(HsVar (PNT pn _ _)) -> case pn of 
                                            PN (UnQual "+") (G (Module "Prelude") "+"  _)
                                                ->let (startLoc,endLoc)=startEndLoc toks ident
                                                  in extendBothSides toks startLoc endLoc isOpenBracket isCloseBracket
                                            PN (UnQual "-") (G (Module "Prelude") "-"  _)
                                                ->let (startLoc,endLoc)=startEndLoc toks ident
                                                  in extendBothSides toks startLoc endLoc isOpenBracket isCloseBracket
                                            PN (UnQual "*") (G (Module "Prelude") "*"  _)
                                                ->let (startLoc,endLoc)=startEndLoc toks ident
                                                  in extendBothSides toks startLoc endLoc isOpenBracket isCloseBracket
                                            _   ->startEndLoc toks ident    
        HsId  x                       ->startEndLoc toks x
     

        HsLit (SrcLoc _ r c) _        -> ((r,c),(r,c))

        HsInfixApp e1 op e2           ->let (startLoc,_)=startEndLoc toks e1
                                            (_, endLoc) =startEndLoc toks e2
                                        in (startLoc,endLoc)

        HsApp e1 e2                   ->let (startLoc,_)=startEndLoc toks e1
                                            (_, endLoc )=startEndLoc toks e2
                                        in (startLoc,endLoc)

        HsNegApp (SrcLoc _ r c) e     ->let (_,endLoc)=startEndLoc toks e
                                        in ((r,c), endLoc)

        HsLambda ps e                 ->let (startLoc,_)=startEndLoc toks (ghead "startEndLoc:HsLambda" ps)  --ps can not be empty
                                            (_,endLoc)  =startEndLoc toks e
                                        in extendForwards toks startLoc endLoc isLambda

        HsIf e1 e2 e3                 ->let (startLoc, _)=startEndLoc toks e1
                                            (_, endLoc)=startEndLoc toks e2
                                        in extendForwards toks  startLoc endLoc isIf

        HsLet ds e                    ->if ds==[]
                                          then
                                            let  (startLoc,endLoc)=startEndLoc toks e
                                            in extendForwards toks startLoc endLoc isLet
                                          else
                                            let  (startLoc,_)=startEndLoc toks (ghead "startEndLoc:HsLet" ds) 
                                                 (_,endLoc)  =startEndLoc toks e 
                                            in extendForwards toks startLoc endLoc isLet

        HsCase e alts                 ->let (startLoc,_)=startEndLoc toks e
                                            (_,endLoc)  =startEndLoc toks (glast "HsCase" alts) --alts can not be empty.
                                        in extendForwards toks startLoc endLoc isCase
                        
        HsDo stmts                    ->let (startLoc, endLoc)=startEndLoc toks  stmts
                                        in extendForwards toks startLoc endLoc isDo

        HsTuple es                    ->if es==[] 
                                         then (defaultPos,defaultPos)  --Check this
                                         else let (startLoc,_)=startEndLoc toks (ghead "startEndLoc:HsTuple" es)
                                                  (_,endLoc)  =startEndLoc toks (glast "startEndLoc:HsTuple" es)
                                              in extendBothSides toks startLoc endLoc isOpenBracket isCloseBracket  

        HsList es                     ->if es==[] 
                                         then (defaultPos,defaultPos)  --Check this
                                         else let (startLoc,_)=startEndLoc toks (ghead "startEndLoc:HsList" es) 
                                                  (_,endLoc)  =startEndLoc toks (glast "startEndLoc:HsList" es)
                                              in extendBothSides toks startLoc endLoc isOpenSquareBracket isCloseSquareBracket  

        HsParen e                     ->let (startLoc,endLoc)=startEndLoc toks  e
                                        in extendBothSides toks startLoc endLoc isOpenBracket isCloseBracket

        HsLeftSection e op            ->let (startLoc,_)=startEndLoc toks e
                                            (_, endLoc )=startEndLoc toks op
                                        in (startLoc,endLoc)

        HsRightSection op e           ->let (startLoc,_)=startEndLoc toks op
                                            (_, endLoc )=startEndLoc toks op
                                        in (startLoc,endLoc)

        HsRecConstr i upds            ->let (startLoc,_)=startEndLoc toks i
                                            (_,endLoc)  =startEndLoc toks (glast "startEndLoc:HsRecConstr" upds) --can 'upds' be empty?
                                        in extendBackwards toks startLoc endLoc isCloseBrace 
                                        
        HsRecUpdate e upds            ->let (startLoc,_)=startEndLoc toks e
                                            (_,endLoc)  =startEndLoc toks (glast "startEndLoc:HsRecUpdate" upds) --ditto
                                        in extendBackwards toks startLoc endLoc isCloseBrace 

        HsEnumFrom e                  ->let (startLoc,endLoc)=startEndLoc toks e
                                        in extendBothSides toks startLoc endLoc isOpenSquareBracket isCloseSquareBracket   

        HsEnumFromTo e1 e2            ->let (startLoc,_)=startEndLoc toks e1
                                            (_,  endLoc)=startEndLoc toks e2
                                        in extendBothSides toks startLoc endLoc isOpenSquareBracket isCloseSquareBracket   

        HsEnumFromThen e1 e2          ->let (startLoc,_)=startEndLoc toks e1
                                            (_,  endLoc)=startEndLoc toks e2
                                        in extendBothSides toks startLoc endLoc isOpenSquareBracket isCloseSquareBracket  
                                  
        HsEnumFromThenTo e1 e2 e3     ->let (startLoc,_)=startEndLoc toks e1
                                            (_,  endLoc)=startEndLoc toks e3
                                        in extendBothSides toks startLoc endLoc isOpenSquareBracket isCloseSquareBracket  
                                 
        HsListComp stmts              ->let (startLoc,endLoc)=startEndLoc toks stmts
                                        in  extendBothSides toks startLoc endLoc isOpenSquareBracket isCloseSquareBracket     

        HsAsPat i e                   ->let (startLoc,_)=startEndLoc toks i
                                            (_,endLoc)=  startEndLoc toks e
                                        in (startLoc,endLoc) 

        HsIrrPat e                    ->let (startLoc,endLoc)=startEndLoc toks e
                                        in extendForwards toks startLoc endLoc isIrrefute

        HsWildCard                    ->(defaultPos,defaultPos)  --SHOULD BE MODIFIED LATER!
  
        
        HsExpTypeSig loc e c t        ->let (startLoc,_)=startEndLoc toks e
                                            (_, endLoc )=startEndLoc toks t
                                        in (startLoc,endLoc)    

instance StartEndLoc HsTypeP where

   startEndLoc toks (Typ p)=
      case p of 
        HsTyFun  t1  t2       ->   let (startLoc,_)=startEndLoc toks t1
                                       (_ , endLoc)=startEndLoc toks t2
                                   in (startLoc,endLoc)  
        --HsTyTuple [t]         ->

        HsTyApp  t1 t2        ->   let (startLoc,_)=startEndLoc toks  t1 
                                       (_ , endLoc)=startEndLoc toks  t2
                                   in (startLoc,endLoc)

        HsTyVar  i            ->   startEndLoc toks i
        HsTyCon  i            ->   startEndLoc toks i

        HsTyForall   is  t   ->   case is of
                                     []  ->let (startLoc,endLoc)=startEndLoc toks t
                                           in  extendForwards toks startLoc endLoc isForall
   
                                     l  -> let (startLoc, _) =startEndLoc toks  $ ghead "StartEndLoc:HsTypeP" is
                                               ( _ , endLoc) =startEndLoc toks t
                                           in extendForwards toks startLoc endLoc isForall
                               
        
                                  


instance StartEndLoc HsPatP where

  startEndLoc toks (Pat p)=
     case p of
       HsPId i                          ->startEndLoc toks i

       HsPLit (SrcLoc _ r c) _          ->((r,c),(r,c))

       HsPNeg (SrcLoc _ r c) p          ->((r,c),(r,c))
   
       HsPInfixApp p1 op p2             ->let (startLoc,_)=startEndLoc toks  p1
                                              (_ , endLoc)=startEndLoc toks p2
                                          in (startLoc,endLoc)

       HsPApp i ps                      ->let (startLoc,_)=startEndLoc toks  i
                                              (_,endLoc)=startEndLoc toks (glast "StartEndLoc:HsPatP" ps)
                                          in (startLoc,endLoc)

       HsPTuple ps                      -> if ps==[]
                                             then  (defaultPos,defaultPos)  --Check this
                                             else let (startLoc,_)=startEndLoc toks (ghead "startEndLoc:HsPTuple"  ps)
                                                      (_,endLoc)=startEndLoc toks (glast "startEndLoc:HsPTuple" ps)
                                                  in extendBothSides toks startLoc endLoc isOpenBracket isCloseBracket  

       HsPList ps                       ->if ps==[] 
                                            then (defaultPos,defaultPos)  --Check this
                                            else let (startLoc,_)=startEndLoc toks (ghead "startEndLoc:HsPList" ps)
                                                     (_, endLoc) =startEndLoc toks (glast "startEndLoc:HsPList" ps)
                                            in  extendBothSides toks startLoc endLoc isOpenSquareBracket isCloseSquareBracket     
                                        
       HsPParen p                       ->let (startLoc,endLoc)=startEndLoc toks p
                                              (toks1,toks2)=break (\t->tokenPos t==startLoc) toks
                                          in extendBothSides toks startLoc endLoc isOpenBracket isCloseBracket  

       HsPRec i upds                    ->let (startLoc,_)=startEndLoc toks i
                                              (_,endLoc)=startEndLoc toks (glast "startEndLoc:HsPRec" upds) --can upds be empty?
                                          in extendBackwards toks startLoc endLoc isCloseBrace 
                                         
       HsPAsPat i p                     ->let (startLoc,_)=startEndLoc toks i
                                              (_,endLoc)=startEndLoc toks p
                                          in (startLoc,endLoc)
 
       HsPIrrPat p                      ->let (startLoc,endLoc)=startEndLoc toks p
                                              toks1=takeWhile (\t->tokenPos t<startLoc) toks  
                                          in extendForwards toks startLoc endLoc isIrrefute
                                       
       HsPWildCard                       ->(defaultPos,defaultPos)

instance StartEndLoc [HsPatP] where

   startEndLoc toks ps = let locs=(nub.(map (startEndLoc toks))) ps \\ [(defaultPos,defaultPos)]
                         in if locs==[] then (defaultPos,defaultPos)
                                        else let (startLoc,_)=ghead "StartEndLoc:HsPatP" locs
                                                 (_,endLoc) =glast "StartEndLoc:HsPatP"  locs
                                             in (startLoc,endLoc)
instance StartEndLoc [HsExpP] where

   startEndLoc toks es=let locs=(nub.(map (startEndLoc toks))) es \\ [(defaultPos,defaultPos)]
                       in if locs==[] then (defaultPos,defaultPos)
                                      else let (startLoc,_)=ghead "StartEndLoc:HsExp" locs
                                               (_,endLoc) =glast "startEndLoc:HsExp" locs
                                           in (startLoc,endLoc)
instance StartEndLoc [HsDeclP] where
  startEndLoc toks ds=if  ds==[] then (defaultPos,defaultPos)
                                 else if length ds==1 
                                        then startEndLoc toks (ghead "StartEndLoc:[HsDeclP]" ds)
                                        else  let (startLoc,_)=startEndLoc toks (ghead "StartEndLoc:[HsDeclP]" ds)
                                                  (_,endLoc) =startEndLoc toks (glast  "StartEndLoc:[HsDeclP]" ds)
                                              in (startLoc,endLoc)

instance StartEndLoc HsMatchP where
   startEndLoc toks (HsMatch loc i ps rhs ds)
         =let (startLoc,_)=startEndLoc toks i
              (_,endLoc)  =if ds==[] then startEndLoc toks rhs
                                     else startEndLoc toks (glast "StartEndLoc:HsMatchP" ds)
          in (startLoc,endLoc)
                                  


instance StartEndLoc HsStmtP where    
   startEndLoc toks stmts=let s=getStmtList  stmts
                          in startEndLoc toks (glast "StartEndLoc:HsStmtP" s)
                  
                              
instance  StartEndLoc (HsStmtAtom HsExpP HsPatP [HsDeclP])  where

    startEndLoc toks stmt=
      case stmt of 
           HsGeneratorAtom (SrcLoc _ r c) p e -> -- Claus
                                  let (startLoc,_)=startEndLoc toks p
                                      (_,endLoc)  =startEndLoc toks e
                                  in ((r,c),endLoc) -- Claus
           HsQualifierAtom e   -> startEndLoc toks e
           HsLetStmtAtom ds    -> if ds==[]
                                   then (defaultPos,defaultPos)
                                   else let (startLoc,_)= startEndLoc toks (ghead "StartEndLoc:HsStmtAtom" ds)
                                            (_,endLoc)  = startEndLoc toks (glast "StartEndLoc:HsStmtAtom" ds)
                                        in (startLoc,endLoc)
           HsLastAtom e        ->startEndLoc toks e


instance (StartEndLoc i,StartEndLoc e)=>StartEndLoc (HsFieldI i e) where

    startEndLoc toks (HsField i e)=let (startLoc,_)=startEndLoc toks i
                                       (_,endLoc)=startEndLoc toks e
                                   in (startLoc,endLoc) 

instance StartEndLoc HsAltP where
   
    startEndLoc toks (HsAlt l p rhs ds)=let (startLoc,_)=startEndLoc toks p
                                            (_,endLoc)=if ds==[] then startEndLoc toks rhs
                                                                 else startEndLoc toks (glast "StartEndLoc:HsAltP" ds)
                                        in (startLoc,endLoc)

instance StartEndLoc RhsP where
    
   startEndLoc toks (HsBody e)=startEndLoc toks e

   startEndLoc toks (HsGuard es)=if es==[] then (defaultPos,defaultPos)
                                           else let (_,e1,_)=ghead "StartEndLoc:RhsP" es
                                                    (_,_,e2)=glast "StartEndLoc:RhsP" es
                                                    (startLoc,_)=startEndLoc toks e1
                                                    (_,endLoc)=startEndLoc toks e2
                                                in (startLoc,endLoc) 

instance StartEndLoc (HsIdentI PNT) where

    startEndLoc toks ident=
       case ident of
           HsVar i  ->startEndLoc toks i
           HsCon i  ->startEndLoc toks i


instance StartEndLoc (HsImportDeclI PNT)  where

     startEndLoc toks (HsImportDecl (SrcLoc _ row col) modName qual  as Nothing)
        = let (startPos,endPos)=((row,col),(row,col))
          in  if isJust as then let (startPos',endPos')=extendBackwards toks startPos endPos isConid
                                 in  extendBackwards toks startPos' endPos' isConid
                           else extendBackwards toks startPos endPos isConid

     
     startEndLoc toks (HsImportDecl (SrcLoc _ row col) modName qual as (Just (_, ents)))
         =let (startPos,endPos)=if ents==[] then (defaultPos,defaultPos)
                                            else startEndLoc toks $ head ents 
          in ((row,col),startPos)  --Attention: This function returns the startPos of the first element in the ents list


instance StartEndLoc  [HsExportSpecI PNT] where

   startEndLoc toks es 
     = if es == [] then (defaultPos, defaultPos)
                   else let (startLoc, _) = startEndLoc toks $ head es 
                            (_, endLoc)   = startEndLoc toks $ last es
                        in extendBothSides toks startLoc endLoc isOpenBracket isCloseBracket 


instance StartEndLoc (HsExportSpecI PNT) where

     startEndLoc toks (EntE ent) =startEndLoc toks ent
    
     startEndLoc toks (ModuleE moduleName) = (defaultPos, defaultPos)  --Problem: There is no location info in moduleName.
              
instance StartEndLoc(EntSpec PNT) where

      startEndLoc toks (Var i)=startEndLoc toks i   --- x (a variable identifier)
 
      startEndLoc toks (Abs i) =startEndLoc toks i   -- T, C

      startEndLoc toks (AllSubs i) =startEndLoc toks i -- T(..), C(..)

      startEndLoc toks (ListSubs i _)=startEndLoc toks i --T (C_1, ...,C_n, f1,...f_n)


instance StartEndLoc [EntSpec PNT] where
 
      startEndLoc toks ents 
        = if ents==[] then (defaultPos,defaultPos)
                      else let (startPos, _)=startEndLoc toks $ head ents
                               (_,  endPos) =startEndLoc toks $ last ents 
                           in (startPos,endPos)
                --         in extendBothSides toks startPos endPos isHiding isCloseBracket

instance StartEndLoc PNT where
    
     startEndLoc toks pnt =
        case pnt of 
          PNT pn  _ (N (Just (SrcLoc _ row col)))   ->((row,col),(row,col))
          _                                         ->(defaultPos,defaultPos)  --ANY PROBLEM HERE?



instance StartEndLoc HsDeclP where    
  
    startEndLoc toks t         --THIS FUNCTION SHOULD BE REVISED LATER.
           =let (startPos,endPos)=startEndLoc' t
                (toks1,toks2)=break (\t->tokenPos t==startPos) toks
                (toks21,toks22)=break (\t->tokenPos t>endPos) toks2 
                brackets1=filter isOpenBracket $
                         takeWhile (\t->isWhite t || isOpenBracket t)  (reverse toks1) 
                startPos'=if brackets1==[] then startPos else (fst.snd.(glast "startEndLoc:HsDeclP")) brackets1 
                brackets2=filter (not.isWhite)$
                             takeWhile (\t->tokenCol t>snd startPos && (not (isIn t)) && (not (isComma t))
                                           || isWhite t) toks22
                brackets2'=filter (not.isWhite) $ dropCloseSqBrackets (extraCloseSqBrackets (toks21++brackets2)) brackets2
                endPos'=if brackets2'==[] then endPos  else (fst.snd.(glast "startEndLoc:HsDeclP")) brackets2'
           in (startPos',endPos')
     where              
         startEndLoc' t=let locs=filter (\loc->loc>(0,0)) (allSortedLocs t) -- this should be changed later.(because the use of srcLoc0)
                        in (ghead "StartEndLoc:HsDeclP" locs, glast "StartEndLoc:HsDeclP" locs)
 
         extraCloseSqBrackets toks=length (filter isCloseSquareBracket toks)-length (filter isOpenSquareBracket toks)
         dropCloseSqBrackets 0 toks=toks
         dropCloseSqBrackets n toks=let (toks1,toks2)=break isCloseSquareBracket (reverse toks)
                                    in dropCloseSqBrackets (n-1) (reverse (tail toks2)) 
 

extendBothSides  toks startLoc endLoc  forwardCondFun backwardCondFun
       =let (toks1,toks2)=break (\t->tokenPos t==startLoc) toks
            toks21=dropWhile (\t->tokenPos t<=endLoc) toks2
            firstLoc=case (dropWhile (not.forwardCondFun) (reverse toks1)) of 
                             [] -> startLoc    -- is this a correct default?
                             l  -> (tokenPos.ghead "extendBothSides:lastTok") l  
            lastLoc =case (dropWhile (not.backwardCondFun) toks21) of
                            [] ->endLoc   --is this a correct default?
                            l -> (tokenPos.ghead "extendBothSides:lastTok") l                    
        in (firstLoc, lastLoc)   

extendForwards toks startLoc endLoc forwardCondFun
       =let toks1=takeWhile (\t->tokenPos t<startLoc) toks          
            firstLoc=case (dropWhile (not.forwardCondFun) (reverse toks1)) of
                       [] ->startLoc
                       l -> (tokenPos.ghead "extendForwards") l
        in (firstLoc, endLoc)

extendBackwards toks startLoc endLoc backwardCondFun
       = let toks1=dropWhile (\t->tokenPos t<=endLoc) toks
             lastLoc=case (dropWhile (not.backwardCondFun) toks1) of
                          [] ->endLoc
                          l ->(tokenPos. ghead "extendBackwards") l
         in (startLoc, lastLoc)     
                  

allSortedLocs t =(sort.nub.getSrcLocs) t \\ [defaultPos]

getSrcLocs=runIdentity.(applyTU (full_tdTU (constTU [] `adhocTU` pnt  
                                                       `adhocTU` literalInExp
                                                       `adhocTU` literalInPat)))
        where 
            pnt (PNT pname _ (N (Just (SrcLoc _ row col))))=return [(row,col)]                
            pnt _=return []
                     
            literalInExp ((Exp (HsLit (SrcLoc _ row col) _))::HsExpP) = return [(row,col)]
            literalInExp (Exp _) =return []

            literalInPat ((Pat (HsPLit (SrcLoc _ row col) _))::HsPatP) = return [(row,col)]
            literalInPat (Pat (HsPNeg (SrcLoc _ row col) _)) = return [(row,col)]
            literalInPat _ =return []  
                  



--get the start&end location of t in the token stream, then extend the end location to 
--include the following comments.
startEndLocIncFowComment toks t
       =let (startLoc,endLoc)=startEndLoc toks t
            toks1=dropWhile (\t->tokenPos t<=endLoc) toks
            (toks11,tok12)=break (\t->tokenRow t/=fst endLoc) toks1
        in  if toks11/=[] && all (\t->isWhite t || endsWithNewLn t) toks11 
              then (startLoc, tokenPos (glast "startEndLocIncFowComment" toks11))
              else (startLoc, endLoc) 

--get the start&end location of t in the token stream, including the following '\n' if there is one.
startEndLocIncFowNewLn toks t
  =let (startLoc,endLoc)=startEndLoc toks t
       toks1=dropWhile (\t->tokenPos t<=endLoc) toks
       nextTok=let t=dropWhile isWhiteSpace toks1
               in if t==[] then defaultToken else ghead "startEndLocIncFowNewLn" t
   in  if isNewLn nextTok  
          then (startLoc, tokenPos nextTok)
          else (startLoc, endLoc)
 
startEndLocIncComments toks t
  =let (startLoc,endLoc)=startEndLoc toks t    
       toks1=takeWhile (\t->tokenPos t<startLoc) toks
       --get all the white tokens before the start location
       groupedToks=groupBy (\x y->tokenRow x==tokenRow y) (reverse toks1)
       emptyLns=takeWhile (all (\t->isWhiteSpace t || isNewLn t )) groupedToks       
       lastComment=if length emptyLns <=1 
                     then takeWhile (all isWhite) $ takeWhile (any isComment) $ dropWhile 
                                                (all (\t->isWhiteSpace t || isNewLn t)) groupedToks    
                     else []
       toks1'=if lastComment /=[] then reverse $ concat (emptyLns ++ lastComment)
                                  else []
       startLoc'=if toks1'==[] then startLoc    --there is no comment before t
                               else tokenPos (ghead "startEndLocIncComments"  toks1') --there is a comment before t
       toks2=dropWhile (\t->tokenPos t<=endLoc) toks
       (toks21,tok22)=break (\t->tokenRow t/=fst endLoc) toks2
   in if toks21==[] 
        then (startLoc',endLoc)
        else if all (\t->isWhite t || endsWithNewLn t) toks21 --get the following white tokens in the same
                                                              --line of the last token of t
                 then (startLoc', tokenPos (last toks21))
                 else (startLoc', endLoc) 