

module RefacMvDefBtwMod(moveDefBtwMod) where

import PosSyntax
import Maybe
import TypedIds
import List 
import TiPNT 
import RefacUtils 
import UniqueNames 
import HsName
import PNT 
import RefacRmDef(removeInTopLevel)
import PFE0 (findFile)
import PrettyPrint

{-This refactoring moves a definition from current module to a user specified module.-}

moveDefBtwMod args
 = do let fileName     = ghead "filename" args        
          --origModName  = Module $ moduleName fileName 
          destModName  = Module (args!!1)
          row          = read (args!!2)::Int
          col          = read (args!!3)::Int
      origModName <- moduleName fileName
      r <-isAnExistingMod destModName
      unless r $  error "The specified module does not exist!"  
      unless (origModName /= destModName) $ error "The destination module name is the same as the current module name!"
      (origInscps, _, origMod, origToks) <- parseSourceFile fileName    
      -- Find the file name of the target module.
      destFile<-findFile destModName
      (destInscps, _, destMod, destToks) <-parseSourceFile destFile 
      let pn = locToPName fileName row col origMod                                        
      unless (pn /= defaultPN) $ error "Invalid cursor position!"
      unless (isTopLevelPN pn) $ error "The selected definition is not a top-level definition!"
      let decl = declDefinesPn pn origMod
          pns  = nub $ concatMap definedNames decl
      unless (not (isDefinedIn (map pNtoName pns)  destMod))
                (error $ "The name definition name is already defined in module "++ show destModName)    
      -- Get the client modules and the corresponding file names of the original module.
      clientsOfOrigMod <- clientModsAndFiles origModName
      -- Get the client modules and the corresponding file names of the original module.
      clientsOfDestMod <- clientModsAndFiles destModName 
      -- get the free variables in the definition to be moved.
      freePnts <- freePNTsIn pns decl                                      
      unless (varsAreInscope freePnts destInscps)
           $ error ("Not all the free variables in the definition to be moved are accessible by module " ++ show destModName)  
      -- Will the moving cause recursive modules?
      r <-causeRecursiveMods decl origModName destModName
      unless (not r) $ error ("Moving this definition to module " ++ show destModName++ " will cause mutually recursive modules!")
      -- Remove the definition from the original module.  
      (origMod', ((origToks',origM),_))<-rmCodeFromMod pns fileName (origMod, origToks) destModName destFile
      -- Parse all those client modules.
      let modsAndFiles=nub (clientsOfOrigMod++clientsOfDestMod) \\ [(origModName, fileName),(destModName, destFile)]
      parsedMods <-mapM parseSourceFile $ map snd modsAndFiles                 
      -- is the definition name used in modules other than the target module?
      let used = any (\(_,_,mod,_)->isUsedBy pns mod) parsedMods  || isUsedBy pns origMod
      -- Add the definition to the destination module.
      (destMod',((destToks',destM),_))<-addCodeToMod pns destFile (origMod, origToks) (destInscps,destMod, destToks) used
      --Do corrsponding modification in the client modules.
      refacClients <- mapM (refactorInClientMod origModName destModName pns) $ zip parsedMods $ map snd modsAndFiles 
      -- output the result.
      writeRefactoredFiles $ [((fileName,origM),(origToks',origMod')),((destFile,destM),(destToks',destMod'))]++ refacClients


-- Return True if any name in 'names' is defined in the scope of t.
isDefinedIn names t
  = fromMaybe False (do (_,rd)<-hsFreeAndDeclared t
                        Just (((map pNtoName rd) `intersect` names)/=[]))

--Test whether all the free variables in the definition to be moved are in scope in the destination module.
varsAreInscope pnts inScopeRel
   = not $ any (\pnt->isNothing (getImpQualifier pnt inScopeRel)) pnts
  
-- get those pnts which are free and different from pns in t.
freePNTsIn pns t 
 = do (pns', _) <- freeAndDeclaredVbls t
      let pnts = nub $ pNTsIn =<< rmLocInfo t 
      return $ filter (\t->elem  (pNTtoPN t) (pns' \\ pns)) pnts


--Return True if moving the definition from origMod to destMod will cause recursive modules.
causeRecursiveMods decl origModName destModName
 =do let pns  = nub $ concatMap definedNames decl
     clients<-clientModsAndFiles origModName
     let clientMods1 = map fst clients
         clientFiles =map snd clients
     if elem destModName clientMods1
      then do servers <-serverMods destModName
              let ms = origModName : (clientMods1 `intersect` servers)
              parsedMods<-mapM  parseSourceFile clientFiles   
              return $ any (\(_,_,mod,_)->isUsedBy pns mod) parsedMods                  
      else do servers <- serverModsAndFiles origModName                
              let serverMods = map fst servers
                  serverFiles =map snd servers
              if elem destModName serverMods
                then do clients <- clientMods destModName
                        let ms = origModName : (clients `intersect` serverMods)                            
                            r  = map pNtoModName $ nub (map pNTtoPN (pNTsIn decl)) \\ pns
                        return $ any (\m-> elem m r ) ms
                else return False  
 
--Return True if any pname in 'pns' is used by 'mod'
isUsedBy pns t
   = fromMaybe False (applyTU (once_tdTU (failTU `adhocTU` worker)) (declsIn t))
        where    
         worker (pnt::PNT)
           | elem (pNTtoPN pnt) pns && pNTinUseLoc pnt && not (findPntInLhs pnt t)
          = Just True
         worker _ = Nothing    


-- Remove the definition from the original module.
rmCodeFromMod pns fileName (mod, tokList) destModName destFileName
  = runStateT (do -- remove the definition.
                  decls'<-removeInTopLevel (ghead "rmCodeFromMod"  pns) (declsIn mod)  
                  -- remove the definition name from the expo
                  mod'<-rmItemFromExport (replaceDecls mod decls') pns 
                  -- is the definition used in the current module?
                  if isUsedBy pns mod' 
                    then -- yes. add the definition name to the import.
                         addImport destFileName destModName pns mod'  
                    else --Not used, do nothing.
                         return mod') 
             ((tokList,unmodified),(fileName,initValue))


-- Add the definition to the client module, and modify the import/export if necessary.
addCodeToMod  pns destFileName (origMod, origToks) (destInscps, destMod, destToks) usedByClients  
 = do let -- get the declaraion. False means not spliting the type signature.
          decl  = definingDecls pns (declsIn origMod) False
           --get the fun binding.
          funBinding = ghead "addCodeToMod" $ filter isFunOrPatBind decl  -- shoudn't be empty.
           --get the type signature if there is any.
          typeSig    = filter isTypeSig decl  
      if causeAbiguity pns funBinding destInscps destMod
       then error "Moving this definition will cause ambiguity problem in the target module! "
       else runStateT
           ( do let toksTobeAdded = if typeSig ==[] || definesSingleId (ghead "addCodeToMod" typeSig)
                                       then let (startPos, endPos) = startEndLocIncComments origToks decl
                                                (toks1,_) = break (\t->tokenPos t > endPos) origToks
                                                --get the token stream for the definition.
                                            in  dropWhile (\t-> tokenPos t <startPos || isNewLn t) toks1 
                                       else let typeSig'= ghead "addCodeToMod" $ getTypeSig pns (head typeSig)
                                                (startPos, endPos) = startEndLocIncComments origToks funBinding
                                                (toks1,_) = break (\t->tokenPos t > endPos) origToks
                                                --get the token stream for the function binfing.
                                                toks2= dropWhile (\t-> tokenPos t <startPos || isNewLn t) toks1
                                                t = mkToken Whitespace (0,0) $ concatMap (\s->s++"\n")((lines.render.ppi) typeSig')
                                            in (t:toks2) 
                    decl' = if typeSig==[] then [funBinding] else [ghead "addCodeToMod" (getTypeSig pns (head typeSig)),funBinding]
                -- Add the definition to the beginning of the destination module. 
             
                mod'<- addDefDecl (decl', toksTobeAdded) =<< rmQualifier pns destMod   --- Check this.
                mod'' <- rmItemFromImport mod' pns
                if usedByClients && not (wholeModIsExported destMod)  -- The definition name is used by other modules, but it is
                  then addVarItemInExport1 mod'' pns     -- not exported by default. so make it exported.
                  else return mod''      -- Do nothing.      
             ) ((destToks,unmodified), (destFileName, initValue))
   where 
        -- Get the type signatures defines the type of pns. 
       getTypeSig pns (Dec (HsTypeSig loc is c tp))
         =[(Dec (HsTypeSig loc (filter (\x->isJust (find (==pNTtoPN x) pns)) is) c tp))]
       getTypeSig pns _ = []


-- modify the import interface in the client modules.
refactorInClientMod origModName destModName pns ((inscps, exps,mod, ts), fileName)
  = do (mod', ((ts',m),_)) <- 
          runStateT ( do mod'<-rmItemFromImport mod pns
                         mod''<- if isUsedBy pns mod
                                  then do let qual=getQualifier destModName mod                                          
                                          addImport fileName destModName pns =<< replaceQualifier pns qual mod'
                                  else if itemIsImportedByDefault destModName mod && sameNameisUsed pns mod'
                                         then addHiding destModName mod' pns
                                         else return mod'
                         return mod'' --replaceQualifier pns destModName mod''
                    ) ((ts,unmodified),(fileName,initValue)) 
       return ((fileName,m),(ts',mod'))

getQualifier modName (HsModule  _ _  _ imp _)
    = let r=(nub.ghead "getQualifier") (applyTU (full_tdTU (constTU [] `adhocTU` inImport)) imp)
      in if r==[] then modName 
                  else head r     
    where 
    inImport ((HsImportDecl _ m _ as _)::HsImportDeclI PNT)
      | show modName == show m 
      = if isJust as then return [fromJust as]
                     else return [] 
    inImport _ = return []


--add a definition name to the import. If the module is not imported yet, then create a new import decl.
addImport destFileName destModName pns mod@(HsModule _ _  _ imp _)
  =if itemIsImportedByDefault destModName mod     -- Is the definition name explicitly imported?
    then return mod                               -- Yes. Do nothing.
    else if itemsAreExplicitlyImported destModName mod  --No. Is the module imported and some of its items are explicitly imported?
          then addVarItemInImport1 destModName pns mod -- Yes. Then add the definition name to the list.
          else addImportDecl mod (mkImportDecl destFileName destModName False (map pNtoName pns)) --Create a new import decl.

  where  
   {- Compose a import declaration which imports the module specified by 'modName',
      and only imports the definition spcified by 'e'.
    -}
   mkImportDecl fileName modName qual ids 
     = (HsImportDecl loc0 modName qual Nothing (Just (False, (map makeEnt ids))))
      where
      makeEnt id= Var $ PNT (PN (UnQual id) (G modName id (N (Just loc0)))) Value (N (Just loc0))
   
   itemsAreExplicitlyImported serverModName (HsModule _ _ _ imps _)
     = any (isExplicitlyImported serverModName) imps
    where
       isExplicitlyImported serverModName ((HsImportDecl _ modName _ _ h)::HsImportDeclI PNT)
         = serverModName == modName && isJust h && not (fst (fromJust h))


-- are items defined in the serverModName imported by the current module by default?
itemIsImportedByDefault serverModName (HsModule _ _ _ imps _) 
  = any (isImportedByDefault'  serverModName) imps
  where
    isImportedByDefault' serverModName ((HsImportDecl _ modName _ _ h)::HsImportDeclI PNT)
       = serverModName == modName && ( isNothing h  || (isJust h && fst(fromJust h)))


addVarItemInImport1 serverModName pns mod
   = applyTP ((once_tdTP (failTP `adhocTP` inImport))  `choiceTP` idTP) mod
  where 
    inImport (imp@(HsImportDecl loc modName qual  as (Just (b,ents))):: HsImportDeclP)
      | serverModName == modName && not b  
     =  do ((toks,_),others)<-get   
           let (_,endPos)=startEndLoc toks (glast "addVarItemInImport1" ents)
               (t, (_,s))=ghead "addVarItemInImport1" $ getToks (endPos,endPos) toks
               newToken=(t,(endPos, s++","++showPns pNtoName pns))
               toks'=replaceToks toks endPos endPos [newToken]
           put ((toks',modified),others) 
           pnts<-mapM (\pn->pNameToPNT pn Value) pns
           let ents'= map (\pnt->Var pnt) pnts          
           return (HsImportDecl loc modName qual as (Just (b, ents++ents')))
    inImport x = mzero

replaceQualifier pns qual  t
    =applyTP (full_tdTP (adhocTP idTP rename)) t
       where
         rename  pnt@(PNT pn ty (N (Just (SrcLoc fileName  row col))))
          |isQualifiedPN pn && elem pn pns 
          = do replaceToks1 (row,col) (row, col) [mkToken Varid  (row,col) (pNameToStr (replaceQual pn  qual))] 
               return (PNT (replaceQual pn qual) ty (N (Just (SrcLoc fileName  row col))))
            where 
              replaceQual (PN (Qual modName  s) l) qual = PN (Qual qual s) l             
              replaceQual pn _ = pn
         rename x=return x

rmQualifier pns t
    =applyTP (full_tdTP (adhocTP idTP rename)) t
       where
         rename  pnt@(PNT pn ty (N (Just (SrcLoc fileName  row col))))
          |isQualifiedPN pn && elem pn pns 
          = do replaceToks1 (row,col) (row, col) [mkToken Varid  (row,col) (pNameToStr (rmQual pn))] 
               return (PNT (rmQual pn) ty (N (Just (SrcLoc fileName  row col))))
            where 
              rmQual (PN (Qual modName  s) l)= PN (UnQual s) l             
              rmQual pn = pn
         rename x=return x

--same name but with different meaning is used.
sameNameisUsed pns mod
   = fromMaybe False (applyTU (once_tdTU (failTU `adhocTU` worker)) (declsIn mod))
        where    
         worker (pnt::PNT)
           | elem (pNTtoName pnt) (map pNtoName pns) && not (elem (pNTtoPN pnt) pns) 
             && pNTinUseLoc pnt && not (findPntInLhs pnt mod)
          = Just True
         worker _ = Nothing   

-- Will adding a new definition (decl), which defines pns, to mod cause any ambiguity problems?
causeAbiguity pns decl destInscps destMod
 = sameNameisUsed pns destMod || isRecursiveDef decl && sameUnqualNameIsScope (ghead "causeAbiquity" pns) destInscps
  


--ToDo: How to replace/remove the qualifiers in the original  and destination module?


