

module RefacNewDef(introNewDef,unfoldDef) where

import PrettyPrint
import PosSyntax
import Maybe
import TypedIds
--import ScopeModule
import UniqueNames hiding (srcLoc)
import PNT 
import TiPNT 

import List  
import RefacUtils
import PFE0 (findFile)

--Introdule a new defintion whose rhs is the user specified expression
introNewDef args
 = do let fileName   = args!!0
          --fileName'  = moduleName fileName
          --modName    = Module fileName'  
          newFunName = args!!1            
          beginRow   = read (args!!2)::Int
          beginCol   = read (args!!3)::Int
          endRow     = read (args!!4)::Int
          endCol     = read (args!!5)::Int
      modName <- moduleName fileName 
      if isVarId newFunName then
         do (inscps, exps, mod, tokList) <- parseSourceFile fileName
            (mod',((tokList',m),_))<-doIntroduce fileName newFunName (beginRow,beginCol)(endRow,endCol) mod tokList           
            writeRefactoredFiles [((fileName,m), (tokList',mod'))]
         else error "Invalid new function name!"
  
  where
   ---Introduce the new defintion, and put it at the end of next lower level declaration list
   doIntroduce fileName newFunName beginPos endPos mod tokList 
      =runStateT (applyTP (once_buTP (failTP `adhocTP` inMatch
                                             `adhocTP` inPat
                                             `adhocTP` inLetExp
                                             `adhocTP` inAlt
                                             `adhocTP` inStmt) `choiceTP` failure) mod) 
                         ((tokList,unmodified),(fileName,initValue))
                         
        where
           --The sub-expression is in the rhs of a match
           inMatch (match@(HsMatch loc1 name pats rhs ds)::HsMatchP)
              |inScope rhs && findExp tokList beginPos endPos rhs /=defaultExp
                     =doIntroduce' match  rhsIn  replaceRhs                 
           inMatch _ =mzero
           
           --The sub-expression is in the rhs of a pattern binding
           inPat (pat@(Dec (HsPatBind loc1 ps rhs ds))::HsDeclP)
              |inScope rhs && findExp tokList beginPos endPos rhs /=defaultExp
                     =doIntroduce' pat  rhsIn   replaceRhs   
           inPat _ =mzero
 
           -- The sub-expression is in the expression of a 'Let ... in'      
           inLetExp (exp@(Exp (HsLet ds e))::HsExpP)
               |inScope e && findExp tokList beginPos endPos e /=defaultExp
                    =doIntroduce'' exp expIn  replaceExpInLet 
           inLetExp _=mzero
                
           -- The sub-expression is in the rhs of a case alternative
           inAlt (alt@(HsAlt loc p rhs ds)::HsAltP)
                |inScope rhs && findExp tokList beginPos endPos rhs /=defaultExp
                    =doIntroduce' alt  rhsIn replaceRhs  
           inAlt _=mzero
      
           -- The sub-expression is in the stmts of a 'let'    
           inStmt (stmt@(HsLetStmt ds stmts):: HsStmtP)
                |inScope stmts && findExp tokList beginPos endPos stmts /=defaultExp
                     =doIntroduce' stmt  stmtsIn replaceStmtsInDo  
           inStmt  _=mzero

           failure=idTP `adhocTP` mod
                 where mod (m::HsModuleP)=error "The highlighted source does not contain a sub-expression!"
           
           inScope t=let (sLoc', eLoc')=startEndLoc tokList t
                         (sLoc,eLoc)=extendBothSides  tokList sLoc' eLoc'  isWhite isWhite
                     in beginPos>=sLoc && beginPos<=eLoc

           --ORDER MATTERS HERE
           doIntroduce' parent getRhs replaceFun  
               =do (f,d)<-freeAndDeclaredVbls parent
                   if elem newFunName (map pNtoName (f `union` d))
                      then 
                         error "The new function name causes name clash or capture,please select another name!"
                      else
                         do 
                            let exp=findExp tokList beginPos endPos (getRhs parent)
                            (freeVars,_)<-hsFreeAndDeclared exp
                            let 
                                params=(freeVars \\ (nub (f `union` d)))  --get parameters to the new function. 
                            (newFun,newFunPName)<-mkNewFun  newFunName params exp
                            addNewFunInTokStream parent newFun
                            --unified foldOneExp and foldAllExps to foldExps, 
                            --then switched back to 1, see comment for foldExps -- Claus
                            --1: only fold the highlighted expression
                            --rhs'<-foldExps False exp newFunPName params (fromJust(getRhs parent))
                            --2: fold all the courrences of the same expression
                            rhs'<-foldExps True exp newFunPName params (fromJust(getRhs parent))
                            return (replaceDecls (replaceFun parent rhs') (declsIn parent ++ [newFun])) 
                            --return (replaceDecls parent (declsIn parent ++ [newFun])) 
                            --3: without folding expression:
                            -- return (replaceDecls parent (newFun:(declsIn parent)))
           --ORDER MATTERS HERE.
           doIntroduce'' parent getRhs replaceFun  
               =do (f,d)<-freeAndDeclaredVbls parent
                   if elem newFunName (map pNtoName (f `union` d))
                      then 
                         error "The new function name causes name clash,please select another name!"
                      else
                         do let exp=findExp tokList beginPos endPos (getRhs parent)
                            (freeVars,_)<-hsFreeAndDeclared exp
                            let 
                                params=(freeVars \\ (nub (f `union` d)))  --get parameters to the new function. 
                            (newFun,newFunPName)<-mkNewFun newFunName params exp
                            --unified foldOneExp and foldAllExps to foldExps, 
                            --then switched back to 1, see comment for foldExps -- Claus
                            --1: only fold the highlighted expression
                            --rhs'<-foldExps False exp newFunPName params (fromJust(getRhs parent))
                            --2: fold all the courrences of the same expression
                            rhs'<-foldExps True exp newFunPName params (fromJust(getRhs parent))
                            addNewFunInTokStream parent newFun
                            return (replaceDecls (replaceFun parent rhs') (declsIn parent ++ [newFun])) 
                            --return (replaceDecls parent (declsIn parent ++ [newFun])) 
                            --3: without folding expression:
                            -- return (replaceDecls parent (newFun:(declsIn parent)))
         
     
           --The rhs of the new definition is pretty printed by the pretty printer.
           --comment this function later.
           addNewFunInTokStream parent newFun
              =do ((toks,_),others)<-get   
                  let localDecls=declsIn parent  --localDecls may be empty.
                      newFunSource=astToSource newFun  --pretty print the AST .
                      (startPos@(_,startCol),endPos'@(endRow',_))  --endPos' does not include the following newline or comment.
                               =if localDecls==[] then startEndLoc toks parent    --The 'where' clause is empty
                                                  else startEndLoc toks localDecls 
                      toks1=dropWhile (\t->tokenPos t<=endPos') toks 
                      ts1=takeWhile (\t->isWhite t && ((not.isMultiLineComment) t) && tokenRow t==endRow')  toks1    
                      --nextTokPos is only used to test whether there is a 'In' or a nested comment.                
                      nextTokPos= case (dropWhile (\t->isWhite t && ((not.isMultiLineComment) t) && tokenRow t==endRow') toks1) of
                                    [] -> defaultPos 
                                    l  -> (tokenPos.ghead "addNewFunInTokStream") l 
                      needNewLn=if nextTokPos==defaultPos  --used to decide whether add a new line character before a introduced fun.
                                  then if toks1==[] then True
                                                    else (not.endsWithNewLn) (last ts1) 
                                  else endRow'==fst nextTokPos 
                    {-  isBraceComment=if nextTokPos==defaultPos --check whether the fun is followed by nested comment.
                                         then False
                                         else endRow'/=fst nextTokPos && snd nextTokPos/=1 -}
                      endPos@(endRow,_)=if ts1==[] then endPos'
                                                   else tokenPos (last ts1)
                      t=if localDecls==[]
                          then (mkTokensWithIndent Special (endRow+1,startCol+2) ["where"] True) ++[newLnToken]++
                               mkTokensWithIndent Varid (endRow+2,startCol+6) newFunSource True -- ++[newLnToken]
                          else mkTokensWithIndent Varid  (endRow+1,startCol) newFunSource True
                      oldToks'=getToks (startPos,endPos) toks
                      newToks'=if needNewLn then oldToks' ++(newLnToken:t)  
                                            else oldToks'++ t++[newLnToken]    
                                          {-  if isBraceComment  then  oldToks'++(newLnToken:t)++[newLnToken]
                                                                    else oldToks'++ t++[newLnToken]  -}           
                      toks'=replaceToks toks startPos endPos newToks'
                      --SHOULD THE ROW NUMBERS IN TOKS2 BE ADJUSTED?
                  --error (show nextTokPos ++ "\n\n" ++ show endRow' ++ "\n\n" ++ show needNewLn)
                  --error (show (startPos,endPos') ++ "----" ++ show (parent)) 
                  put ((toks',modified),others)  
           
           mkNewFun funName params e
             =do (fileName, int)<-nextValue
                 let  loc=SrcLoc fileName int int 
                      funPName=PN (UnQual funName) (S loc)
                      funDecl=Dec (HsFunBind loc0 [HsMatch loc0 (PNT funPName Value (N (Just loc)))
                                   (map pNameToPat  params) (HsBody e) []])
                 --ATTENTION: if length[params]=0, SHOULD CREATE AN PATTERN BINDING.
                 --mkLocsUnique should apply here becaue the location of params need to change.
                 funDecl'<-mkLocsUnique funDecl [(pNameToLoc funPName)]
                 return (funDecl', funPName) 

    --Find the largest leftmost expression in the scope specified by beginPos and endPos.
   findExp toks beginPos endPos t
     =fromMaybe defaultExp (applyTU (once_tdTU (failTU `adhocTU` exp)) t)
       where exp (e::HsExpP)
               |inScope (startEndLoc toks e)=Just e
             exp _ =Nothing                     
             
             inScope (startLoc,endLoc)
                =startLoc>=beginPos &&  endLoc<=endPos                   
   

   -- won't work for all==True! for identifiers, == looks *only* at the location,
   -- so after setting all locations to some default with rmLocInfo,
   -- *all* identifiers look equal 
   foldExps all e funPName paramPNames
       =applyTP (traversal (failTP `adhocTP` exp))  --use once_tdTP if only fold one expression.                     
         where
            traversal   :: MonadPlus m => TP m -> TP m
            traversal    = if all then stop_tdTP else once_tdTP
            compare     :: (Eq t ,Term t) => t -> t -> Bool
            compare e1 e = if all 
                            then toRelativeLocs e1 == toRelativeLocs e
                            --then e1==e --does not handle the cases like [\x->x, \x->x].
                            else e1==e && getSrcLocs e1==getSrcLocs e --comparsion of location info should be enough.
            exp (e1 ::HsExpP)
               | compare e1 e =  do ((toks,_),others)<-get         
                                    let (startPos,endPos)=startEndLoc toks e1 
                                        newExpToken=mkToken Varid startPos  $ foldl (\x y->x++" "++y) 
                                                      (pNtoName funPName) (map pNtoName paramPNames)
                                        toks'=replaceToks toks startPos endPos [newExpToken]
                                    --error (show (startPos,endPos)++ show e1)
                                    put ((toks',modified),others)
                                    funExp<-pNameToExp funPName
                                    paramExps<-mapM pNameToExp paramPNames
                                    return (foldl addParamToExp funExp  paramExps)             
            exp _ =mzero
            --Add a parameter to an expression
            addParamToExp  e param=(Exp (HsApp e param)) 

   foldAllExps e funPName paramPNames
       =applyTP (stop_tdTP (failTP `adhocTP` exp))  --use once_tdTP if only fold one expression.                     
         where
            exp (e1 ::HsExpP)
               | rmLocInfo e1==e
              =do ((toks,_),others)<-get         
                  let (startPos,endPos)=startEndLoc toks e1 
                      newExpToken=mkToken Varid startPos  $ foldl (\x y->x++" "++y) 
                                         (pNtoName funPName) (map pNtoName paramPNames)
                      toks'=replaceToks toks startPos endPos [newExpToken] 
                  put ((toks',modified),others)
                  funExp<-pNameToExp funPName
                  paramExps<-mapM pNameToExp paramPNames
                  return (foldl addParamToExp funExp  paramExps)             
            exp _ =mzero
            --Add a parameter to an expression
            addParamToExp  e param=(Exp (HsApp e param)) 



mkLocsUnique t exceptLocs=do applyTP (full_td (idTP `adhocTP` inLoc)) t
     where 
          inLoc loc@(SrcLoc fileName row col)    --SHOULD ALSO COMPARE FILENAME TO AVOID COINCIDENCE.
             = if isJust (find (==(row, col)) allSourceLocs) --'returning True' means that the id is a declared locally
                 then return (SrcLoc fileName (row+rowOffset) (col+colOffset))
                 else return loc
           
          allSourceLocs=allSortedLocs t \\ ((0,0):exceptLocs)


{-ufolding a definition replaces a function call by its definition, the unfolded definition should be 
 a function/simple pattern binding. -}

unfoldDef args
 =do let fileName = ghead "fileName'" args 
         --fileName'= moduleName fileName
         --modName  = Module fileName'  
         row      = read (args!!1)::Int
         col      = read (args!!2)::Int 
     modName <- moduleName fileName
     (inscps, exps, mod, tokList)<-parseSourceFile fileName
     let pnt = locToPNT fileName row col mod
         pn  = pNTtoPN pnt                                
     if pn /=defaultPN  && (pNTinUseLoc pnt)
       then 
         do decl<-findDefiningDecl pn modName mod
            if decl/=[] 
              then do unInscopeIds <- freeVarsNotInScope decl inscps pnt mod 
                      if unInscopeIds == [] 
                        then do (mod',((tokList',m),_))<-doUnfolding pnt (ghead "unfoldDef" decl) mod 
                                                          ((tokList,unmodified),(fileName, initValue))
                                writeRefactoredFiles [((fileName,m), (tokList',mod'))]
                        else showNotInScopeError unInscopeIds  
              else error "Selected identifer is not a function/pattern binding name!"  
       else error "\nInvalid cursor position!\n"
 
   where 
         
    findDefiningDecl  pn currentModName currentMod
      = let defineModName=pNtoModName pn
        in if isLocalPN pn || defineModName == currentModName
             then return $ declDefinesPn pn currentMod
             else do (inscps, exps, mod, tokList)<-parseSourceFile =<<findFile defineModName
                     return $ declDefinesPn pn mod

    freeVarsNotInScope unfoldedDecl inscps pnt mod
       = do (f,_)<-hsFreeAndDeclared unfoldedDecl 
            dl<-hsVisibleVbls pnt mod         
            let freePnts=filter (\x->elem (pNTtoPN x) f) $ pNTsIn unfoldedDecl
                freePnts'=filter (\x-> not (elem (pNTtoPN x) dl)) freePnts
            return $ nub.(map pNTtoPN) $ filter (\x -> isNothing (getImpQualifier x inscps)) freePnts'


    showNotInScopeError unInscopeIds 
       = if length unInscopeIds>1        
            then error ("The free identifiers:" ++ showPns pNtoQuotedName unInscopeIds ++
                        ", used in the definition to be unfolded, are not in scope in the current module!")
            else error ("The free identifier:" ++ showPns pNtoQuotedName unInscopeIds ++
                        ", used in the definition to be unfolded, is not in scope in the current module!")
        where
          pNtoQuotedName pn =" '"++pNtoName pn++"'" 

                           
    doUnfolding pnt unfoldedDecl mod tokListAndLoc
      =runStateT (do simplifiedDecl'<-simplifyDecl unfoldedDecl
                     simplifiedDecl <-mkLocsUnique simplifiedDecl' []
                     doUnfolding' pnt simplifiedDecl mod ) tokListAndLoc
      where 

        doUnfolding' pnt simplifiedDecl mod
           =applyTP (once_tdTP (failTP `adhocTP` inMatch
                                       `adhocTP` inPat) `choiceTP` failure) mod 
          where
            inMatch (match@(HsMatch loc1 name pats rhs ds)::HsMatchP)
              |findEntity pnt rhs 
                = do match'<-renameInAppSite pnt simplifiedDecl match
                     doUnfolding' pnt  simplifiedDecl  match'
            inMatch _ =mzero
     
            inPat (pat@(Dec (HsPatBind loc p rhs ds))::HsDeclP)
              |findEntity pnt rhs 
                =do pat'<-renameInAppSite pnt simplifiedDecl pat
                    doUnfolding' pnt simplifiedDecl pat'
            inPat _=mzero
          
            failure=idTP `adhocTP` mod
                where mod (m::HsModuleP)=error "Refactoring failed!"  
          
            renameInAppSite pnt unfoldedDecl t
             =do (f,_)<-hsFreeAndDeclared unfoldedDecl 
                 dl<-hsVisibleVbls pnt t
                 --'f\\dl' is used to handle in case that the unfoldedDecl is a local decl of the parent.
                 --see 'ComplexPatIn1.hs' for an example.
                 let clashedNames=filter (\x-> elem (pNtoName x) (map pNtoName (f\\dl))) (nub dl) 
                 t'<-foldM (flip (autoRenameLocalName True)) t clashedNames  
                 return t'
               
            renameInUnfoldedDecl unfoldedDecl subst@(pn, exp) --pn is the formal parameter, exp is the actual parameter.
              =do  ds<-hsVisibleVbls pn unfoldedDecl
                   (f,_)<-hsFreeAndDeclared exp
                   let clashedNames=filter (\x->elem (pNtoName x) (map pNtoName f)) (nub ds)
                   foldM (flip (autoRenameLocalName False)) unfoldedDecl clashedNames

            doUnfolding' pnt unfoldedDecl 
              =applyTP (once_tdTP (failTP `adhocTP` exp)) 
                where
                  exp (e::HsExpP) --Because of the once_td traversal, the 'e' here must be the whole rhs or guard expression.
                    |findEntity pnt e
                    =if isComplexPatBind unfoldedDecl
                        then do replacePatCallByExp pnt unfoldedDecl e
                        else do let formalParams=paramsInDecl unfoldedDecl --get formalParams in unfolded declaration.
                                    actualParams=allParams pnt e     --get actual parameters.
                                    subst=zipWith (,) (map patToPName  formalParams) actualParams                         
                                unfoldedDecl'<-foldM  renameInUnfoldedDecl unfoldedDecl subst
                                --some formal parameters may have been renamed.
                                let formalParams'=paramsInDecl unfoldedDecl' 
                                    subst'=zipWith (,) (map patToPName formalParams') actualParams
                                    extraFormalParams=drop (length actualParams) formalParams'
                                decl'<-foldM replaceExp unfoldedDecl' subst'
                                let rhsExp=if extraFormalParams/=[]  --In case of partial application, convert extra
                                                                   --formal parameters into lambda exprssion.
                                        then (Exp (HsLambda extraFormalParams (fromJust (rhsExpInDecl decl'))))
                                        else fromJust (rhsExpInDecl decl')
                                    localDecls=declsIn decl'
                                replaceFunCallByExp pnt rhsExp localDecls =<<(rmParams pnt e (length subst')) --do unfolding      
                  exp _=mzero
          
       
     --a pattern binding is a complex if the pattern is not a single variable.
    isComplexPatBind decl=case decl of
                   Dec (HsPatBind _ p _ _)->patToPName p ==defaultPN
                   _ ->False

    replacePatCallByExp pnt decl
     =applyTP (full_buTP (idTP `adhocTP` worker))
        where worker (e::HsExpP)
                | samePNT (expToPNT e) pnt   --in 'samePNT', use location matters.
                     =do ((toks,_),others)<-get       --THIS SHOULD BE REFACTORED OUT LATER.
                         let newExp=Exp (HsParen (Exp (HsLet [decl] e)))
                             (startPos,endPos)=startEndLoc toks e 
                             newExpToks=mkTokensNoFstLnIndent Varid startPos ((lines.render.ppi) newExp)
                             toks'=replaceToks toks startPos endPos newExpToks
                         put ((toks',modified),others)  
                         return newExp 
              worker x =return x

 
       
    ---replace the function name by it's defintion
    replaceFunCallByExp pnt exp localDecls
      = applyTP (full_buTP (idTP `adhocTP` worker)) 
         where worker (e::HsExpP)
                 |samePNT (expToPNT e) pnt
                     =do ((toks,_),others)<-get   
                         let newExp=if localDecls/=[]
                                     then Exp (HsParen (Exp (HsLet localDecls exp)))
                                     else Exp (HsParen exp)  --Should check if exp is ready a Hsparen expression!
                             (startPos,endPos)=startEndLoc toks e 
                             newExpToks=mkTokensNoFstLnIndent Varid startPos ((lines.render.ppi) newExp)
                             toks'=replaceToks toks startPos endPos newExpToks
                         put ((toks',modified),others)
                         return newExp  
                         
               worker x=return x

    --get the parameters in the declaration (there should be only once match).
    paramsInDecl decl = case decl of
                     (Dec (HsFunBind _ ((HsMatch _  _ pats _ _):_))) ->pats
                     _ ->[]
 
    --get the right hand side repression in a declaration
    rhsExpInDecl (Dec (HsFunBind _ ((HsMatch _ _  _ (HsBody e) _):_)))=Just e 
    rhsExpInDecl (Dec (HsPatBind _ _ (HsBody e) _))=Just e

    --get all the parameters to pnt (which is a function name) in expression 'exp'
    allParams pnt exp=allParams' pnt exp []
        where allParams' pnt exp initial -- pnt:function/pattern name.
                 =let p=getOneParam pnt exp
                  in if p/=defaultExp  then allParams' pnt (rmOneParam pnt exp) (initial++[p])         
                                       else initial

    --get the first parameter of pnt
    getOneParam pnt t
       =fromMaybe defaultExp (applyTU (once_tdTU (failTU `adhocTU` worker)) t)
         where
           worker (Exp (HsApp e1 e2))
              |samePNT (expToPNT e1) pnt=Just e2
           worker _ =Nothing

    --remove the first parameter of pnt
    rmOneParam  pnt t
        =fromMaybe defaultExp (applyTP (stop_tdTP (failTP `adhocTP` worker)) t)
           where
             worker (Exp (HsApp e1 e2 ))
               |samePNT (expToPNT e1) pnt =Just e1
             worker _=Nothing

    rmOneParamWithUpdToks pnt 
       =applyTP (stop_tdTP (failTP `adhocTP` inExp)) 
           where

             inExp (exp@(Exp (HsParen (Exp (HsApp e1 e2))))::HsExpP)
                |samePNT (expToPNT e1) pnt 
                 =do ((toks,_),others)<-get     --handle the case like '(fun x) => fun "
                     let (startPos,endPos)=startEndLoc toks exp
                         toks'=replaceToks toks startPos endPos $ getToks (startEndLoc toks e1) toks
                     put ((toks',modified),others)
                     return e1     
             inExp exp@(Exp (HsApp e1 e2))
              | samePNT (expToPNT e1) pnt
                 =do ((toks,_),others)<-get   
                     let (startPos,endPos)=startEndLoc toks exp
                         toks'=replaceToks toks  startPos endPos $ getToks (startEndLoc toks e1) toks
                     put ((toks',modified),others)
                     return e1
             inExp  _=mzero
    --remove the first n parameters of pnt in expression 'exp' 
    rmParams pnt exp n=if n==0 then return exp
                               else do exp'<-(rmOneParamWithUpdToks pnt exp)
                                       rmParams pnt exp' (n-1)


{-A simple function binding satisfies : 1. all parameters are variables
                                        2. only one match(equation)
                                        3. The rhs of the match is not in guard style.
  that is:
   HsFunBind SrcLoc ((HsMatch SrcLoc i [var] (HsBody e) ds):[]) ds

  If a function binding is not a simple function binding, then convert it into a simple binding 
  using Case or IfThenElse expressions like this:
    case1: there are multi matches => case expression
    case2: there is only one match, however the parameters are not simple variables =>case expression.
    case3: there is only one match and the parameters are all simple variables, however there is a guard
           in Rhs => If then else

  In case of pattern binding: if there is guard in its Rhs, then convert it into IfThenElse style.
                              
-}
       
simplifyDecl decl
      |isFunBind decl =if (multiMatches decl) || (singleMatchWithComplexParams decl)
                           then matchesToCase decl
                           else return (guardToIfThenElse decl)
      |isPatBind decl=return (guardToIfThenElse  decl)
   where

      multiMatches decl=length (matchesInDecls decl)>1

      singleMatchWithComplexParams (Dec (HsFunBind loc matches@((HsMatch loc1 pnt pats rhs ds):ms)))
            =length matches==1 && any (==defaultPN) (map patToPName pats)
      singleMatchWithComplexParams _ =False

      --convert a multi-match function declaration into a single-match declration using case expression.
      matchesToCase decl@(Dec (HsFunBind loc matches@((HsMatch loc1 pnt pats rhs ds):ms)))
        =do params<-mkNewParamPNames (length pats)
            exp<-pNamesToExp params
            return (Dec (HsFunBind loc [(HsMatch loc1 pnt (map pNameToPat params)
                      (HsBody (Exp (HsCase exp (map matchToAlt matches)))) [])]))

      --make n new parameters like [x_0,x_1, ...,x_n]
      mkNewParamPNames n=mkNewParamPNames' n []
              where mkNewParamPNames' n pNames
                         =if n==0 then return pNames
                                     else do pn'<-idToLocalPName ("x_"++show (n-1))
                                             mkNewParamPNames' (n-1) (pn':pNames)
                                        
      matchToAlt ((HsMatch loc1 pnt pats rhs ds)::HsMatchP)=HsAlt loc0 (listToTuple pats) rhs ds
        where
         listToTuple pats=if (length pats)==1 then head pats   --no problem with head
                                              else (Pat (HsPTuple pats))

      pNamesToExp pns@(p:ps)=if ps==[] then pNameToExp p
                                        else  do exp'<-mapM pNameToExp pns
                                                 return (Exp (HsTuple exp')) 
     
      guardToIfThenElse decl= case decl of
          (Dec (HsPatBind loc p g@(HsGuard gs) ds))->(Dec (HsPatBind loc p (rmGuard g) ds))
          (Dec (HsFunBind loc ((HsMatch loc1 pnt pats  g@(HsGuard gs) ds):[]))) ->
                            (Dec (HsFunBind loc ((HsMatch loc1 pnt pats (rmGuard g) ds):[])))
          _ ->decl
           
          where
           rmGuard ((HsGuard gs)::RhsP)
             = let (_,e1,e2)=glast "guardToIfThenElse" gs
               in  if ((pNtoName.expToPName) e1)=="otherwise" 
                   then  HsBody (foldl mkIfThenElse e2 (tail(reverse gs)))
                   else  HsBody (foldl mkIfThenElse defaultElse (reverse gs)) 
           
           mkIfThenElse e (_,e1, e2)=(Exp (HsIf e1 e2 e)) 

           defaultElse=(Exp (HsApp (Exp (HsId (HsVar (PNT (PN (UnQual "error") (G (Module "Prelude") "error" 
                        (N (Just loc0)))) Value (N (Just loc0)))))) (Exp (HsLit loc0 (HsString "UnMatched Pattern")))))
            



