
module RefacUtils(module DriftStructUtils, module StrategyLib, module RefacTypeSyn,module RefacLocUtils,module RefacUtils, writeHaskellFile)
    where

import Prelude hiding (putStr,putStrLn,writeFile,readFile)
import PFE0 (getEditorCmds,,getCurrentModuleGraph,parserFlags,addToHistory,allModules)
import PFE3(parseModule,parseModule', parseSourceFile')
import PosSyntax
import TypedIds
import ScopeModule
import UniqueNames hiding (srcLoc)
import HsName
import HsLexerPass1
import AbstractIO 
import PNT 
import TiPNT
import Maybe
import List  
import Char
import RefacTypeSyn
import RefacLocUtils
import Graph(reverseGraph,reachable)
import NamesEntities 
import Ents
import Relations
import PosName
import QualNames
import TypedIds
import WorkModule
import SourceNames
import PrettyPrint
import Unlit(readHaskellFile,writeHaskellFile)
import ParserOptions
import MUtils 

import EditorCommands(sendEditorModified)
import qualified MT(lift)
-------------------------
import DriftStructUtils  
import StrategyLib hiding (findFile)
------------------------

{-
moduleName fileName = 
  case span (/='.') fileName of
    (name,".lhs") -> name
    (name,".hs")  -> name
    _ -> error $ "moduleName: can't drop .hs/.lhs from "++fileName -}

-- assume file name is in the current dir.
moduleName fileName = 
  do gf <- getCurrentModuleGraph
     let fileAndMods = [(m,f)|(f,(m,ms))<-gf]
         f = filter (\(m,f) -> f==fileName) fileAndMods
     if f ==[] then error $ "Can't find module name"
                    else return $ (fst.head) f
 

--Some Miscellaneous functions involving module graph.

data NameSpace = ValueName | ClassName | TypeCon | DataCon | Other  deriving (Eq, Show)

{-This function returns all the modules (as well as the file names) that imports module m either directly 
  or indirectly. 
-}

--REFACTORING: COMBINE THESE FOLLOWING FOUR FUNCTIONS.
clientModsAndFiles m =
  do gf <- getCurrentModuleGraph
     let fileAndMods = [(m,f)|(f,(m,ms))<-gf]
         g           = (reverseGraph.(map snd)) gf     
         clientMods  = reachable g [m] \\ [m]
         clients     = concatMap (\m'->[(m,f)|(m,f)<-fileAndMods, m==m']) clientMods
     return clients

clientMods m = 
   do gf <- getCurrentModuleGraph
      let fileAndMods = [(m,f)|(f,(m,ms))<-gf]
          g           = (reverseGraph.(map snd)) gf     
      return (reachable g [m] \\ [m])
  

serverModsAndFiles m =
   do gf <- getCurrentModuleGraph
      let fileAndMods = [(m,f)|(f,(m,ms))<-gf]
          g           = (map snd) gf  
          serverMods  = reachable g [m] \\ [m]
          servers     = concatMap (\m'->[(m,f)|(m,f)<-fileAndMods, m==m']) serverMods
      return servers

serverMods m = 
  do gf <- getCurrentModuleGraph
     let fileAndMods = [(m,f)|(f,(m,ms))<-gf]
         g           = (map snd) gf  
     return (reachable g [m] \\ [m])
 
isAnExistingMod m =  do ms<-allModules
                        return (elem m ms)


--whenever an identifier is imported, the qualified name is imported, the unqualified name
--may or maynot be imported.

getImpQualifier pnt@(PNT pname _ _ ) inScopeRel
  = getQualifyModName.map simplifyEnt. filter (sameEnt pnt) $ relToList inScopeRel
        
   where  
       sameEnt (PNT pname t1 _) (_,(Ent modName1 ident t2))
            =let modName2=hasModName pname
             in isJust modName2 && modName2 == Just modName1 && 
                pNtoName pname==identToName ident && hasSameNamespace t1 t2
 
       simplifyEnt (modQualifier, ent@(Ent (Module modName) ident ty ))
          =(identToName ident, getQualifier modQualifier, modName, hasNameSpace ty)

       getQualifyModName simpleEnts 
           =let r=filter isJust $ map (\(_, x, _,_) -> x) simpleEnts
            in if r==[] then Nothing   --means this pname is not imported.
                        else (head r) 
           
isInScopeAndUnqualified id inScopeRel   = isJust $ find (sameId id) $ relToList inScopeRel
     where 
        sameId id (qual, (Ent modName ident t2))
          = isUnqual qual &&  id==identToName ident


--same name but with different meaning is in scope.
sameUnqualNameIsScope pn inScopeRel = isJust $ find (sameName pn) $ relToList inScopeRel
     where 
        sameName pn (qual, (Ent modName ident t2))
          = isUnqual qual &&  pNtoName pn ==identToName ident && pNtoModName pn /=modName

--Id is defined in two modules: HsNames.hs (type Id = String) and PosName.hs (type Id = SN HsNames.Id)
identToName (HsVar (SN i _)) = i
identToName (HsCon (SN i _)) = i

inScopeNames inScopeRels = map getEntName $ relToList $ unionRels inScopeRels
   where 
     getEntName (qual, (Ent _ ident _))
       =identToName ident
                      
causeAmbiguityInExports pn  newName inscps mod 
  = isInScopeAndUnqualified (pNtoName pn) inscps &&
     findUnqualEntInExport newName (hsModExports mod)
    
  
findUnqualEntInExport name exp
   =(fromMaybe False) $ applyTU (once_tdTU (failTU `adhocTU` worker)) exp
      where
         worker (pnt::PNT)
           | not (isQualifiedPN (pNTtoPN pnt)) &&  pNTtoName pnt==name
          =Just True
         worker _ =Nothing
  
         
isExported pnt@(PNT pn t1 _) exps
   = if isTopLevelPNT pnt 
       then case hasModName pn of 
               Just modName  ->case find (\x-> show (fst x) ==show (pNtoName pn)) exps of
                                    Just (_, (Ent modName1 _ t2)) -> modName==modName1 && hasSameNamespace t1 t2 
                                    Nothing                       -> False  
               Nothing       -> False
       else False 

isExplictlyExported pn mod
   =findPName pn $ hsModExports mod

--Return True if the whole module is imported, otherwise False.
wholeModIsExported mod
   = let exps    = hsModExports mod
         modName = hsModName mod
     in if isNothing exps
           then True
           else isJust $ find (==(ModuleE modName)) (fromJust exps)

--Check this function: why add the new name to the export list?
 --test whether adding the new name to the export list will cause name clash in the exports--
--ExpList :: [(n,Ent n)]-- eg: [("f",Ent "A" (HsVar "f") Value)]

causeNameClashInClientExports  newName mod exps
  = let modNames=nub (concatMap (\(x,Ent modName _ _)->if show x==show newName 
                                                        then [modName]
                                                        else []) exps)
    in any (modIsUnQualifedImported mod) modNames                

      
modIsUnQualifedImported mod modName
 =let imps =hsModImports mod
  in isJust $ find (\(HsImportDecl _ modName1 qualify  _ h)->modName==modName1 && (not qualify)) imps


hasSameNamespace ty1 ty2=hasNameSpace ty1 == hasNameSpace ty2         

hasNameSpace Value           = ValueName
hasNameSpace (ConstrOf _ _ ) = DataCon
hasNameSpace (FieldOf _ _ )  = ValueName
hasNameSpace (MethodOf _ _)  = ValueName
hasNameSpace (Class _)       = ClassName
hasNameSpace (Type _)        = TypeCon     -- It is also possible that it is a type variable.
hasNameSpace  _              = Other       -- We don't care about Assertion & Property so far.


addImportDecl (HsModule loc modName exp imp decls) impDecl
 = do ((toks,_),others)<-get   
      let loc1 = if imp /=[]
                 then let (_, (endRow,_)) = startEndLocIncComments toks (last imp)  
                      in (endRow+1, 0)   
                 else if decls /=[] 
                       then let ((startRow,_),_) = startEndLocIncComments toks (head decls)
                            in (startRow -1, 0)
                       else defaultPos  --This should not happen. No point to add imports to a module without definitions.
          (toks1, toks2) = break (\t->tokenPos t>=loc1) toks
          newTok = mkToken Varid  loc1 $ (render.ppi) impDecl ++"\n"
          toks'  = toks1 ++ [newTok] ++ toks2
      put ((toks',modified),others)
      return (HsModule loc modName exp (imp++[impDecl]) decls)

-- add a definition to at the beginning of the definition declaration list.
addDefDecl (defDecl, defDeclToks) (HsModule loc modName exp imp decls)
  = do ((toks,_),others)<-get   
       let loc1 = if decls/=[] 
            then let ((startRow,_),_) = startEndLocIncComments toks (head decls)
                 in (startRow -1, 0)
            else defaultPos
           (toks1, toks2) = if loc1==defaultPos
                             then (toks, [])
                             else break (\t->tokenPos t>=loc1) toks
           toks' = toks1 ++ [newLnToken] ++ defDeclToks ++ toks2
       put ((toks',modified),others)                                 
       return (HsModule loc modName exp imp (defDecl++decls))


--add an item in the export list 
addVarItemInExport mod@(HsModule loc modName exps imps ds) pn funName
    = case exps  of
       Just ents ->let (e1,e2) = break (findPName pn) ents
                   in if e2 /=[]  
                        then do ((toks,_),others)<-get   
                                let e = (ghead "addVarItemInExport" e2)
                                e'<-renamePN pn (pNtoName funName) (pNameToOrig funName) e
                                let (_,endPos) = startEndLoc toks e
                                    (t, (_,s)) = ghead "addVarItemInExport" $ getToks (endPos,endPos) toks         
                                    newToken   = (t,(endPos, s++","++ (render.ppi) e'))
                                    toks' = replaceToks toks endPos endPos [newToken]
                                put ((toks',modified),others)                                   
                                return (HsModule loc modName (Just (e1++[e,e']++tail e2)) imps ds)
                        else return mod 
                     
       Nothing   -> return mod  
  where
    renamePN oldPN newName newLoc
      =applyTP(full_tdTP(adhocTP idTP rename))  
        where
           rename pn
              | pn==oldPN && pNameToOrig oldPN == pNameToOrig pn 
                    = return (replaceNameAndLocInPN pn newName newLoc)
           rename x = return x    


--add an item in the export list 
addVarItemInExport1 mod@(HsModule loc modName exps imps ds) pns 
    = case exps  of
       Just ents ->let (e1,e2) = break (\e->case e of 
                                            EntE (Var _) -> True 
                                            _  -> False) ents  
                   in if e2 /=[]  
                        then do ((toks,_),others)<-get   
                                pnts <- mapM (\pn->pNameToPNT pn Value) pns
                                let e = (ghead "addVarItemInExport" e2)                                   
                                    e'= map (\pnt->EntE (Var pnt)) pnts 
                                let (_,endPos) = startEndLoc toks e
                                    (t, (_,s)) = ghead "addVarItemInExport" $ getToks (endPos,endPos) toks         
                                    newToken   = (t,(endPos, s++","++ showPns pNtoName pns))
                                    toks' = replaceToks toks endPos endPos [newToken]
                                put ((toks',modified),others)                                   
                                return (HsModule loc modName (Just (e1++e'++ e2)) imps ds)
                        else return mod 
                     
       Nothing   -> return mod  
  
  
{- add items in an import declaration:
 if the generalised function is hided(or explicitly impported) in the import declaration, 
 the auxiliary function will be hided(or explicitly imported) too. -}

addVarItemInImport serverModName mod pn funPName
   = applyTP (full_buTP (idTP `adhocTP` inImport)) mod
  where 
    inImport (imp@(HsImportDecl loc modName qual  as h)::HsImportDeclI PNT)
      | serverModName == modName  &&  findPName pn imp
       = case h of
           Nothing        -> return imp                     
           Just (b, ents) -> do ((toks,_),others)<-get  
                                let (_,endPos)=startEndLoc toks (glast "addVarItemInImport" ents)
                                    (t, (_,s))=ghead "addVarItemInImport" $ getToks (endPos,endPos) toks
                                    newToken=(t,(endPos, s++","++pNtoName funPName))
                                    toks'=replaceToks toks endPos endPos [newToken]
                                put ((toks',modified),others) 
                                pnt<-pNameToPNT funPName Value
                                return (HsImportDecl loc modName qual as (Just (b, ents++[(Var pnt)])))
    inImport imp = return imp
  

--add items in the hiding part of an import declaration which imports the server module..
addHiding serverModName mod pns
   = applyTP (full_tdTP (idTP `adhocTP` inImport)) mod
  where 
    inImport (imp@(HsImportDecl loc modName qual  as h)::HsImportDeclI PNT)
      | serverModName == modName  && not qual
       = case h of
           Nothing ->do ((toks,_),others)<-get  
                        let (_,endPos)=startEndLoc toks imp
                            (t,(_,s)) = head $ getToks (endPos,endPos) toks 
                            newToken=(t, (endPos, s++" hiding ("++showPns pNtoName pns++")"))
                            toks'=replaceToks toks endPos endPos [newToken]
                        put ((toks',modified),others)    
                        return (replaceHiding imp (Just (True, map (mkNewEnt serverModName) pns )))
           Just (True, ents) ->do ((toks,_),others)<-get  
                                  let (_,endPos)=startEndLoc toks imp
                                      (t, (_,s))=head $ getToks (endPos,endPos) toks
                                      newToken=(t,(endPos, showPns pNtoName pns ++","++s))
                                      toks'=replaceToks toks endPos endPos [newToken]
                                  put ((toks',modified),others) 
                                  return (replaceHiding imp  (Just (True, (map (mkNewEnt serverModName) pns)++ents))) 
           Just (False, ent)  -> return imp
  
    inImport x = return x

    mkNewEnt modName pn = (Var (PNT pn Value (N (Just loc0))))

    replaceHiding (HsImportDecl loc modName qual as h) h1 = HsImportDecl loc modName qual as h1    


--remove the idenitifers specified by 'pns' from the hiding clause in client modules.
rmItemFromImport mod pns
  = applyTP (full_buTP (idTP `adhocTP` inImport)) mod
   where 
     inImport (imp@(HsImportDecl loc modName qual  as h)::HsImportDeclP)
      | any (flip findPName imp) pns
       = case h of
           Just (b, ents) ->
             do let matchedEnts=findEnts pns ents
                if  matchedEnts==[] 
                  then return imp
                  else if length matchedEnts == length ents   -- Problem: Why not remove the whole import decl when it is not hiding?
                         then do ((toks,_),others)<-get  
                                 let (startPos,endPos)=startEndLoc toks ents
                                     toks'=deleteToks toks startPos endPos
                                 put ((toks',modified),others)
                                 return (HsImportDecl loc modName qual as Nothing)  -- This is wrong?
                         else do ((toks,_),others)<-get  
                                 let remainedEnts=concatMap (\pn->filter (not.isVarAndSame pn) ents) pns 
                                     toks'=foldl deleteEnt toks $ map (startEndLoc toks) matchedEnts
                                 put ((toks',modified),others)
                                 return (HsImportDecl loc modName qual as (Just (b, remainedEnts)))
           _ ->return imp 
     inImport x = return x
   
     findEnts pns ents =concatMap (\pn->filter (isVarAndSame pn) ents) pns


isVarAndSame pn (Var i)  = pNTtoPN i == pn
isVarAndSame _       _  = False

--delete an ent from the hiding list, the following comma should also be removed if there is any.
deleteEnt toks (startPos, endPos)
  = let (toks1,toks2)=break (\t->tokenPos t>=startPos) toks
        previousTok=ghead "deleteEnt" $ dropWhile isWhiteSpace $ reverse toks1
        startPos'=if isComma previousTok then tokenPos previousTok else startPos
        toks' = dropWhile isWhiteSpace $ dropWhile (\t->tokenPos t<=endPos) toks2
        nextTok = ghead "deleteEnt" toks'
    in if isComma nextTok                    
         then let remainedToks = gtail "deleteEnt" toks' 
              in if remainedToks /= [] 
                   then let whites = takeWhile isWhiteSpace remainedToks
                        in if whites /= [] then deleteToks toks startPos' (tokenPos (last whites))
                                           else deleteToks toks startPos' (tokenPos nextTok)
                   else deleteToks toks startPos' (tokenPos nextTok)
         else deleteToks toks startPos' endPos  

--remove the idenitifers specified by 'pns' from the hiding clause in client modules.
--ATTENTION: Module name does not have location information. (Can this be added? ask programmatica)
rmItemFromExport mod@(HsModule loc modName exps imps ds) pns
  = do exps'<- case exps of
        Just ents -> let matchedEnts = findEnts pns ents
                     in if matchedEnts == [] 
                         then return exps
                         else
                           if length matchedEnts == length ents
                            then 
                              do ((toks,_),others)<-get  
                                 let (startPos,endPos) = startEndLoc toks ents
                                     toks'= deleteToks toks startPos endPos
                                 put ((toks',modified),others)
                                 return Nothing
                            else do ((toks,_),others)<-get  
                                    let remainedEnts = concatMap (\pn->filter (\e-> case e of
                                                                             EntE e' ->(not (isVarAndSame pn e'))
                                                                             _       -> True ) ents)  pns 
                                        toks' = foldl deleteEnt toks $ map (startEndLoc toks) matchedEnts
                                    put ((toks',modified),others)
                                    return (Just remainedEnts) 
        Nothing -> return exps 
       return (HsModule loc modName exps' imps ds)

    where
    findEnts pns ents =concatMap (\pn->filter (\e ->case e of 
                                                     ModuleE _ -> False
                                                     EntE e'    -> isVarAndSame pn e') ents) pns


parseSourceFile 
     =checkScope @@ parseSourceFile'         
   where 
     checkScope (ts,(((wm,_),mod),refs))
        = check (checkRefs refs) >> return (inscpRel wm, exports wm, mod, expandNewLnTokens ts) 

     check [] = done
     check errs = fail $ pp $ "Scoping errors" $$ vcat errs      

{- 
writeRefactoredFiles files
  = do addToHistory $ map fst files
       mapM_ modifyFile files
     where
       modifyFile (fileName,(ts,mod)) = do
           let source = concatMap (snd.snd) ts
           seq (length source) $ writeHaskellFile fileName source
           editorCmds <-getEditorCmds 
           MT.lift $ sendEditorModified editorCmds fileName        
           --seq (length source) $ writeFile (createNewFileName "_TokOut" fileName) source 
           writeFile (createNewFileName "AST" fileName) ((render.ppi.rmPrelude) mod)   
       createNewFileName str fileName 
          =let (name, posfix)=span (/='.') fileName
           in (name++str++posfix)
-}

writeRefactoredFiles files
  = do let modifiedFiles = filter (\((f,m),_)->m==modified) files
       addToHistory $ map (fst.fst) modifiedFiles
       mapM_ modifyFile modifiedFiles
       mapM_ writeTestDataForFile files
     where
       modifyFile ((fileName,_),(ts,mod)) = do
           let source = concatMap (snd.snd) ts
           seq (length source) $ writeHaskellFile fileName source
           editorCmds <-getEditorCmds 
           MT.lift $ sendEditorModified editorCmds fileName        
       writeTestDataForFile ((fileName,_),(ts,mod)) = do
           --seq (length source) $ writeFile (createNewFileName "_TokOut" fileName) source 
           writeFile (createNewFileName "AST" fileName) ((render.ppi.rmPrelude) mod)   
       createNewFileName str fileName 
          =let (name, posfix)=span (/='.') fileName
           in (name++str++posfix) 


expandNewLnTokens::[HToken]->[HToken]
expandNewLnTokens ts = concatMap expand ts
       where expand tok@(Whitespace,((row,col),s))
                = doExpanding (row,col) s                
             expand x = [x]

             doExpanding (row,col) [] =[]
             doExpanding (row,col) (t:ts)
                   =case t of 
                        ' ' -> (Whitespace, ((row,col),[t])):(doExpanding (row,col+1) ts)
                        '\n'-> (Whitespace, ((row,col),[t])):(doExpanding (row+1,1) ts)
                        '\t'-> (Whitespace, ((row,col),[t])):(doExpanding (row,col+1) ts)

 --show a list of PNames.
showPns f [] = ""
showPns f (pn:pns) = if pns==[] then f pn
                                else (f pn ++ "," ++ showPns f pns)
      
--These values is used to make uniqe locations

initValue= (-1000)

stepValue=100


locOffset@(rowOffset,colOffset)=(1000,1000)

nextValue=do (toks,(filename,int))<-get
             put (toks, (filename, int-stepValue))
             return (filename,int)


------Remove the 'Prelude' imports added by Programatica------------------------------
rmPrelude (HsModule s m e is ds) = HsModule s m e (is \\ prelimps) ds
  where
    prelimps = [HsImportDecl loc0 (Module "Prelude") True Nothing Nothing,
                HsImportDecl loc0 (Module "Prelude") False Nothing Nothing]
    
class (Term t) => HasModName t where
  
    hasModName :: t->Maybe ModuleName

instance HasModName PNT  where

    hasModName (PNT (PN _ (G modName s1 loc)) _ _)=Just modName
    hasModName _ =Nothing   

instance HasModName PName where
  
    hasModName (PN _ (G modName s1 loc))=Just modName
    hasModName _ =Nothing


--Test whether a string is a lexically valid variable/constructor identifier.
isVarId,isConId::String->Bool
isVarId id=id/=[] && isSmall (ghead "isVarId" id) && isLegalIdTail (tail id) && not (isReservedId id)
       
       
isConId id=id /=[] && isUpper (ghead "isConId" id) && isLegalIdTail (tail id) && not (isReservedId id)

isLegalIdTail tail=all isLegal tail
        where isLegal c=isSmall c|| isUpper c || isDigit c || c=='\''

isSmall c=isLower c || c=='_'

isReservedId id=elem id reservedIds 
   where reservedIds=["case", "class", "data", "default", "deriving","do","else" ,"if",
                       "import", "in", "infix","infixl","infixr","instance","let","module",
                       "newtype", "of","then","type","where","_"]

isVarSym id = id /= [] && isSym (ghead "isVarOp" id) && isLegalSymTail (tail id) && not (isReservedOp id)

isConSym id = id /=[]  && (ghead "isConSymbol" id ==':') && isLegalSymTail (tail id) && not (isReservedOp id)
  
isLegalSymTail tail = all isLegal tail
     where isLegal c = isSym c || c==':'

isSym id = elem id symbols
   where symbols = ['!', '#', '$', '%', '&', '*', '+','.','/','<','=','>','?','@','\'','^','|','-','~']

isReservedOp id = elem id reservedOps
   where reservedOps = ["..", ":","::","=","\"", "|","<-","@","~","=>"]

-----------------------------------------------------------------------------

isTopLevelPN,isLocalPN::PName->Bool

--return True if an identifier is a top level name
isTopLevelPN (PN i (G _ _ _))=True
isTopLevelPN _ =False

--return True if an identifier is a local name
isLocalPN (PN i (S _))=True
isLocalPN _ =False

isTopLevelPNT = isTopLevelPN.pNTtoPN

isQualifiedPN (PN (Qual mod id) _)=True
isQualifiedPN _ =False 

getQualFromPN (PN (Qual mod id) _ ) = Just mod
getQualFromPN _ = Nothing

--returns True if pn is a function/pattern name
isFunOrPatName pn 
   =isJust.(applyTU (once_tdTU (failTU `adhocTU` worker)))
     where
        worker (decl::HsDeclP)
           | defines pn decl=Just True
        worker _ =Nothing

--return True if pn is a local function/pattern name
isLocalFunOrPatName pn scope
 =isLocalPN pn && isFunOrPatName pn scope

--------------------------------------------------------------------------------
isTypeSig ::HsDeclP->Bool
isTypeSig (Dec (HsTypeSig loc is c tp))=True
isTypeSig _=False

isFunBind (Dec (HsFunBind loc matches))=True
isFunBind _ =False

isPatBind (Dec (HsPatBind _ _ _ _))=True
isPatBind _=False

isSimplePatBind decl=case decl of
     Dec (HsPatBind _ p _ _)->patToPName p /=defaultPN
     _ ->False

isComplexPatBind decl=case decl of
     Dec (HsPatBind _ p _ _)->patToPName p ==defaultPN
     _ -> False

isFunOrPatBind decl=isFunBind decl || isPatBind decl

isClassDecl (Dec (HsClassDecl _ _ _ _ _)) = True
isClassDecl _ = False

isInstDecl (Dec (HsInstDecl _ _ _ _ _)) = True
isInstDecl _ = False
        
isSimpleExp (Exp (HsId _))=True
isSimpleExp (Exp (HsLit _ _))=True
isSimpleExp _ =False

isParenExp (Exp (HsParen _))=True
isParenExp _=False

isLast (HsLast e)=True
isLast _=False

expInLast (HsLast e)=Just e
expInLast _=Nothing

definedNames (Dec (HsFunBind _ ((HsMatch _ (PNT pname _ _) _ _ _):_))) =[pname]
definedNames (Dec (HsPatBind _ p _ _)) =pNamesInPat p
definedNames _=[] 

isRecursiveDef (Dec (HsFunBind loc ms)) 
   = any isUsedInDef ms 
   where isUsedInDef (HsMatch loc1 fun pats rhs ds)
            =findPName (pNTtoPN fun) rhs
isRecursiveDef _ = False

----------------------------------------------------------------------------------
defaultPN=PN (UnQual "--") (G (Module "--") "--" (N Nothing)) :: PName

defaultPNT=PNT defaultPN Value  (N Nothing) :: PNT 

defaultModName = Module "--"

defaultExp=Exp (HsId (HsVar defaultPNT))

defaultPat=Pat (HsPId (HsVar defaultPNT))

samePNT (PNT pname1 ty1 (N loc1)) (PNT pname2 ty2 (N loc2)) = pname1==pname2 && loc1==loc2
                         
pNTtoPN :: PNT->PName
pNTtoPN (PNT pname _ _)=pname

--This function does not take the qualifier into account.
pNtoName (PN (UnQual i) orig)=i
pNtoName (PN (Qual modName i) orig)=i   

pNTtoName=pNtoName.pNTtoPN

pNtoQuotedName pn
  = let (r,c)=pNameToLoc pn
    in  " '"++pNtoName pn++"'" ++"(at row:"++show r ++ " col:" ++ show c ++")"

pNtoId :: PN (HsName.Id) ->String
pNtoId (PN i orig)=i

unqualPNtoQualPN  qual pn 
 = case pn of 
      PN (UnQual n) ty -> PN (Qual qual n ) ty
      _                -> pn


---------------------------------------------------------------------------------

--PROBLEM :: Loc0 can not be used here!!

idToLocalPName name=do (filename,int)<-nextValue
                       let loc=SrcLoc filename int int 
                       return (PN (UnQual name) (S loc))

idToTopLevelPName name modName
   = do (filename,int)<-nextValue
        let loc=SrcLoc filename int int 
        return (PN (UnQual name) (G modName name (N (Just loc))))

idToTopLevePNT name modName
   = do (filename,int)<-nextValue
        let loc=SrcLoc filename int int 
        return (PNT (PN (UnQual name) (G modName name (N (Just loc)))) Value (N (Just loc)))


--idToPat converts an identifier into a pattern.
idToPat name
    =do (filename, int)<-nextValue
        let loc=SrcLoc filename int int 
        return (Pat (HsPId (HsVar (PNT (PN (UnQual name) (S loc)) Value (N (Just loc)))))) 

pNameToOrig (PN _ (G _ _ (N (Just loc)))) = loc
pNameToOrig (PN _ (G _ _ (N Nothing)))=loc0
pNameToOrig (PN _ (S loc)) = loc 

--this function should be removed later. the calling functions should use pNameToOrig.
pNameToLoc (PN _ (G _ _ (N (Just (SrcLoc _ row col))))) = (row,col)
pNameToLoc (PN _ (S (SrcLoc _ row col))) = (row,col) 
pNameToLoc (PN _ (G _ _ (N Nothing))) = defaultPos

pNtoModName (PN _ (G modName _ _)) = modName
pNtoModName (PN _ (S _))           = defaultModName

pNTtoModName = pNtoModName.pNTtoPN

pNameToPat :: PName -> HsPatP
pNameToPat pname
    =let loc=pNameToOrig pname
     in (Pat (HsPId (HsVar (PNT pname Value (N (Just loc))))))

pNameToExp pname
   =do (filename, int)<-nextValue 
       let loc=SrcLoc filename int int
       return (Exp (HsId (HsVar (PNT pname Value (N (Just (SrcLoc filename int int)))))))

pNameToPNT pname ty
   =do (filename, int)<-nextValue 
       let loc=SrcLoc filename int int
       return (PNT pname ty (N (Just (SrcLoc filename int int))))

--idToExp converts an identifer into an expression.
idToExp ::String ->HsExpP
idToExp name
    =Exp (HsId (HsVar (PNT (PN (UnQual name) (S loc0)) Value (N (Just loc0)))))

pnToExp::PName->HsExpP
pnToExp pn =Exp (HsId (HsVar (PNT pn  Value (N (Just loc0)))))


expToPNT (Exp (HsId (HsVar pnt)))=pnt 
expToPNT (Exp (HsParen e))=expToPNT e 
expToPNT _ =defaultPNT

expToPName =pNTtoPN.expToPNT

patToPNT (Pat (HsPId (HsVar pnt)))= pnt
patToPNT (Pat (HsPParen p))=patToPNT p
patToPNT _=defaultPNT

patToPName=pNTtoPN.patToPNT

mkNewName oldName fds init
  =let newName=if init==0 then oldName
                          else oldName++"_"++ show init
   in if elem newName fds
        then mkNewName oldName fds (init+1)
        else return newName 

{- Create the auxiliary function name. Suppose the original function name is 'f', 
   then the auxiliary function name is like 'f_gen_i', where i is an integer. -}

mkNewFunPName pn e fileName modName inscopeNames
   =do  (f1,d1) <- freeAndDeclaredNames e
        name    <- mkNewName ((pNtoName pn)++"_gen") (nub (f1 `union` d1 `union` inscopeNames)) 0  
        let loc = SrcLoc fileName initValue initValue 
        return (PN (UnQual name) (G modName name (N (Just loc))))

-----Find the identifier (in PNT format) whose start position is (row,col) in fileName ---
locToPNT  fileName row col 
  =(fromMaybe defaultPNT). applyTU (once_buTU (failTU `adhocTU` worker))
     where 
        worker pnt@(PNT pname ty (N (Just (SrcLoc fileName1 row1 col1))))
           |fileName1==fileName && row1==row && col1==col =Just pnt
        worker _ =Nothing 

locToPName fileName row col =pNTtoPN.(locToPNT fileName row col)


----------------------------------------------------------------------------
{-This function returns True if the pnt is in a use site, otherwise returns False.
  It is different from the one defined in Class findEntity.  -}

pNTinUseLoc pnt=case pnt of
                   (PNT pname ty (N (Just loc)))->srcLoc pname/=loc
                   _->False

findPName pn 
  =(fromMaybe False).(applyTU (once_tdTU (failTU `adhocTU` worker)))
     where 
        worker (pn1::PName)
           |pn == pn1 && pNameToOrig pn == pNameToOrig pn1 = Just True
        worker _ =Nothing 

findNameInUseLoc name t
   =(fromMaybe False) (applyTU (once_tdTU (failTU `adhocTU` worker)) t)
      where
         worker (pnt::PNT)
           |pNTtoName pnt ==name && pNTinUseLoc pnt && not (isQualifiedPN (pNTtoPN pnt)) && not (findPntInLhs pnt t)
          = Just True
         worker _ =Nothing


--a multiparameter class
class (Term b)=>FindEntity a b  where
  
  findEntity :: a->b->Bool

instance (Term b)=>FindEntity HsExpP b  where

  findEntity e b
    =(fromMaybe False)(applyTU (once_tdTU (failTU `adhocTU` inExp)) b)
     where
       inExp (e1::HsExpP)
         | e==e1 =Just True
       inExp _ =Nothing

instance (Term b)=>FindEntity PName b  where

  findEntity pn b   
   =(fromMaybe False)(applyTU (once_tdTU (failTU `adhocTU` worker)) b)
     where 
        worker (Exp (HsId (HsVar (PNT pname _ _))))
           |pname==pn= Just True
        worker _ =Nothing 

instance (Term b)=>FindEntity PNT b where

  findEntity pnt b
   =(fromMaybe False)(applyTU (once_tdTU (failTU `adhocTU` worker)) b)
      where 
        worker (pnt1::PNT)
           |samePNT pnt pnt1 = Just True

        worker _ =Nothing 


 --Find the largest leftmost expression in the scope specified by beginPos and endPos.
locToExp1 toks beginPos endPos t
  =fromMaybe defaultExp (applyTU (once_tdTU (failTU `adhocTU` exp)) t)
      where
         exp (e::HsExpP)
           |inScope (startEndLoc toks e) =Just e
         exp _ =Nothing                     
   
         inScope (startLoc,endLoc)=startLoc>=beginPos && startLoc <= endPos && endLoc>=startPos &&  endLoc<=endPos 


 --Find the largest leftmost expression in the scope specified by beginPos and endPos.
locToExp toks  beginPos endPos t 
  = fromMaybe defaultExp (applyTU (once_tdTU (failTU `adhocTU` exp)) t)
      where
         exp (e::HsExpP)
           |inScope e =Just e
         exp _ =Nothing                     
   
         inScope (e::HsExpP)
            = let (startLoc', endLoc') = startEndLoc toks e
                  allLocs            = allSortedLocs e
                  (startLoc, endLoc)  
                      = if allLocs == [] then (startLoc', endLoc')
                                         else ( if startLoc' == defaultPos then  head allLocs  else  startLoc',
                                                if endLoc'   == defaultPos then  last allLocs  else endLoc')
              in  startLoc>=beginPos && startLoc <= endPos && endLoc>=beginPos &&  endLoc<=endPos && 
                  length (rmWhites (render (ppi e))) <= length (rmWhites ( concatMap (snd.snd) subToks))
                  -- comparing the length is a bit inacurrate, it's better to use subString.
 
         rmWhites s = deleteBy (\x->x==' ') s 
            where 
            deleteBy f (y:ys) = if f y then deleteBy f ys else y:(deleteBy f ys)
            deleteBy f [] = [] 

         subToks = getToks (beginPos, endPos) toks 

--------------------------------------------------------------------------------  
--Attention: Rename this function name to pNsIn 
--find all the pNames in a pattern---  
pNamesInPat =(nub.ghead "pNamesInPat").applyTU (full_tdTU (constTU [] `adhocTU` inPnt)) --no problem with 'head'
  where 
     inPnt (PNT pname ty loc)
          =return [pname]
     

pNTsIn =(nub.ghead "pNTsIn").applyTU(full_tdTU (constTU [] `adhocTU` inPnt))
   where
     inPnt pnt@(PNT _  _ _)
        = return [pnt]


-------------------------------------------------------------------

matchesInDecls ((Dec (HsFunBind loc matches))::HsDeclP)=matches
matchesInDecls x=[]          
       
patsInMatch ((HsMatch loc1 name pats rhs ds)::HsMatchP)
   =pats


--return True if there is a declaration definind pn. 
findDecl pn decls typeSigInc
  =if typeSigInc then isJust (find (\d->defines pn d || definesType pn d ) decls)  
                 else isJust (find (defines pn) decls)

---return True if a function/pattern binding defines pn
defines pn decl@(Dec (HsFunBind loc ((HsMatch loc1 (PNT pname ty loc2) pats rhs ds):ms))) 
          =pname==pn
defines pn decl@(Dec (HsPatBind loc p rhs ds)) =elem pn (pNamesInPat p)
defines pn decl=False

--return True if this is a type signature defining pn's type
definesType pn (Dec (HsTypeSig loc is c tp))=elem pn (map pNTtoPN is)  --consider the name again
definesType pn decl=False


definesSingleId  (Dec (HsTypeSig loc is c tp)) = length is ==1 

--get the declaration (function or pattern binding) defining pn
declDefinesPn pn t=let decls=applyTU (once_tdTU (failTU `adhocTU` decl)) t 
                   in if decls/=[] then head decls     
                                   else []
     where
       decl (d::HsDeclP)
        |defines pn d=return [d]
       decl d=mzero 

--Find those declarations(function/pattern binding and type signature) which define pn-
--splitTypeSig indicates whether the corresponding type signature will be splited.
definingDecls::[PName]->[HsDeclP]->Bool->[HsDeclP]
definingDecls pns ds incTypeSig=concatMap (defines pns) ds 
      where 
       defines pn decl@(Dec (HsFunBind loc ((HsMatch loc1 (PNT pname ty loc2) pats rhs ds):ms)))
          | isJust (find (==pname) pns)=[decl]
       defines pn decl@(Dec (HsPatBind loc p rhs ds))    ---CONSIDER AGAIN----
          |(pNamesInPat p) `intersect` pns /=[]=[decl]
       defines pn decl@(Dec (HsTypeSig loc is c tp)) --handle cases like  a,b::Int 
          |(map pNTtoPN is) `intersect` pns /=[]
          =if incTypeSig then [(Dec (HsTypeSig loc (filter (\x->isJust (find (==pNTtoPN x) pns)) is) c tp))]
                           else [decl]
       defines pn decl=[]   



{- Divide a declaration list(ds) into three parts (before, parent, after) according to pn,
   where 'parent' is the first decl containing 'pn', 'before' are those decls before 'parent' and
   'after' are those decls after 'parent'. -}

--REFACTOR THESE TWO FUNCTIONS:

divideDecls ds pn
  =let (before, after)=break (\x->(findPName pn x)) ds
   in if (after/=[])
         then (before, [ghead "divideDecls" after],tail after)  
         else (ds,[],[])

divideDeclsAccordingToPNT ds pnt
  = let (before,after)=break (\x->isJust (findPNT pnt x)) ds
    in if (after/=[])
         then (before, [ghead "divideDeclsAccoringToPNT" after], tail after)
         else (ds,[],[])
 
---------------remove type signature for pn----------------------------------

removeTypeSig ::PName->[HsDeclP]->[HsDeclP]
removeTypeSig pn decls=concatMap (removeTypeSig' pn) decls
      where removeTypeSig' pn sig@(Dec (HsTypeSig loc is c tp))
             =if definesType pn sig && length is==1 
                 then []
                 else [Dec (HsTypeSig loc (filter (\x-> (pNTtoPN x)/=pn) is) c tp)]
            removeTypeSig' pn x=[x]   


{-remove source location information in the syntax tree, that is to replace all
  locations by default location: loc0 -}       
rmLocInfo t =runIdentity (applyTP (full_tdTP (idTP `adhocTP` exp
                                                   `adhocTP` pnt
                                                   `adhocTP` alt
                                                   `adhocTP` lit)) t) 
   where 
         exp ((Exp (HsNegApp loc e)) ::HsExpP)=return (Exp (HsNegApp loc0 e))
         exp (Exp (HsExpTypeSig loc e c t))=return (Exp (HsExpTypeSig loc0 e c t))
         exp x=return x 

         alt ((HsAlt loc p e ds)::HsAltP)=return (HsAlt loc0 p e ds)

         lit (HsInt loc int)=return (HsInt loc0 int)
         lit (HsFrac loc rat)=return (HsFrac loc0 rat)

         pnt (PNT pname ty _)=return (PNT pname ty (N Nothing))

{-only remove source location instead of both source location and defining location-}   
rmLocInfo1 fileName t =runIdentity (applyTP (full_tdTP (idTP `adhocTP` atSrcLoc)) t) 
   where 
         atSrcLoc ((N (Just loc@(SrcLoc fileName1 row col)))::OptSrcLoc)
            |fileName==fileName1 =return (N (Just (SrcLoc fileName1 (-1) (-1))))    
         atSrcLoc x = return x

 -----SOME FUNCTIONS INVOLVING RENAMEING/REPLACING---------------------------------        
------------------------------------------------------------------------------------------ 
replaceNameInPN (PN (UnQual s)(S loc)) newName
              =(PN (UnQual newName) (S loc))   
replaceNameInPN (PN (Qual modName  s)(S loc)) newName
              =(PN (Qual modName newName)(S loc))
replaceNameInPN (PN (UnQual s) (G modName s1 loc)) newName
              =(PN (UnQual newName) (G modName newName loc)) 
replaceNameInPN (PN (Qual modName s) (G modName1 s1 loc)) newName
              =(PN (Qual modName newName) (G modName1 newName loc)) 


replaceNameInPN1 qualifier (PN (UnQual s)(S loc))  newName
              =(PN (Qual qualifier newName) (S loc)) 
replaceNameInPN1 qualifier (PN (Qual modName  s)(S loc))  newName
              =(PN (Qual modName newName)(S loc))
replaceNameInPN1 qualifier (PN (UnQual s) (G modName s1 loc))  newName
              =(PN (Qual qualifier  newName) (G modName newName loc)) 
replaceNameInPN1 qualifier (PN (Qual modName s) (G modName1 s1 loc))  newName
              =(PN (Qual modName newName) (G modName1 newName loc)) 

replaceNameAndLocInPN (PN (UnQual s)(S loc)) newName newLoc
              =(PN (UnQual newName) (S newLoc)) 

replaceNameAndLocInPN (PN (Qual modName  s)(S loc)) newName newLoc
              =(PN (Qual modName newName)(S newLoc))
replaceNameAndLocInPN (PN (UnQual s) (G modName s1 (N (Just loc)))) newName newLoc
              =(PN (UnQual newName) (G modName newName (N (Just newLoc))))
replaceNameAndLocInPN (PN (Qual modName s) (G modName1 s1 (N (Just loc)))) newName newLoc
              =(PN (Qual modName newName) (G modName newName (N (Just newLoc))))
replaceNameAndLocInPn pn newName newLoc = pn


pNameToStr (PN (UnQual s) _) =s
pNameToStr (PN (Qual (Module qualifier) s) _)=qualifier ++ "."++ s

renamePN oldPN newName
    =applyTP(full_tdTP(adhocTP idTP rename))  
        where
           rename pn
              | pn==oldPN && pNameToOrig oldPN == pNameToOrig pn
             = return (replaceNameInPN pn newName)
           rename x=return x



--Renaming an identifier by  creating a new name automatically--------------------
--if updateToks is true, the token stream will be updated, otherwise only 
--the AST is updated.

renamePNwithUpdLexer oldPN newName t
    =applyTP (full_tdTP (adhocTP idTP rename)) t
        where
           rename  (PNT pn ty (N (Just (SrcLoc fileName  row col))))
            | (pn ==oldPN) && (pNameToLoc oldPN == pNameToLoc pn)
               = do ((toks,_),others)<-get
                    let toks'=replaceToks toks (row,col) (row,col) 
                               [mkToken Varid  (row,col) (pNameToStr (replaceNameInPN pn newName))]
                    put ((toks',modified),others)
                    return (PNT (replaceNameInPN pn newName) ty (N (Just (SrcLoc fileName  row col))))
           rename x=return x


renamePNwithUpdLexer1 oldPN qualifier newName  t
    =applyTP (full_tdTP (adhocTP idTP rename)) t
       where
         rename  pnt@(PNT pn ty (N (Just (SrcLoc fileName  row col))))
          | (pn ==oldPN) && (pNameToLoc oldPN == pNameToLoc pn)
          = do ((toks,_),others)<-get                           
               let toks'=replaceToks toks (row,col) (row,col)
                         [mkToken Varid  (row,col) (pNameToStr (replaceName pn  newName))]
               put ((toks', modified),others)
               return (PNT (replaceName pn newName) ty (N (Just (SrcLoc fileName  row col))))
           where replaceName = if not (pNTinUseLoc pnt) || findPntInLhs pnt t || findPntInImp pnt t
                                  then replaceNameInPN   
                                  else (replaceNameInPN1 qualifier) 
         rename x=return x



findPntInLhs pnt
  = (fromMaybe False).(applyTU (once_tdTU (failTU `adhocTU` inMatch
                                                  `adhocTU` inDecl)))
                                                   
   where
    inMatch ((HsMatch loc1 name pats rhs ds)::HsMatchP)
      | isJust (find (samePNT pnt) [name]) = Just True
    inMatch _ =Nothing

    inDecl ((Dec (HsTypeSig loc is c tp))::HsDeclP)
       |isJust (find (samePNT pnt) is)   = Just True
    inDecl _ =Nothing
    
findPntInImp pnt 
  =(fromMaybe False).(applyTU (once_tdTU (failTU `adhocTU` inImp)))
     where 
        inImp ((HsImportDecl loc modName qual  as h)::HsImportDeclI PNT)
          = findPNT pnt h
        
findPNT pnt 
   =applyTU (once_tdTU (failTU `adhocTU` worker))
     where
       worker (pnt1::PNT)
         |samePNT pnt pnt1 =Just True
       worker _ =Nothing  




--Renaming an identifier by  creating a new name automatically--------------------
--if updateToks is true, the token stream will be updated, otherwise only 
--the AST is updated.

autoRenameLocalName updateToks oldPN 
  =applyTP (once_buTP (failTP `adhocTP` renameInMatch 
                              `adhocTP` renameInPat
                              `adhocTP` renameInExp
                              `adhocTP` renameInAlt
                              `adhocTP` renameInStmts))  
      where 
         renameInMatch (match::HsMatchP)
           |isDeclaredIn oldPN match=worker match
         renameInMatch _ =mzero
     
         renameInPat (pat::HsDeclP)
          |isDeclaredIn oldPN pat=worker pat                 
         renameInPat _ =mzero
     
         renameInExp (exp::HsExpP)
          |isDeclaredIn oldPN exp=worker exp                 
         renameInExp _ =mzero
 
         renameInAlt (alt::HsAltP)
          |isDeclaredIn oldPN alt=worker alt                 
         renameInAlt _ =mzero
       
         renameInStmts (stmt::HsStmtP)
          |isDeclaredIn oldPN stmt=worker stmt
         renameInStmts _=mzero
         
         worker t =do (f,d)<-freeAndDeclaredNames t
                      ds<-hsVisibleNames oldPN (declsIn t)
                      newName<-mkNewName (pNtoName oldPN) (nub (f `union` d `union` ds)) 1
                      if updateToks
                        then renamePNwithUpdLexer oldPN newName t 
                        else renamePN oldPN newName t     
              
   
--substitute an old expression by new expression
replaceExp decls subst
   = applyTP (full_buTP (idTP `adhocTP` worker)) decls
         where worker (e::HsExpP)
                 |(expToPName e/=defaultPN) &&  (expToPName e) == (fst subst)
                     =return (snd subst)
               worker x=return x 


--replaceDecls replaces a list of old declaraions by a list of new declarations---                                    
class (Term t) => ReplaceDecls t where
  
    replaceDecls :: t->[HsDeclI PNT]->t

instance ReplaceDecls HsExpP where
    replaceDecls (Exp (HsLet ds e)) ds'
            =if ds'==[]
                then e 
                else (Exp (HsLet ds' e))
                     
    replaceDecls (Exp (HsListComp (HsLetStmt ds stmts))) ds'
            =if ds'==[] && isLast stmts 
               then (Exp (HsList [fromJust (expInLast stmts)]))
               else (Exp (HsListComp (HsLetStmt ds' stmts)))
    replaceDecls (Exp (HsDo (HsLetStmt ds stmts))) ds'
            =if ds'==[] 
                then (Exp (HsDo stmts))
                else (Exp (HsDo (HsLetStmt ds' stmts)))
    replaceDecls x ds'=x
   
   
instance ReplaceDecls HsMatchP where   
    replaceDecls (HsMatch loc1 fun pats rhs ds) ds'
                 =(HsMatch loc1 fun pats rhs ds')
           
instance ReplaceDecls HsDeclP where
    replaceDecls (Dec (HsPatBind loc p rhs ds)) ds'
                 =Dec (HsPatBind loc p rhs ds')
    replaceDecls x ds' =x

instance ReplaceDecls HsModuleP where
    replaceDecls (HsModule loc modName exps imps ds) ds'
                  =HsModule loc modName exps imps ds'

instance ReplaceDecls HsAltP where
   replaceDecls (HsAlt loc p rhs ds) ds'=HsAlt loc p rhs ds'

instance ReplaceDecls HsStmtP where
   replaceDecls (HsLetStmt ds stmts) ds'
                    =if ds'/=[] then  HsLetStmt ds' stmts         
                                else stmts

--replaceRhs replace the old rhs by new rhs---- 
class (Term t) => ReplaceRhs t where
  
    replaceRhs:: t->RhsP->t

instance ReplaceRhs HsMatchP where   
    replaceRhs (HsMatch loc1 fun pats rhs ds) rhs'
                 =(HsMatch loc1 fun pats rhs' ds)
           
instance ReplaceRhs HsDeclP where
    replaceRhs (Dec (HsPatBind loc p rhs ds)) rhs'
                =Dec (HsPatBind loc p rhs' ds)
    replaceRhs x _ =x
         
instance ReplaceRhs HsAltP where
    replaceRhs (HsAlt loc p rhs ds) rhs'
                 =HsAlt loc p rhs' ds


replaceExpInLet (Exp (HsLet ds e)) e'=(Exp (HsLet ds e'))
replaceExpInLet x _=x 

replaceStmtsInDo (HsLetStmt ds stmts) stmts'=HsLetStmt ds stmts'
replaceStmtsInDo x _=x

----------------------------------------------------------------------
-----isDeclaredIn tests whether pn is declared in scope t--------
class (Term t) => IsDeclaredIn t where
  
    isDeclaredIn :: PName->t->Bool 

instance IsDeclaredIn HsMatchP where   
    isDeclaredIn  pn match@(HsMatch loc1 (PNT fun _ _) pats rhs ds)
             =fromMaybe False ( do (_,d)<-freeAndDeclaredVbls match
                                   Just (elem pn (d \\ [fun])))

instance IsDeclaredIn HsDeclP where
    isDeclaredIn pn (Dec (HsPatBind loc p rhs ds))
            =fromMaybe False (do (_, rd)<-hsFreeAndDeclared rhs
                                 (_, dd)<-hsFreeAndDeclared ds
                                 Just (elem pn (rd `union` dd)))               
    isDeclaredIn pn _ =False 

instance IsDeclaredIn HsModuleP where
    isDeclaredIn pn (HsModule loc modName exps imps ds)
           =fromMaybe False  (do (rf,rd)<-hsFreeAndDeclared ds
                                 Just (elem pn rd))

instance IsDeclaredIn HsExpP where
                                                                 
    isDeclaredIn pn (Exp (HsLambda pats body))
            = fromMaybe False (do (pf,pd) <-hsFreeAndDeclared pats
                                  Just (elem pn  pd))

    isDeclaredIn pn (Exp (HsLet decls e))        
           =fromMaybe False (do (df,dd)<- hsFreeAndDeclared decls
                                Just (elem pn dd))
        
    isDeclaredIn pn _=False 
                   
  
instance IsDeclaredIn HsAltP where
    isDeclaredIn pn (HsAlt _ pat exp decls)
       =fromMaybe False ( do (pf,pd) <- hsFreeAndDeclared pat
                             (df,dd) <- hsFreeAndDeclared decls
                             Just (elem pn (pd `union` dd)))
    
    
instance IsDeclaredIn HsStmtP where
      
    isDeclaredIn pn (HsGenerator _ pat exp stmts) -- Claus
        =fromMaybe False (do (pf,pd) <-hsFreeAndDeclared pat
                             Just (elem pn pd))

    isDeclaredIn pn (HsLetStmt decls stmts)
        =fromMaybe False (do (df,dd) <-hsFreeAndDeclared decls
                             Just (elem pn dd))

    isDeclaredIn pn _=False
                                    
-- *********FUNCTIONS ADDING NEW PARAMETERS**********************

---Adding new parameters to declarationa, which involves adding parameters to both expressions and patterns.


addParamsToDecls decls pn paramPNames
    =if (paramPNames/=[])
       then  do decls'<-addParamsToExps pn paramPNames decls
                let decls''=addParamsToPats paramPNames decls'                                                            
                mkLocsUnique (map pNameToLoc paramPNames) decls''
       else return decls 

  where 
    mkLocsUnique locs = applyTP (stop_tdTP (failTP `adhocTP` inMatch
                                                   `adhocTP` inPatBind))
        where
             inMatch (match@(HsMatch loc name pats rhs ds)::HsMatchP)
                  =do (_,offset)<-nextValue 
                      mkLocsUnique' locs offset match 
             inPatBind (pat@(Dec (HsPatBind loc p rhs ds))::HsDeclP)
                  =do  (_,offset)<-nextValue
                       mkLocsUnique' locs offset pat                      
             inPatBind _=mzero
       
             mkLocsUnique' locs offset t=do applyTP (full_td (idTP `adhocTP` inLoc)) t
               where 
                 inLoc loc@(SrcLoc fileName row col)      --SHOULD ALSO COMPARE FILENAME TO AVOID COINCIDENCE.
                   = if isJust (find (==(row, col)) locs) --return True means current loc belongs to locs.
                        then return (SrcLoc fileName (row+offset) (col+offset))
                        else return loc          

mkDefaultLocsUnique =applyTP (stop_tdTP (failTP `adhocTP` inLoc))
   where 
      inLoc loc@(SrcLoc fileName (-1) (-1))
         =do (_,int)<-nextValue
             return (SrcLoc fileName int int)
      inLoc _=mzero

     
--adding new parameters to expressions
addParamsToExps pn paramPNames
     =applyTP(stop_tdTP (failTP `adhocTP` worker))  
        where                   
          worker exp@(Exp (HsId (HsVar (PNT pname ty loc))))
                  | pname==pn
                      =do exps<-mapM pNameToExp paramPNames
                          return (foldl addParamToExp exp exps)
          worker x =mzero
          addParamToExp  exp param
              =(Exp (HsApp exp param))
 
--adding new parameters to patterns.
addParamsToPats paramPNames decls 
     =map (addParamToDecl (map pNameToPat paramPNames)) decls
  where
    addParamToDecl params  (Dec (HsFunBind loc matches)) 
        =(Dec (HsFunBind loc (map(addParamtoMatch params) matches)))
       where
         addParamtoMatch param (HsMatch loc  fun  pats rhs  decls) 
            =(HsMatch loc  fun  (param++pats)  rhs  decls)

    addParamToDecl params (Dec (HsPatBind loc p rhs ds))
          |patToPNT p/=defaultPNT =(Dec (HsFunBind loc [HsMatch loc (patToPNT p) params rhs ds]))
       -- add parameters in lambda style to pattern binding
       --PROBLEM : how about guard cases?
    addParamToDecl params (Dec (HsPatBind loc p (HsBody e) ds))
        =(Dec (HsPatBind loc p (HsBody (Exp (HsLambda params e))) ds)) 
    addParamToDecl params x=x 



-- *************BEGIN OF  COLLECTING FUNCTIONS**********************-

--hsDataConstrs collects data constuctors--
hsDataConstrs modName  
    = ghead "hsDataConstrs". (applyTU (stop_tdTU (failTU `adhocTU` pnt)))  
       where
          pnt (PNT pname (ConstrOf _  _) _)
              = if hasModName pname==Just modName then return ([],[pname])
                                             else return ([pname],[])
          pnt _ = mzero


--hsTypeConstrAndClasses collects type constructors and class names--
hsTypeConstrsandClasses modName
    = ghead "hsTypeConstrAndClasses".(applyTU (stop_tdTU (failTU `adhocTU` pnt))) 
        where 
           pnt (PNT (PN i (G  modName1 id loc)) (Type _) _)
                 =if modName == modName1
                     then return ([],[(PN i (G  modName id loc))])
                     else return ([(PN i (G  modName id loc))], [])
           pnt (PNT pname (Class _) _)=if hasModName pname==Just modName 
                                          then return ([],[pname])
                                          else return ([pname],[])
           pnt _ =mzero

--hsTypeVbls collects type variables--
hsTypeVbls _ =ghead "hsTypeVbls".(applyTU (stop_tdTU (failTU `adhocTU` pnt))) 
        where 
           pnt (PNT (PN i (S loc)) (Type _) _)
                = return ([], [(PN i (S loc))])
           pnt _ =mzero

hasFreeVars e = fromMaybe False (do (f,_)<-hsFreeAndDeclared e
                                    if f/=[] then Just True
                                             else Nothing)
--hsClassMembers collects class instance names--
hsClassMembers className modName 
    = ghead "hsClassMembers". (applyTU (stop_tdTU (failTU `adhocTU` pnt))) 
         where
            pnt(PNT pname (MethodOf i _) _) -- Claus
                | pNtoId i==className
                  = if hasModName pname == Just modName 
                      then return ([],[pname])
                      else return ([pname],[])
            pnt _ = mzero    
            

--hsFreeAndDeclared collects free and declared variables visible outside (mostly by Ralf Lammel)
hsFreeAndDeclared:: (Term t, MonadPlus m)=> t-> m ([PName],[PName])
hsFreeAndDeclared t=do (f,d)<-hsFreeAndDeclared' t
                       return (nub f, nub d)
   where 
          hsFreeAndDeclared'=applyTU (stop_tdTU (failTU  `adhocTU` exp
                                                         `adhocTU` pat
                                                         `adhocTU` match
                                                         `adhocTU` patBind
                                                         `adhocTU` alt
                                                         `adhocTU` decls
                                                         `adhocTU` stmts
                                                         `adhocTU` recDecl))  
                          
          exp (Exp (HsId (HsVar (PNT pn _ _))))=return ([pn],[])
          exp (Exp (HsId (HsCon (PNT pn _ _))))=return ([pn],[])
          exp (Exp (HsInfixApp e1 (HsVar (PNT pn _ _)) e2))
              =addFree pn (hsFreeAndDeclared [e1,e2])
          exp (Exp (HsLambda pats body))
              = do (pf,pd) <-hsFreeAndDeclared pats
                   (bf,_) <-hsFreeAndDeclared body
                   return ((bf `union` pf) \\ pd, [])
          exp (Exp (HsLet decls exp))
              = do (df,dd)<- hsFreeAndDeclared decls
                   (ef,_)<- hsFreeAndDeclared exp 
                   return ((df `union` (ef \\ dd)),[])
          exp (Exp (HsRecConstr (PNT pn _ _) e))
               =addFree  pn  (hsFreeAndDeclared e)   --Need Testing
          exp (Exp (HsAsPat (PNT pn _ _) e))
              =addFree  pn  (hsFreeAndDeclared e)  
          exp _ = mzero

          
          pat (Pat (HsPId (HsVar (PNT pn _ _))))=return ([],[pn])
          pat (Pat (HsPInfixApp p1 (HsVar (PNT pn _ _)) p2))=addFree pn (hsFreeAndDeclared [p1,p2])
          pat (Pat (HsPApp (PNT pn _ _) pats))=addFree pn (hsFreeAndDeclared pats)
          pat (Pat (HsPRec (PNT pn _ _) fields))=addFree pn (hsFreeAndDeclared fields)
          pat _ =mzero
                               

          match ((HsMatch _ (PNT fun _ _)  pats rhs  decls)::HsMatchP)
            = do (pf,pd) <- hsFreeAndDeclared pats
                 (rf,rd) <- hsFreeAndDeclared rhs
                 (df,dd) <- hsFreeAndDeclared decls
                 return ((pf `union` ((rf `union` df) \\ (dd `union` pd `union` [fun]))),[fun])

         -------Added by Huiqing Li-------------------------------------------------------------------

          patBind ((Dec (HsPatBind _ pat (HsBody rhs) decls))::HsDeclP)
             =do (pf,pd) <- hsFreeAndDeclared pat
                 (rf,rd) <- hsFreeAndDeclared rhs
                 (df,dd) <- hsFreeAndDeclared decls
                 return (pf `union` ((rf `union` df) \\(dd `union` pd)),pd)
          patBind _=mzero
         ------------------------------------------------------------------------------------------- 

          alt ((HsAlt _ pat exp decls)::(HsAlt (HsExpP) (HsPatP) [HsDeclP]))
             = do (pf,pd) <- hsFreeAndDeclared pat
                  (ef,ed) <- hsFreeAndDeclared exp
                  (df,dd) <- hsFreeAndDeclared decls
                  return (pf `union` (((ef \\ dd) `union` df) \\ pd),[])


          decls (ds :: [HsDeclP])
             =do (f,d) <-hsFreeAndDeclaredList ds
                 return (f\\d,d)
          
          stmts ((HsGenerator _ pat exp stmts) :: HsStmt (HsExpP) (HsPatP) [HsDeclP]) -- Claus
             =do (pf,pd) <-hsFreeAndDeclared pat
                 (ef,ed) <-hsFreeAndDeclared exp
                 (sf,sd) <-hsFreeAndDeclared stmts
                 return (pf `union` ef `union` (sf\\pd),[])

          stmts ((HsLetStmt decls stmts) :: HsStmt (HsExpP) (HsPatP) [HsDeclP])
             =do (df,dd) <-hsFreeAndDeclared decls
                 (sf,sd) <-hsFreeAndDeclared stmts
                 return (df `union` (sf \\dd),[])
          stmts _ =mzero

          recDecl ((HsRecDecl _ _ _ _ is) :: HsConDeclI PNT (HsTypeI PNT) [HsTypeI PNT])
                =do let d=map pNTtoPN $ concatMap fst is
                    return ([],d)
          recDecl _ =mzero
            
       
addFree free mfd=do (f,d)<-mfd
                    return ([free] `union` f, d)

hsFreeAndDeclaredList l=do fds<-mapM hsFreeAndDeclared l
                           return (foldr union [] (map fst fds),
                                   foldr union [] (map snd fds))
 

instance (Monoid a, Monoid b) => Monoid (a,b) where
   mappend (a,b) (a',b') = (mappend a a', mappend b b')
   mempty                = (mempty, mempty)
           

hsFreeAndDeclaredNames t =do (f1,d1)<-hsFreeAndDeclared t
                             return ((nub.map pNtoName) f1, (nub.map pNtoName) d1)

--freeAndDeclaredVbls collects free and declared variables visible inside

----ATTENTION:THIS CLASS IS NOT COMPLETE----

class (Term t)=>FreeAndDeclaredVbls t where

   freeAndDeclaredVbls :: (MonadPlus m)=>t->m ([PName],[PName])

instance FreeAndDeclaredVbls HsMatchP where 
    freeAndDeclaredVbls (HsMatch loc1 (PNT fun _ _) pats rhs ds) 
     = do(pf, pd)<-hsFreeAndDeclared pats
         (rf, rd)<-hsFreeAndDeclared rhs
         (df, dd)<-hsFreeAndDeclared ds
         return ((pf `union` ((rf `union` df) \\ (dd `union` pd `union` [fun]))), 
                 (pd `union` rd `union` dd `union` [fun]))

instance FreeAndDeclaredVbls HsDeclP where 
   freeAndDeclaredVbls (Dec (HsPatBind loc p rhs ds)) 
    = do (pf, pd)<-hsFreeAndDeclared p
         (rf, rd)<-hsFreeAndDeclared rhs
         (df, dd)<-hsFreeAndDeclared ds 
         return (pf `union` ((rf `union` df) \\ (dd `union` pd)),
                (pd `union` rd `union` dd))

   freeAndDeclaredVbls (Dec (HsFunBind loc matches))
      =do fds <-mapM freeAndDeclaredVbls matches
          return (concatMap fst fds, concatMap snd fds)  
   
   freeAndDeclaredVbls _=return ([],[]) 


instance FreeAndDeclaredVbls HsModuleP where
   freeAndDeclaredVbls (HsModule loc modName exps imps ds)
      =hsFreeAndDeclared ds  

instance FreeAndDeclaredVbls [HsDeclP] where
   freeAndDeclaredVbls decls=hsFreeAndDeclared decls 

instance FreeAndDeclaredVbls HsExpP where  

      freeAndDeclaredVbls (Exp (HsLet decls exp))
              = do (df,dd)<- hsFreeAndDeclared decls
                   (ef,_)<- hsFreeAndDeclared exp 
                   return ((df `union` (ef \\ dd)),dd)
      freeAndDeclaredVbls  (Exp (HsLambda pats body))
               = do (pf,pd) <-hsFreeAndDeclared pats
                    (bf,_) <-hsFreeAndDeclared body
                    return ((bf `union` pf) \\ pd, pd)      
      freeAndDeclaredVbls _ = return ([],[])

instance FreeAndDeclaredVbls HsAltP where
      
   freeAndDeclaredVbls (HsAlt _ pat exp decls)
             = do (pf,pd) <- hsFreeAndDeclared pat
                  (ef,ed) <- hsFreeAndDeclared exp
                  (df,dd) <- hsFreeAndDeclared decls
                  return (pf `union` (((ef \\ dd) `union` df) \\ pd),pd `union` dd)

instance FreeAndDeclaredVbls HsStmtP where
      freeAndDeclaredVbls (HsLetStmt decls stmts) 
           =do (df,dd) <-hsFreeAndDeclared decls
               (sf,sd) <-hsFreeAndDeclared stmts
               return (df `union` (sf \\dd),dd)

      freeAndDeclaredVbls (HsGenerator _ pat exp stmts) -- Claus
             =do (pf,pd) <-hsFreeAndDeclared pat
                 (ef,ed) <-hsFreeAndDeclared exp
                 (sf,sd) <-hsFreeAndDeclared stmts
                 return (pf `union` ef `union` (sf\\pd),pd)

      freeAndDeclaredVbls _ =return ([],[])    
     

freeAndDeclaredNames t =do (f1,d1)<-freeAndDeclaredVbls t
                           return ((nub.map pNtoName) f1, (nub.map pNtoName) d1)


--Return those names (in String format) visible to pn in scope t.
hsVisibleNames pn t =do d<-hsVisibleVbls pn t
                        return ((nub.map pNtoName) d)

hsVisibleVbls t =applyTU (full_tdTU (constTU [] `adhocTU` mod 
                                                `adhocTU` exp
                                                `adhocTU` match
                                                `adhocTU` patBind
                                                `adhocTU` alt
                                                `adhocTU` stmts))
      where         
          mod ((HsModule loc modName exps imps decls)::HsModuleP)
            |findEntity t decls
           =do (df,dd)<-hsFreeAndDeclared decls
               return dd
          mod _=return []
                                                             
          exp ((Exp (HsLambda pats body))::HsExpP)
            |findEntity t body
             = do (pf,pd) <-hsFreeAndDeclared pats
                  return pd         

          exp (Exp (HsLet decls e))
             |findEntity t e || findEntity t decls
             = do (df,dd)<- hsFreeAndDeclared decls
                  return dd      
          exp _ =return [] 
                   
          match (m@(HsMatch _ (PNT fun _ _)  pats rhs  decls)::HsMatchP)
            |findEntity t rhs || findEntity t decls
            = do (pf,pd) <- hsFreeAndDeclared pats
                 (df,dd) <- hsFreeAndDeclared decls
                 return  (pd `union` dd `union` [fun])
          match _=return []
 
          patBind (p@(Dec (HsPatBind _ pat rhs decls))::HsDeclP)
            |findEntity t rhs || findEntity t decls 
             =do (pf,pd) <- hsFreeAndDeclared pat
                 (df,dd) <- hsFreeAndDeclared decls
                 return (pd `union` dd)
          patBind _=return [] 

          alt ((HsAlt _ pat exp decls)::HsAltP)
             |findEntity t exp || findEntity t decls
             = do (pf,pd) <- hsFreeAndDeclared pat
                  (df,dd) <- hsFreeAndDeclared decls
                  return (pd `union` dd)
          alt _=return []
                    
          stmts ((HsGenerator _ pat exp stmts) :: HsStmtP) -- Claus
            |findEntity t stmts
             =do (pf,pd) <-hsFreeAndDeclared pat
                 return pd

          stmts (HsLetStmt decls stmts)
            |findEntity t decls || findEntity t stmts
             =do (df,dd) <-hsFreeAndDeclared decls
                 return dd
          stmts _ =return []

------------------------------------------------------------------
--declsIn collects declarations in scope t ------

class (Term t) => DeclsIn t where
  
    declsIn :: t->[HsDeclI PNT]

instance DeclsIn HsMatchP where   
    declsIn (HsMatch loc1 fun pats rhs ds)=ds 

instance DeclsIn HsDeclP where
    declsIn (Dec (HsPatBind loc p rhs ds))=ds
    declsIn (Dec (HsFunBind loc matches))=concatMap declsIn matches  
    declsIn _ =[]

instance DeclsIn [HsDeclP] where
    declsIn ds=concatMap declsIn ds

instance DeclsIn HsModuleP where
    declsIn (HsModule loc modName exps imps ds)=ds

instance DeclsIn RhsP where
    declsIn rhs=fromMaybe [] (applyTU (stop_tdTU (failTU `adhocTU` inLet
                                                         `adhocTU` inAlt
                                                         `adhocTU` inStmt)) rhs) 
             where inLet ((Exp (HsLet ds e)) ::HsExpP)=Just ds
                   inLet _ =mzero

                   inAlt ((HsAlt _ p rhs ds)::HsAlt HsExpP HsPatP [HsDeclP])=Just ds

                   inStmt ((HsLetStmt ds _)::HsStmt HsExpP HsPatP [HsDeclP])=Just ds
                   inStmt _=mzero

instance DeclsIn HsExpP where
    declsIn rhs=fromMaybe [] (applyTU (stop_tdTU (failTU `adhocTU` inLet
                                                         `adhocTU` inAlt
                                                         `adhocTU` inStmt)) rhs) 
             where inLet ((Exp (HsLet ds e)) ::HsExpP)=Just ds
                   inLet (Exp (HsListComp (HsLetStmt ds stmts)))=Just ds
                   inLet (Exp (HsDo (HsLetStmt ds stmts)))=Just ds
                   inLet _ =Nothing

                   inAlt ((HsAlt _ p rhs ds)::HsAlt HsExpP HsPatP [HsDeclP])=Just ds

                   inStmt ((HsLetStmt ds _)::HsStmt HsExpP HsPatP [HsDeclP])=Just ds
                   inStmt _=Nothing 

instance DeclsIn HsStmtP where
    declsIn (HsLetStmt ds stmts)=ds
    declsIn  _ = []
instance DeclsIn HsAltP where
    declsIn (HsAlt _ p rhs ds)=ds

-- *****************END OF COLLECTING FUNCTIONS********************************

hsDeclaredConstrs=(applyTU (stop_tdTU (failTU `adhocTU` pnt)))
       where  pnt (PNT pn@(PN name (G mod id (N (Just (SrcLoc file1 row1 col1)))))
                       (ConstrOf _ _)  (N (Just (SrcLoc file2 row2 col2))))
                 |file1==file2 && row1==row2 && col1==col2
                =return [pn]

              pnt (PNT pn@(PN name (G mod id (N (Just (SrcLoc file1 row1 col1)))))
                      (Type _)  (N (Just (SrcLoc file2 row2 col2))))
                 |file1==file2 && row1==row2 && col1==col2
                =return [pn]   
                   
              pnt _ =mzero

  
class (Term t) =>RhsIn t where
  
    rhsIn:: t->Maybe RhsP

instance RhsIn HsMatchP where   
    rhsIn (HsMatch loc1 fun pats rhs ds)=Just rhs
           
instance RhsIn  HsDeclP where
    rhsIn (Dec (HsPatBind loc p rhs ds))=Just rhs                
    rhsIn _ =Nothing

instance RhsIn HsModuleP where
    rhsIn mod=Nothing

instance RhsIn HsAltP where
    rhsIn (HsAlt loc p rhs ds)=Just rhs

expIn ((Exp (HsLet ds exp))::HsExpP)=Just exp
expIn _=Nothing

stmtsIn ((HsLetStmt ds stmts):: HsStmtP)=Just stmts
stmtsIn _=Nothing
  

-- change the absolute locations of local variabls to relative ones.
toRelativeLocs e = runIdentity (applyTP (full_tdTP (idTP `adhocTP` inLoc)) e)
  where
    inLoc (SrcLoc f row col)
      | elem (row,col) locs = let index=fromJust (elemIndex (row,col) locs) + 1 
                              in return (SrcLoc f index index)
    inLoc loc@(SrcLoc f row col) = return loc

    locs = getDefLocs e

    getDefLocs = (nub.ghead "toRelativeLoc").applyTU (full_tdTU (constTU [] `adhocTU` inPnt))
        where
         inPnt pnt@(PNT pn ty loc)
            |not (pNTinUseLoc pnt) = return [pNameToLoc pn]
         inPnt _ = return []
 

 
