module HsModule(module HsModule,ModuleName(..)) where
import List(sort,nub)
import SrcLoc
import HsName(ModuleName(..))
import HsIdent(HsIdentI(..))
import HsConstants(prelude_mod)

data HsModuleI i ds
    = HsModule { hsModSrcLoc  :: SrcLoc,
		 hsModName    :: ModuleName,
		 hsModExports :: Maybe [HsExportSpecI i],
		 hsModImports :: [HsImportDeclI i],
		 hsModDecls   :: ds }
      deriving (Eq, Show)

instance HasSrcLoc (HsModuleI i ds) where srcLoc = hsModSrcLoc

data HsExportSpecI i
    = EntE (EntSpec i)          -- only unqualified names allowed for i
    | ModuleE ModuleName        -- module M
    deriving (Eq, Show)

data HsImportDeclI i
    = HsImportDecl SrcLoc ModuleName
          {-qualified-} Bool {-as-} (Maybe ModuleName)
          (Maybe (Bool, [EntSpec i]))
      deriving (Eq, Show)

--hsImpFrom :: HsImportDecl -> ModuleName
hsImpFrom (HsImportDecl _ x _ _ _)  = x
hsModImportsFrom = sort . nub . map hsImpFrom . hsModImports

type HsImportSpecI i = EntSpec i -- qualified names allowed for i

data EntSpec i
    = Var i                   -- x (a variable identifier)
    | Abs i                   -- T, C (or P-Logic assertion/predicate)
    | AllSubs i               -- T(..), C(..)
    | ListSubs i [HsIdentI i] -- T(C_1, ..., C_n, f_1, ..., f_n)
                              -- C(m_1, ..., m_2)
    deriving (Eq, Show)

exportVar = EntE . Var

--------------------------------------------------------------------------------

--Functions for making the implicit import of Prelude explicit:

optAddPrelude prel = if prel then addPrelude else id

addPrelude m@HsModule{hsModImports=is} = m{hsModImports=prelimps++is}
  where
    prelimps = prelqual:if prelude_mod `elem` map hsImpFrom is
			then []
			else [prelunqual]

    prelqual = prel True
    prelunqual = prel False
    prel qual = HsImportDecl loc0 prelude_mod qual Nothing Nothing
