module HsName(module HsName) where

import Char(isAlpha,isUpper)
import List(intersperse)
import PrettyPrint
--import MUtils(( # ))

type Id = String

newtype ModuleName = Module String deriving (Eq, Ord)

instance Show ModuleName where
  showsPrec n (Module s) = showsPrec n s
instance Read ModuleName where
  readsPrec n s = [(Module m,r)|(m,r)<-readsPrec n s]

data HsName
    = Qual ModuleName Id
    | UnQual Id
      deriving (Eq, Ord, Show, Read)

hsUnQual = accHsName id

--- Map, Acc & Seq for the HsName functor --------------------------------------

mapHsName idf (Qual q id) = Qual q (idf id)
mapHsName idf (UnQual id) = UnQual (idf id)

accHsName idf (Qual q id) = idf id -- ??
accHsName idf (UnQual id) = idf id


--- Pretty printing for HsName and Module ---- ---------------------------------


--instance Show HsName where show = pp

instance Printable ModuleName where
    ppi (Module m) = ppi m
    wrap           = ppi

instance Printable HsName where
    ppi (Qual (Module "Prelude") "[]") = ppi "[]" -- hack
    ppi (Qual (Module "Prelude") "()") = ppi "()" -- hack
    ppi (Qual (Module "Prelude") "(,)") = ppi "(,)" -- hack
    ppi (Qual (Module "Prelude") "->") = ppi "(->)" -- hack
    ppi (Qual q id) = q <> '.' <> id
    ppi (UnQual id) = ppi id

    wrap n | isSymbolOp n = parens n
           | otherwise    = ppi n

instance PrintableOp HsName where
    isOp = isSymbolOp
    ppiOp n = if isAlphaOp n
	      then backQuotes n
	      else ppi n

isSymbolOp, isAlphaOp, isConOp :: HsName -> Bool
isSymbolOp = isSymbol . head . hsUnQual
isAlphaOp  = isAlpha  . head . hsUnQual
isConOp    =  (==':') . head . hsUnQual

ppInfixOp n = ppiOp n
ppInfixName n = ppiOp n

isSymbol c = elem c ":!#$%&*+./<=>?@\\^|-~"

------------

splitQualName = uncurry (Qual . Module) . splitQualName'

{-
-- Haskell 98 version:
splitQualName' s = Qual (Module m) n
  where (m,'.':n) = break (=='.') s
-}

-- For the "hierachical module name" extension:
splitQualName' s = (m,n)
  where
    m = concat (intersperse "." (init parts))
    n = last parts
    parts = chop s

    chop "" = []
    chop ('.':s@(c:_)) = if isUpper c
			 then []:chop s
			 else []:[s]
    chop (c:s) = 
	  case chop s of
	    [] -> [[c]]
	    s:ss -> (c:s):ss
