-- Type checking sequences of declarations
module TiDs where
import List(nub,(\\),intersect,partition)
import Maybe(mapMaybe,fromMaybe)

import HasBaseStruct(HasBaseStruct(..),GetBaseStruct(..))
import SrcLoc(HasSrcLoc(..))
import HsLiteral(HsLiteral)
import HsDeclStruct
import HsPatStruct(PI)
import TI hiding (Subst)
import TiNameMaps
import TiDkc(Dinst,Einst)
import TiSolve
import TiSCC(sccD)
import TiContextReduction(contextReduction,entails)
import TiInstanceDB(instType)
import TiDerivedInstances
import Substitute
import SubstituteBaseStruct
--import QualNames

import MUtils
import PrettyPrint

explicit kenv ctx ds = (concat kss,concat tss)
  where (kss,tss) = unzip $ map (explicitlyTyped kenv ctx) ds

instance DeclInfo i d => DeclInfo i [d] where
  explicitlyTyped  = explicit -- hmm
---
instance (
 TypeId i,ValueId i,Printable i, Fresh i,HasSrcLoc i,
 TypedId PId i,
 DeclInfo i d, FreeNames i d, TypeCheckDecl i d ds2,HasSrcLoc d,
 HasCoreSyntax i p,HasCoreSyntax i e,HasDef ds d,
 HasLit HsLiteral e,HasLit HsLiteral p,
 HasBaseStruct e (Einst i e p ds),HasBaseStruct p (PI i p),
 DefinedNames i d,AccNames i d,
 HasBaseStruct d (Dinst i e p ds), GetBaseStruct d (Dinst i e p ds),
 HasDef ds2 d2,  HasLocalDef i e2 d2,  Types i d2,
 HasAbstr i ds2, HasLocalDef i e2 ds2, Types i ds2, AddDeclsType i ds2,
 HasCoreSyntax i e2,
 MapExp e2 ds2,Subst i e2,HasTypeApp i e2, -- for dicts in mono rec calls
 KindCheck i d ()) => TypeCheckDecls i [d] ds2
  where
    tcTopDecls rewrite ds = decorate # tcDs rewrite True ds
    tcInstDecls mts ds = do ds':>:bs <- tcExplDs mts ds
			    return (addDeclsType ([],bs) ds')

--}

decorate (ds:>:(insts,dt)) = addDeclsType dt ds:>:(insts,dt)

--sccD ds = [ds] -- dummy

-- How to type check a (possibly recursive) group of declarations:
tcDs rewrite rec ds0 =
    do -- Infer the kinds of all class/type identifers, including
       -- type variables.
       -- (Names are assumed to have been made unique in an earlier pass.)
       ks <- kgeneralise $ do kbs <- kintro (allTypeNames ds0)
			      let kbs' = map (fmap (flip (,) undefined)) kbs
                              extendkts kbs' (kc ds0>>return kbs)
       -- Find additional info about classes and types
       -- and the types of all explicitly typed value identifiers...
       let (kinfo,expls) = explicit ks [] ds0
	   vkinfo = [v:>:(k,Tyvar)|v@(HsVar _):>:k<-ks]
       expl <- mapM checkTypeClash (collectTyped expls)
       -- ...and add them to the environment (if in a recursive group).
       extendkts (kinfo++vkinfo) $ opt rec (extendts expl) $
	do -- Extend the instance database with the declared instances
           (insts,ds) <- getInstances ds0
	   extendIEnv insts $
	    do -- Infer the types of the implicitly typed declarations,
	       -- one strongly connected component (scc) at a time...
	       let (dsImpl,dsExpl) = partition (isImplicit ns) ds
		   dsccs = sccD dsImpl
		   ns:>:_=unzipTyped expl
	       ds1:>:bs1 <- tcImplBGs rec expl dsccs
	       -- ...and add them to the environment (if in a recursive group)
	       opt rec (extendts bs1) $
		do -- Infer the types of the explicitly typed declarations
		   -- and check that they agree with their explicit types
		   ds2:>:bs2 <- tcExplDs expl dsExpl
		   let env = (kinfo,its++filter noDup bs0++expl)
		       its=map (instType ks) insts
		       bs0 = bs1++bs2
		       noDup (x:>:_) = x `notElem` ns
		   -- Finally, return the translated declarations, the
		   -- instances, and the resulting environment
		   ((ds1 `appendDef` ds2)>:(insts,env))
  where	
    getInstances ds =
        do modmap <- getModule
           kenv <- getKEnv
           let insts1 = instDecls ds modmap kenv
	   extendIEnv insts1 $
	    do (insts2,ds2) <- unzip # derivedInstances modmap kenv ds
	       return (insts1++insts2,ds++rewrite ds2)

    instDecls ds modmap env =
         [instDecl s optn ctx tp|HsInstDecl s optn ctx tp _<-mapMaybe basestruct ds]
       where
        instDecl s optn ctx tp = (syn tp,(n,map syn ctx))
	  where n = fromMaybe (instName' modmap s) optn
        syn = expandSynonyms env

    isImplicit ns d = any (`notElem` ns) (definedValueNames d)

-- Type check a sequence of implicitly typed binding groups:
tcImplBGs rec expl [] = noDef>:[]
tcImplBGs rec expl (ds:dss) =
  do ds1:>:bs1 <- tcImplBG rec expl ds
     ds2:>:bs2 <- extendts bs1 $ tcImplBGs rec expl dss
     ds1 `appendDef` ds2>:bs1++bs2

tcImplBG rec expl ds =
    checkExplicit rec (srcLoc ds) expl =<< tcBG expl rec False ds

-- Type check declarations with explicit type signatures
-- pre: the explicitly given types have already been added to the environment
tcExplDs expl ds =conc . unzipTyped # mapM (tcExplD expl) ds
  where conc (dss:>:bss) = concatDefs dss:>:concat bss

tcExplD expl d =
  checkExplicit False (srcLoc d) expl =<< tcBG expl False True [d]

-- Typecheck one (mutually recursive) binding group:
tcBG all_ebs rec isExpl ds =
  addErrorContext $
  do (dns,ds'):>:bs' <-
       generalise' keepambig unr mapTs $
       do bs1 <- tintro (values\\ens)
	  bs2 <- tintro ens
	  let bs=bs1++bs2
	      lbs = (map.fmap) mono bs1
	  (:>:bs) # opt rec (extendts lbs) (mapM (tcDecl bs) ds)
     (dns,concatDefs ds')>:bs'
  where
    ens=values `intersect` all_ens where all_ens:>:_=unzipTyped all_ebs
    (types,values) = definedNamesSplit ds
    unr = all (isUnrestricted isExpl) ds
    keepambig = all keepAmbigTypes ds

    mapTs = map . fmap
      --where mapTM f (x:>:t) = (x:>:) # f t
    --typed = uncurry (:>:)

    addErrorContext =
      if not (null types && null values)
      then errmap (render
	    (sep [(if unr then "In an un" else "In a ")
		   <>"restricted definition of",
		  nest 4 (fsep types $$ fsep values)<>":\n"])++)
      else id

opt b f = if b then f else id

checkExplicit rec loc explicit ((dns,d):>:inferred) =
    errmap ((pp loc++": type signature vs inferred type:\n")++) $
    do let xs:>:_ = unzipTyped inferred
	   ixs = xs \\ exs
           (exs,eqscs) =
	       unzip [ (x,(it,et)) |x:>:it<-inferred,y:>:et<-explicit,y==x]
           pick x = if x `elem` exs
		    then head [ t |t@(y:>:_)<-explicit,x==y]
		    else head [ t |t@(y:>:_)<-inferred,x==y]
           -- abstract' adjusts monomorphic recurive calls:
	   abstract' ixs dns d =
	       if rec
	       then abstract dns (esubst addDicts d)
	       else abstract dns d
	     where
	       addDicts x = if HsVar x `elem` ixs
			    then foldl1 app (x':map var dns)
			    else if HsVar x `elem` exs
			         then x' 
				 else var x -- Grr!! Type info is lost
		 where
		   x' = spec (HsVar x) sc (map tyvar (tdom gs))
		   _:>:sc@(Forall ags gs qt) = pick (HsVar x)
       if null eqscs then abstract' ixs dns d>:map pick xs
	 else do (ids,eds,eqts) <- freshInsts dns eqscs
		 s@(S su) <- matchEqns eqts =<< getKEnv
		 let ids' = apply s ids
		 -- Although the context ids should already be reduced,
                 -- the instance ids' may still not be in normal form,
		 -- so we have to apply context reduction again.
		 (ids'',r1) <- do contextReduction ids'
		 r2 <- eds `entails` ids'' =<< getKEnv
		 let edns:>:_ = unzipTyped eds
		     r = r2 . r1
		 ngvs <- tv # getTEnv
                 mapM_ constrain [tyvar v:=:t|(v,t)<-su,v `elem` ngvs]
		 abstract' ixs edns (r (apply s d))>:apply s (map pick xs)
  where
    freshInsts dns eqscs =
       do ((ctx,eds):_,eqts) <- unzip # mapM freshInst' eqscs
	  return (zipTyped (dns:>:ctx),eds,eqts)

    freshInst' (isc,esc) =
       do -- Don't freshen the inferred type, the quantified tvars can appear in the body:
          --(ids,it) <- freshInst isc
          --let _:>:ctx = unzipTyped ids
          let Forall _ _ (ictx:=>it) = isc
	      Forall _ _ (ectx:=>et) = esc
          --(eds,et) <- freshInst esc
          ds <- map dictName # freshlist (length ectx)
	  let eds = zipTyped (ds:>:ectx)
          return ((ictx,eds),(it,et))

checkTypeClash (x:>:ts0) =
    case nub ts0 of -- Equality of type schemes. Hmm...
      [t] -> x>:t
      ts -> fail (render msg)
	where
	  msg = "Conflicting type signatures for " <+> x $$ vcat ts
