module PfeBrowserCerts where
import Prelude hiding (putStr,putStrLn,print,readFile,ioError,catch,readIO)
import AllFudgets as Fud hiding (getModificationTime)
import Maybe(isJust)
import MUtils
import AbstractIO
import DirUtils(latestModTime)
import List((\\))

import PfePlumbing
import PfeBrowserMonad
import PPU(pp{-,ppi,ppu,withPPEnv,($$),(<>),(<+>),PPHsMode-})
import Pfe0Cmds(getSubGraph)
import PFE0
import PFE_Certs
import CertServers(certFiles,lookupServer)
import PfePropCmds(tassertionSignature)
import CertAttrs(label)
import DrawLex(cstatusIcon,certIcon)
import QualNames(getQualified)
import TypedIds

{-+ #evidenceGfx#: this function computes the table that is shown in the
Evidence window. Every existing certificate, certificate annotation and
assertion will be mentioned in one of the table entries.
-}
evidenceGfx o dir (bad,icons) certs asserts =
    htableD hdrs (concatMap certD certs++concatMap missingD missing)
  where
    hdrs = ["Icon","Name","Sequent (or other info)"]
    missing = [a|(a,_)<-asserts,pp a `notElem` concs]
    concs =[conc|Just conc<-map (lookup "conc" @@ fst.snd) certs]

    missingD a = [g bad,createCertD "" s,g ("? |-"++s)]
      where s = pp a

    certD (cert,(optattrs,status)) =
      case optattrs of
	Nothing -> [g bad,certNameD cert,
		    g "certificate does not exist (not created yet?)"]
	Just attrs ->
	  case (`lookup` icons) =<< lookup "type" attrs of
	    Just cicons ->
		[certIconD cert icon,certNameD cert,g sequent]
	      where
		icon = cstatusIcon cicons status
	    _ -> [g bad,certNameD cert,g "bad cert/unknown cert type"]
	  where
	    conc = fromMaybe "-" (lookup "conc" attrs)
	    hyp = fromMaybe "" (lookup "hyp" attrs)
	    --sequent = conc++" -| "++hyp
	    sequent = hyp++" |- "++conc

    certIconD cert icon = labelD (CertLbl cert) (g icon)


{-+ #getAInfo#: compute the extra info that is shown when the user clicks
on an assertion name.
-}
getAInfo (n,Assertion) =
  do allcs <- certs # getStBr
     cicons <- icons # getStBr
     let a = pp (getQualified n)
	 certs = [c|c@(_,(Just attrs,_))<-allcs,lookup "conc" attrs==Just a]
	 newc = head ((a:[a++show n|n<-[1..]])\\map fst allcs)
	 cis = [(cert,certIcon cicons cert)|cert<-certs]
     return [hboxcD' 10 (g "Certificates:":
			 if null certs
			 then [g "none.",createCertD newc a]
			 else map certNameIconD cis)]
getAInfo _ = return []

{-+: #showCertInfo#: computes the certificate information displayed when the
user clicks on a certificate icon. This includes checking whether the
certificate is valid, invalid or outofdate.
-}
showCertInfo cert =
    do m <- modname # getStBr
       certinfo <- getCertDescription m cert
       pfePut (toRefs (linesD (briefTxt certinfo)))
       putInfoWindow (CertInfo,fullGfx certinfo)
       popupInfoWindow (CertInfo,True)
  where
    briefTxt (Left (msg,actionD)) = [g msg,actionD]
    briefTxt (Right (fields,_)) = map (g.join) (take 3 fields)
      where join (x,y) = x++": "++y

    fullGfx (Left (msg,actionD)) = vboxD [Fud.g msg,actionD]
    fullGfx (Right (fields,actionD)) =
	vboxD [htableD ["Field","Value"] flatfields,actionD]
      where flatfields = concatMap (\(x,y)->[g x,g y]) fields
					    
    getCertDescription m cert =
      withWaitCursor $ 
      do let msg =g ("Checking status of certificate "++cert)
         putInfoWindow (CertInfo,msg)
	 pfePut (toRefs msg)
         dir <- checkProject
	 servers <- certServers # getStBr
	 (revalidate,status,info) <- certUpdStatus dir m =<< certInfo m cert
	 --let status = "Unknown"
         return (certDescription m revalidate status servers info)

    certDescription _ _ _ _ (cert,(Nothing,_)) =
        Left (cert++": certificate does not exist (not created yet?)", actionD)
      where
        actionD = createCertD cert ""

    certDescription m revalidate status' servers (cert,(Just attrs,status)) =
      Right ([("Certificate",(cert++" :: "++ctype)),
              ("Current Status",status')]
--	     ++ cfield "Certifies" "conc"
--	     ++ cfield "Depends on" "hyp"
	     ++ showStatus status
	     ++ concat [cfield (label attrs) name|(name,attrs)<-cattrs]
	     ++ cfield "Created by" "who"
	     ++ sfield "About this certificate type" "description",
	     actionD)
      where
	showStatus Nothing = cfield "Date" "date"
	showStatus (Just (valid,when)) =
          [("Marked "++(if valid then "" else "in")++"valid on",show when)]

        ctype = fromMaybe "missing!!" (lookup "type" attrs)
	optserver = lookupServer servers ctype
	sattrs = maybe [] serverAttrs optserver
	cattrs = maybe [] certAttrs optserver
        cfield = field attrs
	sfield = field sattrs
        field attrs hdr name = maybe [] ((:[]).((,) hdr)) (lookup name attrs)

        actionD = hboxD' 20 (revalD++editD++removeD)
          where
	    revalD  = if revalidate then [cmdD reval "Validate"] else []
	    removeD = [cmdD rm "Remove"]
	    editD =
	      if isJust optserver
	      then [labelD (CertEdit (ctype,(cert,attrs))) (ulinkD (g "Edit"))]
	      else []

            cmdD cmd lbl = labelD (CertCmd cmd) (ulinkD (g lbl))

            reval = "cert validate "++mcert
	    rm    = "cert rm "      ++mcert
	    mcert = quote (pp m++"/"++cert)

    --showType = pfePut . toRefs @@ getType

revalidateAll =
    withWaitCursor $
    do dir <- checkProject
       m <- modname # getStBr
       mapM_ (certUpdStatus dir m @@ certInfo m . fst) . certs =<< getStBr

certUpdStatus dir m info@(cert,(a@(Just attrs),Just (valid,when))) =
    ifM auxchange (revalidate "auxiliary file changed/missing") $
    ifM srcchange
	(if valid
	 then certStatusFromSig
	 else revalidate "was invalid, some source file has been changed") $
    return (not valid,if valid then "Valid" else "Invalid",info)
  where
    auxchange = let auxfiles = certAuxFiles m cert attrs dir
		in anyChangeSince when auxfiles
    srcchange = do srcfiles <- map fst # getSubGraph [m]
		   anyChangeSince when srcfiles -- not cached??

    revalidate why = return (True,"Needs revalidation ("++why++")",info)

    isUnconditionalCert =
      case (lookup "conc" attrs,lookup "hyp" attrs) of
	(Just conc,Just "..") -> Just conc
	_ -> Nothing

    certStatusFromSig =
	maybe unknownStatus uncondCertStatusFromSig isUnconditionalCert

    unknownStatus = return (True,"unknown",info)

    uncondCertStatusFromSig conc =
      do optcertsig <-
	   maybeM (readIO =<< readFile (certDir m cert dir++"/deps"))
	 case optcertsig of
	   Nothing -> return (True,"New, not yet validated",info)
	   Just certsig ->
	     do srcsig <- tassertionSignature ((m,conc),Nothing)
		if certsig==srcsig
		  then do markValid m cert
			  st <- certStatus m cert
			  return (False,"Valid",(cert,(a,st)))
		  else revalidate "source changed"

    -- vaid/invalid status missing:
certUpdStatus _ _ info@(_,(Just _,_)) =
    return (True,"New, not yet validated",info)

    -- attributes missing:
certUpdStatus _ _ info = return (False,"unknown (attributes missing?!)",info)

quickCertsStatus m certs dir =
  do srcdate <- latestModTime . map fst =<< getSubGraph [m]
     mapM (quickCertStatus m srcdate dir) certs

quickCertStatus m srcdate dir cert =
  do (optattrs,status) <- snd # certInfo m cert
     case (,) # optattrs <# status of
       Nothing -> return Nothing
       Just (attrs,status@(valid,when)) ->
	 if srcdate>when
	 then return Nothing
	 else ifM (anyChangeSince when (certAuxFiles m cert attrs dir))
		  (return Nothing)
		  (return (Just status))

--allCertFiles m cert dir = (++) # certAuxFiles m cert dir <# certSrcFiles m
certAuxFiles m cert attrs dir = certAttrsPath m cert dir:certFiles attrs
--certSrcFiles m = map fst # getSubGraph [m]

createCertD cert concl = labelD (CreateCert (cert,concl))
			        (ulinkD (g "Create a new certificate!"))

certNameD cert = labelD (CertLbl cert) (ulinkD (g cert))

certNameIconD ((c,_),icon) = labelD (CertLbl c) (hboxD [ulinkD (g c),g icon])

ulinkD s = fgD "blue" $
           spacedD centerS $
           stackD [spacedD bottomS (g (hFiller 1)),s]

quote s = q++s++q where q="\""

htableD hdrs cells = tableD' 0 (length hdrs) (map b hdrs++map f cells)
  where
    b x = stackD [fgD "grey" (g $ filler False False 5),padD 3 (g x)]
    f x = stackD [fgD "grey" (g frame),padD 3 x]
    --f x = padD 3 x

linesD = fontD defaultFont . vboxlD' 0 -- . map g

anyChangeSince when paths =
   ifM (allM doesFileExist paths)
       ((when<) # latestModTime paths)
       (return True) -- a file has been deleted
