module PfeBrowserGUI(module PfeBrowserGUI,posLabel) where
import List(groupBy,sortBy)
--import Char(isSpace)

import AllFudgets
import FudExtras(popupMenuF')
import ContribFudgets(menuIcon,item,itemValue)--menuBarF,delayedSubMenuItem,idT,cmdItem
import HyperGraphicsF2
import TreeBrowser as B(treeBrowserF',Tree(..))
import HistoryButtonsF
import ListUtil(chopList,breakAt,mapFst)

import PfePlumbing
import PfeBrowserColors as C
import DrawLex(drawLex,fakeLex,posLabel,rootLabel,isCertAnnot)

import HsLexerPass1(lexerPass0)
import HsTokens as T
import HsModule(ModuleName(..))

import OpTypes(cmpBy)
import PrettyPrint(pp)
  
newCertPopupF :: F ToNewCertPopup FromNewCertPopup
newCertPopupF =
    snd>^=<inputPopupF title newCertF Nothing >=^< pre
  where
    title = "Create a new certificate"
    pre inp = (Just title,Just inp)

newCertF = tableF 2 (typeF `inputPairF` (nameF `inputPairF` conclF))
  where
    typeF = labelF "Type" >*<
              (inputChange>^=<dynF nullLF
               >=^< Left . radioButtonsF>=^^<idempotSP>=^<map certTypeItem)
    radioButtonsF [] = labelF "No certificate types available?!"
    radioButtonsF items@(item1:_) =
      radioGroupF [(itemValue item,item)|item<-items] (itemValue item1)

    nameF = labelF "Name" >*< stringF
    conclF = labelF "Conclusion" >*< stringF

    certTypeItem (cert,(icon,_,_)) = item cert (hboxD [g icon,g cert])

graphDispF :: F ToGraph FromGraph
graphDispF =
    mapFilterSP leaf>^^=<
    vScrollF (treeBrowserF' g0)>=^<graphTree.sortBy (cmpBy (fst.snd))
  where
    g0 = graphTree []
    leaf (B.Leaf l) = Just (itemValue l)
    leaf _ = Nothing

    graphTree g =
	Node "Module Graph" [Node "Files" (filepaths2trees g),
			     Node "Modules" (map mitem g)]
      where
	mitem n@((_,(m,_))) = B.Leaf (item n (pp m))

    filepaths2trees = list2trees . map withpath
      where
	list2trees = map tree . groupBy eqHead . sortBy (cmpBy fst)

	--tree ([]:_) = ??
	tree [([f],p)] = B.Leaf (item p f)
	tree ((x:xs,p):ys) = Node x (list2trees ((xs,p):mapFst tail ys))

	eqHead (x:xs,_) (y:ys,_) = x==y
	eqHead ([],_) ([],_) = True
	eqHead _ _ = False

	withpath n@(path,_) = (chopList (breakAt '/') path,n)

refDispF :: F ToRefs InfoLbl
refDispF = scrollF (hyperGraphicsF' pm initD>=^<Left)
  where pm = setSizing Dynamic . setFont refFont
        initD = padD 20 $ g $ "Welcome to the "++myname++"!"

nodeInfoF :: F ToNodeInfo FromNodeInfo
nodeInfoF =
    --simpleGroupF [] $
    vBoxF (fileF>|<hBoxF (moduleF>|<dynImportMenusF))
     >=^< \ ((f,(n,is)),es) -> (f,(n,(is,es)))
  where
    fileF       = Right         >^=< ldispF aLeft  "File:"
    moduleF     = Left . Module >^=< (historyButtonsF>*<
                                      ldispF aRight "Module:")     >=^< pp
    --importsF    =                    dynModuleMenuF "Imports"
    --importedByF =                    dynModuleMenuF "Imported By"

    ldispF a l = l `labLeftOfF` strF
      where strF = stringInputF --' (setSizing Dynamic)

{-
    dynImportMenusF =
      Left >^=< dynF (importsMenuF ([],[])) >=^< Left . importsMenuF

    importsMenuF (is,es) = menuBarF menu
       where
         menu = [modmenu is "Imports",modmenu es "Imported By"]
	 modmenu ms = delayedSubMenuItem idT [cmdItem m (pp m)|m<-ms]

   dynModuleMenuF lbl = Left >^=< dynF (mF [])>=^< Left . mF
      where mF ms = menuF lbl [(m,pp m)|m<-take 30 ms]

--}
--{- 
    dynImportMenusF = dynModuleMenuF "Imports" >|<dynModuleMenuF "Imported By"
 
    dynModuleMenuF lbl = popupMenuF' 1 [] (menuLblF lbl)>=^<pre
      where pre ms = Left [(m,pp m)|m<-take 30 ms]
			   -- !! no indication if menu is truncated
--}

    menuLblF lbl = labelF' pm (hboxcD [g lbl,g menuIcon])
      where pm = setBorderWidth 1 . setBgColor paperColor

    fud1>|<fud2 = stripEither>^=<prodF fud1 fud2

sourceDispF :: F ToSource FromSource
sourceDispF =
    wCreateGCtx rootGCtx (gcFontA sourceFont) $ \ gctx ->
    loadColorsF gctx sourceDispF'

sourceDispF' colgc =
    scrollF (mousePointSP>^^=<hyperGraphicsF2' pm init)>=^<either replace id
  where
    pm = setFont sourceFont . setGfxEventMask [GfxButtonMask,GfxMotionMask]
    init = drawLex undefined undefined undefined colgc [] [] empty
    empty = lexerPass0 (replicate 24 '\n'++replicate 80 ' ') -- for initial size
    replace (dir,icons,(path,(m,_)),ts,rs,cs) =
      replaceGfx rootLabel $ drawLex dir icons m colgc rs cs ts

    mousePointSP = idempotSP -==- mapFilterSP label
      where
        label event =
          do (lbl,_):_ <- gfxEventPaths event
	     case event of
	       GfxButtonEvent {gfxType=Pressed,gfxButton=b} -> Just (Just b,lbl)
	       GfxMotionEvent {} -> Just (Nothing,lbl)
	       _ -> Nothing

highlightGfxs off on = ChangeGfx $ change off False++change on True
  where change paths on = [(path,(on,Nothing))|path<-paths]

hilite off on = Right (highlightGfxs off on)
mkvis lbl = Right (ShowGfx lbl (Nothing,Nothing))
setwaitcursor =  Right (ChangeGfxFontCursor 150)
setlinkcursor =  Right (ChangeGfxFontCursor 60)
setnormalcursor = Right (ChangeGfxCursor cursorNone)
showsrctext = Right . replaceGfx rootLabel . fakeLex

isCertTokenRef = isCertToken . fst
isCertToken token =
  do (NestedComment,(_,s)) <- return token
     isCertAnnot s


-- Moved file path definitions to ../../property/pfe/PFE_Certs.hs

sourceFont = argKey "sourcefont" defaultFont
refFont = argKey "reffont" defaultFont
