module PFE4(
  PFE4MT,PFE4Info,runPFE4Cmds,runPFE4,clean4,
  getSt4ext,updSt4ext,setSt4ext,
  typeCheckModule,rewriteAndTypeCheckModule,
  typeCheck,rewriteAndTypeCheck,topTypes,typeCheck',
  ) where
import Prelude hiding (putStrLn,readFile)
--import Monad(when)
import List(intersect)
import Maybe(fromMaybe,fromJust,isNothing)

import HsModule(HsModuleI,ModuleName(..),hsModName)
import HsModuleMaps(mapDecls)
import TiModule(tcModuleGroup,representative,monomorphismRestriction)
import TiPNT()
import TI(DeclsType,extendIEnv,extendEnv,(+++),errmap,Typing(..),unzipTyped)
import TiInstanceDB(Instance)

import PNT(PNT(..))
import TypedIds(IdTy(Value))
import UniqueNames(noSrcLoc)

import PFE3(parseScopeSourceFile)
import PFE3(PFE3MT,runPFE3Cmds,runPFE3,getSt3ext,updSt3ext,clean3)
import PFE0(Upd,epput,newerThan,optSubGraph,sortCurrentModuleGraph,projPath,projectDir,maybeF,withProjectDir)

import PrettyPrint
import qualified Lift
import MUtils
--import StateMT(getSt,updSt_)
import Graph(reachable)
import AbstractIO
import FileUtils(updateFile')
import DirUtils(optCreateDirectory,getModificationTimeMaybe,latestModTime,rmR)

type PFE4Info i ds = [(ModuleName,(ClockTime,
				   (Maybe (HsModuleI i ds),
				    ([Instance i],DeclsType i))))]
pfe4info0 = []::PFE4Info i ds

type PFE4MT n i1 i2 ds1 ds2 ext m = PFE3MT n i1 ds1 (PFE4Info i2 ds2,ext) m

runPFE4Cmds ext = runPFE3Cmds (pfe4info0,ext)
runPFE4 ext = runPFE3 (pfe4info0,ext)

getSt4 :: Monad m => PFE4MT n i1 i2 ds1 ds2 ext m (PFE4Info i2 ds2)
updSt4 :: Monad m => Upd (PFE4Info i2 ds2)->PFE4MT n i1 i2 ds1 ds2 ext m ()
getSt4ext :: Monad m => PFE4MT n i1 i2 ds1 ds2 ext m ext
updSt4ext :: Monad m => Upd ext->PFE4MT n i1 i2 ds1 ds2 ext m ()

getSt4 = fst # getSt3ext
updSt4 = updSt3ext . apFst 
setSt4 = updSt4 . const

getSt4ext = snd # getSt3ext
updSt4ext = updSt3ext . apSnd
setSt4ext = updSt4ext . const


typeCheckModule m = rewriteAndTypeCheckModule id m

rewriteAndTypeCheckModule rewrite m =
   fromJust.fst.snd.fromJust.lookup m # rewriteAndTypeCheck rewrite (Just [m])

rewriteAndTypeCheck f = rewriteAndTypeCheck' f True
typeCheck             = typeCheck' True
topTypes              = typeCheck' False

{-
typeCheck' :: ... => Bool -> [ModuleName] -> t m (PFE4Info i2 ds)
-}
typeCheck' = rewriteAndTypeCheck' id

rewriteAndTypeCheck' rewrite wantAST optms =
  do subsg <- flip optSubGraph optms # sortCurrentModuleGraph
     oldTypes <- getSt4
     types <- rewriteAndTypeCheck'' rewrite wantAST optms oldTypes subsg
     let newTypes = types++[old|old@(m,_)<-oldTypes,m `notElem` newms]
	 newms = map fst types
     setSt4 newTypes
     return types

rewriteAndTypeCheck'' rewrite wantAST optms oldTypes subsg =
    do optCreateDirectory `projPath` typesdir
       tcSccs [] [] subsg
  where
    subg = map snd (concat subsg)
    ms = fromMaybe (map fst subg) optms
    rms = map (fst.snd.head) subsg -- representative from each scc

    tcSccs types updated [] = return types
    tcSccs types updated (scc:sccs) =
	  do let sccms = map (fst.snd) scc
		 wanted = wantAST && (not.null $ sccms `intersect` ms)
	     (types',updated') <- tcScc wanted sccms types updated scc
	     tcSccs (types'++types) (updated'++updated) sccs

    tcScc wanted mns types updated scc@(_:_) =
        if wanted && any isNothing [fst.snd=<<lookup m oldTypes|m<-mns]
           || not (null updated_imports)
	then tcAgain
	else useOld
      where
	updated_imports = imports_approx `intersect` updated

        -- A coarse approximation to what is needed by a module:
        -- (Advantage: this is the right thing for instances)
        imports_approx = reachable subg mns

	useOld =
	  do srct <- latestModTime srcfs
	     let opt_oldtypes = mapM (flip lookup oldTypes) mns
		 oldtt = minimum.map fst # opt_oldtypes
	     if srct `newerThan` oldtt
		then do tt <- maybeF getModificationTimeMaybe typesf
	                if srct `newerThan` tt
			  then tcAgain
			  else do --pput $ "Using old types:"<+>fsep mns
                                  types' <- readTypeFiles
				  let t = fromJust tt
				  -- !!! Type info is duplicated here:
				  return ([(m,(t,(Nothing,types')))|m<-mns],[])
		else return (zip mns (fromJust opt_oldtypes),[])

	tcAgain =
	  do ms <- parseSCM scc
             epput $ "Type checking:"<+>fsep mns
             tms:>:types' <- tcScc' ms
	     optdir <- projectDir
	     (updated,t) <-
               case optdir of
		 Nothing ->
	           do --No access to the old types here, so we have to assume
	              --they have changed!! (This should be easy to improve.)
                      t <- getClockTime
		      return (True,t)
		 Just dir ->
		   do updated <- updateTypeFiles dir types'
		      t <- getModificationTime (typesf dir)
		      return (updated,t)
	     -- !!! Type info is duplicated here:
	     return ([(hsModName tm,(t,(Just tm,types')))|tm<-tms],
		     if updated then mns else [])


        tcScc' ms =
           do mono <- (/=Just "no") # maybeM (getEnv "PFE_monomorphism")
	      Lift.lift (tcScc'' mono ms)

	tcScc'' mono ms@(_:_) =
	    monomorphismRestriction mono $
	    extendIEnv (concat insts) $ extendEnv (concenvs envs) $
	    errmap (ctx++) $ tcModuleGroup rewrite ms
	  where
	    (insts,envs) = unzip [mt|(m,(_,(_,mt)))<-types,
				     m `elem` imports_approx,m `elem` rms]
	    concenvs = foldr (+++) ([],[])

	srcfs = map fst scc
	mn = representative mns -- pick one representative from the scc
	kindsf = kindsFile mn
	typesf = typesFile mn
	instsf = instsFile mn
	ctx = pp ("In module(s)"<+>fsep mns)++":\n"

        updateTypeFiles dir (insts,(kinds,types)) =
	  do uk <- updateFile' (kindsf dir) (showTypings kinds)
	     ut <- updateFile' (typesf dir) (showTypings types)
             ui <- updateFile' (instsf dir) (show insts)
	     return (uk||ut||ui)

        readTypeFiles =
	  do dir <- fromJust # projectDir
             kinds <- readTypings # readFile (kindsf dir)
	     types <- readTypings # readFile (typesf dir)
	     insts <- read # readFile (instsf dir)
	     return (insts,(kinds,types))

    parseSCM = mapM ((mapDecls rewrite . snd # ). parseScopeSourceFile .fst)

showTypings xys = shl (length xs).shls (map (fmap pn) xs) . shls ys $ ""
  where
     shls xs r = foldr shl r xs
     shl x = shows x.showChar '\n'
     xs:>:ys = unzipTyped xys
     pn (PNT n _ _) = n -- hmm

readTypings s = rdls (lines s)
  where
     rdls (n:ls) = uncurry zipTypings (splitAt (read n) ls)
     zipTypings [] _ = []
     zipTypings (x:xs) ~(y:ys) = (fmap pnt (read' x):>:read' y):zipTypings xs ys
     pnt n = PNT n Value noSrcLoc -- hmm!

read' s =
  case reads s of
    (x,""):_ -> x
    _ -> error $ "no parse: "++take 60 s

--------------------------------------------------------------------------------
clean4 = withProjectDir clean
  where
    clean dir = do rmR [typesdir dir]
		   clean3

--------------------------------------------------------------------------------

typesdir dir=dir++"types/"

typesFile (Module s) dir = typesdir dir++s++".types"
kindsFile (Module s) dir = typesdir dir++s++".kinds"
instsFile (Module s) dir = typesdir dir++s++".insts"
