-- Programatica Front-End Commands, level 0
module Pfe0Cmds where
import Prelude hiding (print,putStr,putStrLn,catch)
import List((\\))
import Monad(unless)

import PFE0

import HsName(ModuleName(..))
import HsConstants(main_mod)

import CmdLineParser
import PrettyPrint
import MUtils
import Graph(reverseGraph,reachable)
import DirUtils(optCreateDirectory)
import AbstractIO

pfe0 parseModule args = runPFE0Cmds () pfe0Cmds parseModule args

pfe0Cmds = projman++graphqueries

addHelpCmd prg cmds0 = cmds where cmds=cmds0++helpCmd prg cmds

helpCmd prg cmds =
  [("help",(Null (putStrLn (usage prg cmds)),"list available commands"))]

projman =
  -- Project management (a project is a set of source files)
  [("new"     , (fileArgs   new, "create a new project containing (zero or more) <files>")),
   ("add"     , (fileArgs   add, "add files to the project")),
   ("remove"  , (fileArgs   remove, "remove files from the project")),
   ("prune"   , (moduleArgs prune,"remove unreachable modules from the project")),
   ("files"   , (Null       files,"list files in the project")),
   ("options" , (Null       options,"show options in effect"))]

graphqueries =
  -- Module graph queries
  [("modules" , (moduleArgs modules,"show a topologically sorted list of modules")),
   ("graph"   , (moduleArgs graph,"show module dependecy (sub)graph")),
   ("dotgraph", (moduleArgs dotgraph,"dot format module dependency graph")),
   ("revgraph", (moduleArgs revgraph,"show reverse module dependecy (sub)graph")),
   ("unused"  , (moduleArgs unused,"list unimported and unreachable modules")),
   ("file"    , (moduleArg  file,"which file is the module in")),
   ("module"  , (fileArg    module',"which module does the file contain"))]

moduleArg f = moduleArgs (mapM_ f)
moduleArgs f = Args "<modules>" (f @@ mapM checkModuleName)

checkModuleName s =
  do ms <- map (fst.snd) # getCurrentModuleGraph
     let m = Module s
     unless (m `elem` ms) $ fail (s++": unknown module")
     return m

just ms = if null ms then Nothing else Just ms

--- Project management ---------------------------------------------------------
new args    = do newProject; addPaths args
add args    = addPaths args
remove args = removePaths args

files = putStr . unlines =<< allFiles

options = print =<< parserFlags

--- Module graph queries -------------------------------------------------------
modules ms = do g <- getSortedSubGraph ms
	        let sccs = map (map (fst.snd)) g
	        putStrLn "Topologically sorted strongly connected components:"
	        putStr (unlines (map (unwords.map pp) sccs))

graph = showGraph id
revgraph = showGraph reverseGraph
showGraph f ms = do g <- f . map snd # getSubGraph ms
		    pput $ vcat [m<>":"<+>fsep is|(m,is)<-g]

getSubGraph ms = concat # getSortedSubGraph ms
getSortedSubGraph ms = flip optSubGraph (just ms) # sortCurrentModuleGraph

-- Contributed by Claus Reinke:
dotgraph = showDotGraph id
showDotGraph f ms = 
  do g <- f . map snd # getSubGraph ms
     pput $   "digraph ModuleGraph {size=\"7.5,10.5\";ratio=fill;"
           $$ vcat [show m<>"->{"<+>fsep [show i<>";"|i<-is]<>"}"|(m,is)<-g]
           $$ "}"
 
file m =
    do g <- getCurrentModuleGraph
       pput $ vcat [m<>":"<+>f|(f,(m',_))<-g,m'==m]

module' f =
    do g <- getCurrentModuleGraph
       pput $ vcat [f<>":"<+>m|(f',(m,_))<-g,f'==f]

getUnused ms =
  do fg <- getCurrentModuleGraph
     let g = map snd fg
         unimported = map fst g \\ [m|(m,_:_)<-reverseGraph g]
	 roots = if null ms then [main_mod] else ms
	 r = reachable g roots
	 unreached = map fst g \\ r
     return (fg,roots,unimported,unreached)

unused ms =
  do (_,roots,us,unr) <- getUnused ms
     if null us -- unlikely...
       then putStrLn "All modules are imported somewhere"
       else pput $ "The following modules are never imported:" $$
	           nest 4 (fsep us)
     unless (null unr) $
       do pput $
	    sep [ppi "The following modules are unreachable from",
		 nest 2 (fsep roots<>":")]
	  pput $ nest 4 (fsep unr)

prune ms =
  do (fg,_,_,unreached) <- getUnused ms
     remove [f|(f,(m,_))<-fg,m `elem` unreached]
