module PfeDepCmds where
import Prelude hiding (print)
import List(nub,intersect)
import Char(isUpper)
import Monad(unless)

import HsName
import HsIdent(getHSName)
import HsConstants(main_mod)
import SourceNames(SN(..))
import SrcLoc(loc0)
import PFEdeps(depModules',tdepModules')
import Pfe0Cmds(moduleArgs,checkModuleName,just,getSubGraph)
import PFE0(allModules,pput,getCurrentModuleGraph)
import PFE2(getModuleExports)
import AbstractIO
import MUtils
import UniqueNames(HasOrig(..),PN(..),Orig(..),noSrcLoc,origModule)
import PNT(PNT(..))
import TiPNT() -- instances for PNT
import TiNames(instName)
import TypedIds(IdTy(..),idTy)
import Ents(Ent(..))
import Relations(applyRel)
import CmdLineParser
import Graph(graph,nodes,reachable,isReachable,listReachable,reverseGraph)
--import OrdGraph --slower than Graph!
import PrettyPrint

import PFE3(refsModule)
import RefsTypes
import ConvRefsTypes

pfeDepCmds =
    [("deps" ,        (moduleArgs deps,
	               "compute dependency graph for definitions")),
     ("tdeps" ,       (moduleArgs tdeps,
	               "compute dependency graph for definitions")),
     ("dotdeps" ,     (moduleArgs dotdeps,
	               "dot format dependency graph for definitions")),
     ("tdotdeps" ,    (moduleArgs tdotdeps,
	               "dot format dependency graph for definitions")),
     ("needed",       (qualIds needed,"needed definitions")),
     ("tneeded",      (qualIds tneeded,"needed definitions")),
     ("neededmodules",(qualIds neededmodules,
		       "names of modules containing needed definitions")),
     ("tneededmodules",(qualIds tneededmodules,
		       "names of modules containing needed definitions")),
     ("dead",         (qualIds dead,"dead code (default: Main.main)")),
     ("tdead",        (qualIds tdead,"dead code (default: Main.main)")),
     ("uses",         (qualId uses,"find uses of an entity"))]

qualIds f = Args "<M1.x1 ... Mn.xn>" (f @@ mapM parseQualId)

parseQualId s =
    if isQual s
    then let (m,n) = splitQualName' s
         in (,) # checkModuleName m <# return n
             -- TODO: also check that m.n is defined!
    else fail $ "Qaulified name required: "++s
  where
    isQual s =
      case break (=='.') s of
	(c:_,'.':_:_) -> isUpper c
	_ -> False

qualId f = Args "<M.x>" (f @@ parseQualId @@ one)
  where
     one [q] = return q
     one qs = fail $ "Exactly one qualified name is required: "++unwords qs

needed = needed' depModules'
tneeded = needed' tdepModules'

needed' dm qids =
  do (_,_,used) <- snd # depgraph' dm qids
     pput $ ppOrigs (listReachable used)

neededmodules = neededmodules' depModules'
tneededmodules = neededmodules' tdepModules'

neededmodules' dm qids =
  do (_,_,used) <- snd # depgraph' dm qids
     pput $ fsep (nub $ map origModule (listReachable used))

dead = dead' depModules'
tdead = dead' tdepModules'

dead' dm [] = dead'' dm [(main_mod,"main")]
dead' dm qids = dead'' dm qids

dead'' dm qids =
  do (g,ids,used) <- snd # depgraph' dm qids
     let unused = [n|n<-nodes g,not (isReachable used n)]
     pput $ fsep (map ppOrig unused)

uses q@(m,n) =
  do ms <- flip reachable [m] . reverseGraph . map snd # getCurrentModuleGraph
     --qs <- origpnt q
     --print qs
     mapM_ (usesIn q) ms

usesIn (Module mn,n) m =
  do refs <- simplifyRefsTypes' # refsModule m
     let uses = [ p |(p,(_,r,_,os))<-refs,
                   not (isDef r),
		   (mn,n) `elem` [ o |(_,Left o)<-os]]
     unless (null uses) $ pput (m<>":"<+>fsep uses)

intersects xs ys = not . null $ intersect xs ys

depgraph = depgraph' depModules'
tdepgraph = depgraph' tdepModules'

depgraph' depModules' qids =
  do let ms = nub (map fst qids)
     deps <- depModules' . Just . map (fst.snd) =<< getSubGraph ms
     ids <- concatMapM origpnt qids
     let g = depGraph [(n,ns)|(m,(t,ds))<-deps,(ns1,(ns,h))<-ds,n<-ns1]
         always = [n|(m,(t,ds))<-deps,([],(ns,h))<-ds,n<-ns]
	 used = reachable g (nub (ids++always))
     return (deps,(g,ids,used))

depGraph = graph . mapSnd concat . collectByFst
                           -- because of names from typesigs...

origpnt (m,n) =
  do (t,rel) <- getModuleExports m
     return $ case map ent2pnt (applyRel rel (SN n loc0)) of
                [] -> [pnt m n]
                ns -> ns
  where
    ent2pnt (Ent m i ty) = pnt' (conv # ty) m n
      where SN n _ = getHSName i
	    conv (SN n s) = PN n (S s)
    pnt = pnt' Value
    pnt' ty m n = PNT (PN (Qual m n) (g m n)) ty noSrcLoc
    g m n = G m n noSrcLoc

dotdeps = pput.pdotdeps @@ depModules' . just
tdotdeps = pput.pdotdeps @@ tdepModules' . just

pdotdeps deps =
    "digraph DepGraph"$$
    braces ("size=\"7.5,10.5\";ratio=fill;"$$
            vcat [p d<>"->"<>braces (fsep [p f<>";"|f<-fs])
                  |(m,(t,mdeps))<-deps,(ds,(fs,h))<-mdeps,d<-ds])
  where
    p = doubleQuotes . ppOrig

deps = pput.vcat.map pdeps @@ depModules' . just
tdeps = pput.vcat.map pdeps @@ tdepModules' . just

pdeps  (m,(t,deps)) =
  sep ["module"<+>m<>":",nest 2 (vcat $ map pdep deps)]
  where
    pdep (ds,(fs,h)) = sep [fsep ds <> ":",nest 2 (fsep (map ppOrig fs))]
    ppOrig = ppOrig' (Just m)

ppOrig = ppOrig' Nothing

ppOrig' optm n =
    if Just m'==optm then ppi x else m'<>"."<>x
  where
    (m',x) = origId n

origId n =
    case orig n of
      G m' n' _ -> (m', n')
--    I m' loc  -> (m',instName m' loc)
      _ -> error ("Bug: PfeDepCmds.origId "++show n)

ppOrigs qs = vcat [m<>":"<+>fsep xs|(m,xs)<-collectByFst (map origId qs)]
